
#include "precomp.h"
#include "chat.h"
#include "chatline.h"
#include "ircconnection.h"

/////////////////////////////////////////////////////////////////////////////
// Chat construction:

Chat::Chat(IRCConnection *connection, const std::string &filter, CL_Component *parent) :
	CL_Component(parent),
	connection(connection),
	filter(filter),
	font("InputBox/font", get_style_manager()->get_resources())
{
	slots.connect(sig_paint(), this, &Chat::on_paint);
	slots.connect(connection->sig_command_received(), this, &Chat::on_connection_command_received);
	slots.connect(connection->sig_unknown_command_received(), this, &Chat::on_connection_unknown_command_received);
	slots.connect(connection->sig_numeric_reply(), this, &Chat::on_connection_numeric_reply);
	slots.connect(connection->sig_name_reply(), this, &Chat::on_connection_name_reply);
	slots.connect(connection->sig_nick(), this, &Chat::on_connection_nick);
	slots.connect(connection->sig_join(), this, &Chat::on_connection_join);
	slots.connect(connection->sig_part(), this, &Chat::on_connection_part);
	slots.connect(connection->sig_mode(), this, &Chat::on_connection_mode);
	slots.connect(connection->sig_topic(), this, &Chat::on_connection_topic);
	slots.connect(connection->sig_invite(), this, &Chat::on_connection_invite);
	slots.connect(connection->sig_kick(), this, &Chat::on_connection_kick);
	slots.connect(connection->sig_privmsg(), this, &Chat::on_connection_privmsg);
	slots.connect(connection->sig_notice(), this, &Chat::on_connection_notice);
	slots.connect(connection->sig_ping(), this, &Chat::on_connection_ping);

	color_text = CL_Color(0,0,0);
	color_channel = CL_Color(0,255,0);
	color_system = CL_Color(0,0,255);
}

Chat::~Chat()
{
	std::list<ChatLine *>::iterator it;
	for (it = lines.begin(); it != lines.end(); ++it) delete *it;
}

/////////////////////////////////////////////////////////////////////////////
// Chat attributes:


/////////////////////////////////////////////////////////////////////////////
// Chat operations:

void Chat::add_line(const std::string &text, const CL_Color &color)
{
	add_line("", text, color);
}

void Chat::add_line(const std::string &nick, const std::string &text, const CL_Color &color)
{
	ChatLine *line = new ChatLine;
	line->nick = nick;
	line->text = text;
	line->color = color;
	lines.push_back(line);
}

/////////////////////////////////////////////////////////////////////////////
// Chat implementation:

void Chat::on_paint()
{
	CL_Display::clear(CL_Color(255,255,255));

	int border = 5;
	int y = get_height() - border;

	int indent = 150;

	std::list<ChatLine *>::reverse_iterator it;
	for (it = lines.rbegin(); it != lines.rend(); ++it)
	{
		ChatLine *line = *it;

		font.set_color(
			line->color.get_red()/255.0f,
			line->color.get_green()/255.0f,
			line->color.get_blue()/255.0f,
			line->color.get_alpha()/255.0f);

		CL_Size nick_size = font.get_size("<"+line->nick+"> ");
		if (nick_size.width < indent) nick_size.width = indent;

		CL_Size bounding_size = font.get_size(line->text, CL_Size(get_width()-nick_size.width-2*border, 0));
		y -= bounding_size.height;

		if (!line->nick.empty()) font.draw(border, y, "<"+line->nick+">");
		font.draw(CL_Rect(CL_Point(border+nick_size.width, y), bounding_size), line->text);
	}
}

void Chat::on_connection_command_received(const std::string &prefix, const std::string &command, const std::vector<std::string> &params)
{
}

void Chat::on_connection_unknown_command_received(const std::string &prefix, const std::string &command, const std::vector<std::string> &params)
{
	add_line(prefix, CL_String::format("Unknown command %1 received.", command), color_system);
}

void Chat::on_connection_numeric_reply(const std::string &prefix, int command, const std::vector<std::string> &params)
{
	std::string param_line;
	for (int i=0; i<params.size(); i++) param_line += params[i] + " ";
	add_line(prefix, CL_String::format("%1 %2", command, param_line), color_system);
}

void Chat::on_connection_name_reply(const std::string &self, const std::string &channel, const std::vector<std::string> &users)
{
}

void Chat::on_connection_nick(const std::string &old_nick, const std::string &new_nick)
{
	add_line(CL_String::format("%1 is now known as %2", old_nick, new_nick), color_channel);
}

void Chat::on_connection_join(const std::string &nick, const std::string &channel)
{
	add_line(CL_String::format("%1 joined %2", nick, channel), color_channel);
}

void Chat::on_connection_part(const std::string &nick, const std::string &channel)
{
	add_line(CL_String::format("%1 left %2", nick, channel), color_channel);
}

void Chat::on_connection_mode(const std::string &prefix, const std::string &receiver, const std::string &mode, const std::vector<std::string> &params)
{
	add_line(CL_String::format("%1 sets mode %3 on %2", prefix, receiver, mode), color_channel);
}

void Chat::on_connection_topic(const std::string &prefix, const std::string &channel, const std::string &topic)
{
	add_line(CL_String::format("Topic on %2 is '%3', set by %1", prefix, channel, topic), color_channel);
}

void Chat::on_connection_invite(const std::string &prefix, const std::string &nick, const std::string &channel)
{
	add_line(CL_String::format("%1 invites %2 to %3", prefix, nick, channel), color_channel);
}

void Chat::on_connection_kick(const std::string &prefix, const std::string &channel, const std::string &user, const std::string &comment)
{
	add_line(CL_String::format("%1 kicked %3 from %2 (%4)", prefix, channel, user, comment), color_channel);
}

void Chat::on_connection_privmsg(const std::string &prefix, const std::string &receiver, const std::string &text)
{
	add_line(prefix, text, color_text);
}

void Chat::on_connection_notice(const std::string &prefix, const std::string &receiver, const std::string &text)
{
	add_line(prefix, text, color_text);
}

void Chat::on_connection_ping(const std::string &daemon1, const std::string &daemon2)
{
	add_line("Ping...uh pong!? - we gotta clone mirc's most important feature ;)", color_system);
}
