#include <ClanLib/display.h>
#include <ClanLib/gl.h>

#include "sprites.h"

Application app;

int Application::main(int argc, char** argv)
{
	#ifdef _DEBUG
	// Create a console window for text-output if not available
	CL_ConsoleWindow console("Debug Console", 80, 1000); // 1000 allows a y-scrollbar to be present
	console.redirect_stdio();
	#endif

	try
	{
		// Init stuff
		CL_SetupCore::init();
		CL_SetupDisplay::init();
		CL_SetupGL::init();
		
		// Create a window
		CL_DisplayWindow window("Sprite Example", 640, 480);

		// Setup resources
		CL_ResourceManager resources("resources.scr", false);

		// Create some sprites
		man_walk = CL_Sprite("spr_man_walk", &resources);
		man_still = CL_Sprite("spr_man_still", &resources);

		man.set_image_data(man_still);
		man.set_alignment(origin_center);
		
		man2.set_image_data(man_walk);
		man2.set_alignment(origin_center);
		man2.set_angle(75.0f);

		is_walking = false;		
		cur_x = 320;
		cur_y = 200;
		delta_x = delta_y = 0;
		
		// Receive mouse clicks in function on_mouse_down()
		slot_mouse = CL_Mouse::sig_key_down().connect(this, &Application::on_mouse_down);

		// Run the main loop
		run();

		// Deinit stuff
		CL_SetupGL::deinit();
		CL_SetupDisplay::deinit();
		CL_SetupCore::deinit();
	}
	catch (CL_Error e)
	{
		std::cout << e.message.c_str() << std::endl;

		#ifdef _DEBUG
		console.display_close_message();
		#endif
	}

	return 0;
}

void Application::on_mouse_down(const CL_InputEvent &key)
{
	// Set new destination
	if(key.id == CL_MOUSE_LEFT)
		set_destination(key.mouse_pos.x, key.mouse_pos.y);

	// Stop walking
	if(key.id == CL_MOUSE_RIGHT)
	{
		is_walking = false;
		man.set_image_data(man_still);
	}
}

// Sets a new destination for our walking sprite
void Application::set_destination(int x, int y)
{
	CL_Vector mouse = CL_Vector(cur_x - x, cur_y - y);
	CL_Vector up = CL_Vector(0, 1);

	// Calculate angle from current sprite position to mouse position
	angle = up.angle(mouse);
	angle = angle * 180.0 / PI;
	if(x < cur_x)
		angle = 360 -angle;

	// Calculate delta position (to be used per frame-update)
	delta_x = -(float)mouse[0] / (float)(mouse.norm() / 75.0f);
	delta_y = -(float)mouse[1] / (float)(mouse.norm() / 75.0f);

	dest_x = x;
	dest_y = y;

	if(is_walking == false)
	{
		man.set_image_data(man_walk);	// Change to walking animation
		is_walking = true;		
	}

	man.set_angle(angle);
}

void Application::run()
{
	while(!CL_Keyboard::get_keycode(CL_KEY_ESCAPE))
	{
		update();
		draw();

		CL_Display::flip();

		CL_System::keep_alive();
	};
}

void Application::update()
{
	// Update sprite animation
	man2.update();
	float time = man.update();

	// Move man if walking
	if(is_walking)
	{
		cur_x += time * delta_x;
		cur_y += time * delta_y;
	}
}

void Application::draw()
{
	// Clear screen
	CL_Display::clear(CL_Color(200,200,200));
		
	// Draw sprites
	man.draw((int)cur_x, (int)cur_y);

	man2.draw(20, 20);
}
