/*
	$Id: point.h,v 1.18 2002/11/13 21:29:58 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	See http://www.clanlib.org
	------------------------------------------------------------------------
*/

//! clanCore="Math"
//! header=core.h

#ifndef header_point
#define header_point

#include "math.h"

//: 2D (x,y) point structure.
class CL_Point
{
// Construction:
public:
	//: Constructs a point.
	//param x: Initial x value.
	//param y: Initial y value.
	//param p: Point to use for initial values.
	CL_Point()
	{ return; }

	CL_Point(int x, int y)
	: x(x), y(y) { }

	CL_Point(const CL_Point &p)
	{ x = p.x; y = p.y; }

// Operations:
public:
	//: Return a rotated version of this point.
	//param hotspot: The point around which to rotate.
	//param angle: The amount of degrees to rotate by, clockwise.
	CL_Point rotate(
		const CL_Point &hotspot,
		float angle) const
	{
		//Move the hotspot to 0,0
		CL_Point r(x - hotspot.x, y - hotspot.y);
		
		//Do some Grumbel voodoo.
		const float c = (float) sqrt((float)r.x*(float)r.x + (float)r.y*(float)r.y);
		
		const float nw = (float)(atan2((float)r.y, (float)r.x) + ((angle + 180) * M_PI / 180));
		
		r.x = (int)((sin(1.5 * M_PI - nw) * c) + 0.5) + hotspot.x;
		r.y = -(int)((sin(nw) * c) + 0.5) + hotspot.y;
		
		return r;
	}
	
	//: Translate point.
	CL_Point &operator+=(const CL_Point &p)
	{ x += p.x; y += p.y; return *this; }

	//: Translate point negatively.
	CL_Point &operator-=(const CL_Point &p)
	{ x -= p.x; y -= p.y; return *this; }
	
	//: Point + Point operator.
	CL_Point operator+(const CL_Point &p) const
	{ return CL_Point(x + p.x, y + p.y); }

	//: Point - Point operator.
	CL_Point operator-(const CL_Point &p) const
	{ return CL_Point(x - p.x, y - p.y); }

// Attributes:
public:
	//: X coordinate.
	int x;

	//: Y coordinate.
	int y;
};

//: 2D (x,y) floating point point structure.
class CL_Pointf
{
// Construction:
public:
	//: Constructs a point.
	//param x: Initial x value.
	//param y: Initial y value.
	//param p: Point to use for initial values.
	CL_Pointf()
	{ return; }
	
	CL_Pointf(float x, float y)
	: x(x), y(y) { }
	
	CL_Pointf(const CL_Pointf &p)
	{ x = p.x; y = p.y; }
	
// Operations:
public:
	//: Return a rotated version of this point.
	//param hotspot: The point around which to rotate.
	//param angle: The amount of degrees to rotate by, clockwise.
	CL_Pointf rotate(
		const CL_Pointf &hotspot,
		float angle) const
	{
		//Move the hotspot to 0,0
		CL_Pointf r(x - hotspot.x, y - hotspot.y);
		
		//Do some Grumbel voodoo.
		const float c = (float)(sqrt(r.x*r.x + r.y*r.y));
		
		const float nw = (float)((float)atan2(r.y, r.x) + ((angle + 180) * M_PI / 180));
		
		r.x = ((float)(sin(1.5 * M_PI - nw)) * c) + hotspot.x;
		r.y = -((float)(sin(nw)) * c) + hotspot.y;
		
		return r;
	}
	
	//: Translate point.
	CL_Pointf &operator+=(const CL_Pointf &p)
	{ x += p.x; y += p.y; return *this; }
	
	//: Translate point negatively.
	CL_Pointf &operator-=(const CL_Pointf &p)
	{ x -= p.x; y -= p.y; return *this; }
	
	//: Point + Point operator.
	CL_Pointf operator+(const CL_Pointf &p) const
	{ return CL_Pointf(x + p.x, y + p.y); }

	//: Point - Point operator.
	CL_Pointf operator-(const CL_Pointf &p) const
	{ return CL_Pointf(x - p.x, y - p.y); }

// Attributes:
public:
	//: X coordinate.
	float x;
	
	//: Y coordinate.
	float y;
};

#endif
