 /*
	$Id: clanstring.h,v 1.49 2003/02/14 11:33:58 sphair Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	See http://www.clanlib.org
	------------------------------------------------------------------------
*/

//! clanCore="System"
//! header=core.h

#ifndef header_clanstring
#define header_clanstring

#include <string>
#include <vector>

class CL_ResourceManager;

class CL_String
//: ClanLib string class.
//- <p>This is a general purpose string class with formatting capabilities.</p>
{
//! Operations:
public:
	//: Loads a string from a resource.
	//param res_id: Resource to load string from.
	//param resources: Resource manager used to load string.
	//param def_value: Default value if resource is not found.
	static std::string load(const std::string &res_id, CL_ResourceManager *resources);
	static std::string load(const std::string &res_id, CL_ResourceManager *resources, const std::string &def_value);
	
	//: Returns the first n characters of the string
	static std::string left(std::string string, int n);
	
	//: Returns the last n characters in the string
	static std::string right(std::string string, int n);

	//: String formatting.
	//- <p>The format specified is of the form "Some text %1, some more text %2".
	//- The %n markers can then be search/replaced using the arg() functions.</p>
	//- <pre>
	//- Example 1:
	//-   CL_String str;
	//-   str.format("%1 world! Meaning of life is: %2");
	//-   str.arg("Hello");
	//-   str.arg(42);
	//- Example 2:
	//-   CL_String str;
	//-   str.format("%1 is better than %2", "ClanLib 0.7", "ClanLib 0.6");
	//- </pre>
	//static std::string format(const std::string &format);

	template<class Param1>
	static std::string format(const std::string &str_format, const Param1 &p1)
	{
		std::string ret = str_format;
		arg(ret, p1, 1);
		
		return ret;
	}

	template<class Param1, class Param2>
	static std::string format(const std::string &str_format, const Param1 &p1, const Param2 &p2)
	{
		std::string ret = str_format;
		arg(ret, p1, 1);
		arg(ret, p2, 2);

		return ret;
	}

	template<class Param1, class Param2, class Param3>
	static std::string format(const std::string &str_format, const Param1 &p1, const Param2 &p2, const Param3 &p3)
	{
		std::string ret = str_format;
		arg(ret, p1, 1);
		arg(ret, p2, 2);
		arg(ret, p3, 3);
		return ret;
	}

	template<class Param1, class Param2, class Param3, class Param4>
	static std::string format(const std::string &str_format, const Param1 &p1, const Param2 &p2, const Param3 &p3, const Param4 &p4)
	{
		std::string ret = str_format;
		arg(ret, p1, 1);
		arg(ret, p2, 2);
		arg(ret, p3, 3);
		arg(ret, p4, 4);
		return ret;
	}
	
	template<class Param1, class Param2, class Param3, class Param4, class Param5>
	static std::string format(const std::string &str_format, const Param1 &p1, const Param2 &p2, const Param3 &p3, const Param4 &p4, const Param5 &p5)
	{
		std::string ret = str_format;
		arg(ret, p1, 1);
		arg(ret, p2, 2);
		arg(ret, p3, 3);
		arg(ret, p4, 4);
		arg(ret, p5, 5);
		return ret;
	}

	template<class Param1, class Param2, class Param3, class Param4, class Param5, class Param6>
	static std::string format(const std::string &str_format, const Param1 &p1, const Param2 &p2, const Param3 &p3, const Param4 &p4, const Param5 &p5, const Param6 &p6)
	{
		std::string ret = str_format;
		arg(ret, p1, 1);
		arg(ret, p2, 2);
		arg(ret, p3, 3);
		arg(ret, p4, 4);
		arg(ret, p5, 5);
		arg(ret, p6, 6);
		return ret;
	}

	template<class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7>
	static std::string format(const std::string &str_format, const Param1 &p1, const Param2 &p2, const Param3 &p3, const Param4 &p4, const Param5 &p5, const Param6 &p6, const Param7 &p7)
	{
		std::string ret = str_format;
		arg(ret, p1, 1);
		arg(ret, p2, 2);
		arg(ret, p3, 3);
		arg(ret, p4, 4);
		arg(ret, p5, 5);
		arg(ret, p6, 6);
		arg(ret, p7, 7);
		return ret;
	}

	//: Converts from an integer to string.
	static std::string from_int(int value);

	//: Converts from an float to string.
	static std::string from_float(float value);

	//: Converts from an double to string.
	static std::string from_double(double value);

	//: Converts string to an integer.
	static int to_int(const std::string&);

	//: Converts string to a floating point number.
	static float to_float(const std::string&);
	
	//: Converts string to a double precision floating point number.
	static double to_double(const std::string&);
	
	//: Returns the string with all characters converted to lowercase.
	static std::string to_lower(const std::string&);
	
	//: Returns the string with all characters converted to uppercase.
	static std::string to_upper(const std::string&);

	//: Breaks up the string into tokens.
	static std::vector<std::string> tokenize(const std::string &str, const std::string &delimiter);

	//: Does a case insensitive compare of string to other string.
	static bool compare_nocase(const std::string&, const std::string&);

//! Implementation:
private:
	// Private Constructor to prevent people from creating instances of CL_String
	CL_String() {;}

	// Replaces argument %(num) in format with arg
	static void arg(std::string &format, const std::string &arg, int num);

	static void arg(std::string &format, const char *arg, int num);

	static void arg(std::string &format, int number, int num);

	static void arg(std::string &format, float number, int num);

	static void arg(std::string &format, double number, int num);
};

#endif
