/*
www.sourceforge.net/projects/tinyxml
Original code (2.0 and earlier )copyright (c) 2000-2002 Lee Thomason (www.grinninglizard.com)

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any
damages arising from the use of this software.

Permission is granted to anyone to use this software for any
purpose, including commercial applications, and to alter it and
redistribute it freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must
not claim that you wrote the original software. If you use this
software in a product, an acknowledgment in the product documentation
would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and
must not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source
distribution.


This is a modified version for ClanLib. Its essentially like the original, except
the class names have been altered to match clanlib's syntax.
*/

//! clanCore="XML"
//! header=core.h

#ifndef header_tinyxml
#define header_tinyxml

// Enable STL stuff
#define CL_XML_USE_STL

#ifdef _MSC_VER
#pragma warning( disable : 4530 )
#pragma warning( disable : 4786 )
#endif

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

// Help out windows:
#if defined( _DEBUG ) && !defined( DEBUG )
#define DEBUG
#endif

#if defined( DEBUG ) && defined( _MSC_VER )
#include <windows.h>
#define CL_XML_LOG OutputDebugString
#else
#define CL_XML_LOG printf
#endif


#ifdef CL_XML_USE_STL
	#include <string>
	#define CL_XML_STRING	std::string
	#define CL_XML_ISTREAM	std::istream
	#define CL_XML_OSTREAM	std::ostream
#else
	#include "tinystr.h"
	#define CL_XML_STRING	CL_XmlString
	#define CL_XML_OSTREAM	CL_XmlOutStream
#endif

class CL_XmlDocument;
class CL_XmlElement;
class CL_XmlComment;
class CL_XmlUnknown;
class CL_XmlAttribute;
class CL_XmlText;
class CL_XmlDeclaration;


//: CL_XmlBase is a base class for every class in TinyXml.
//- <p>It does little except to establish that TinyXml classes
//-	can be printed and provide some utility functions.</p>
//-
//- <p>	In XML, the document and elements can contain
//-	other elements and other types of nodes.</p>
//-
//-<pre>
//-	A Document can contain:	Element	(container or leaf)
//-							Comment (leaf)
//-							Unknown (leaf)
//-							Declaration( leaf )
//-
//-	An Element can contain:	Element (container or leaf)
//-							Text	(leaf)
//-							Attributes (not on tree)
//-							Comment (leaf)
//-							Unknown (leaf)
//-
//-	A Decleration contains: Attributes (not on tree)
//-</pre>
class CL_XmlBase
{
	friend class CL_XmlNode;
	friend class CL_XmlElement;
	friend class CL_XmlDocument;

//! Construction:
public:
	CL_XmlBase()								{}
	virtual ~CL_XmlBase()					{}

//! Attributes:
public:
	//: Changes the white space condense mode.
	//- <p> The world does not agree on whether white space should be kept or
	//-	not. In order to make everyone happy, these global, static functions
	//-	are provided to set whether or not TinyXml will condense all white space
	//-	into a single space or not. The default is to condense. Note changing these
	//-	values is not thread safe.</p>
	static void SetCondenseWhiteSpace( bool condense )		{ condenseWhiteSpace = condense; }

	//: Return the current white space setting.
	static bool IsWhiteSpaceCondensed()						{ return condenseWhiteSpace; }

//! Operations:
public:
	//: Print node to file.
	//- <p>	All TinyXml classes can print themselves to a filestream.
	//-	This is a formatted print, and will insert tabs and newlines.</p>
	//-	
	//- <p>	For an unformatted stream, use the << operator.</p>
	virtual void Print( FILE* cfile, int depth ) const = 0;

//! Implementation:
protected:
	// See STL_STRING_BUG
	// Utility class to overcome a bug.
	class StringToBuffer
	{
	  public:
		StringToBuffer( const CL_XML_STRING& str );
		~StringToBuffer();
		char* buffer;
	};

	static const char*	SkipWhiteSpace( const char* );
	inline static bool	IsWhiteSpace( int c )		{ return ( isspace( c ) || c == '\n' || c == '\r' ); }

	virtual void StreamOut (CL_XML_OSTREAM *) const = 0;

	#ifdef CL_XML_USE_STL
	    static bool	StreamWhiteSpace( CL_XML_ISTREAM * in, CL_XML_STRING * tag );
	    static bool StreamTo( CL_XML_ISTREAM * in, int character, CL_XML_STRING * tag );
	#endif

	/*	Reads an XML name into the string provided. Returns
		a pointer just past the last character of the name,
		or 0 if the function has an error.
	*/
	static const char* ReadName( const char* p, CL_XML_STRING* name );

	/*	Reads text. Returns a pointer past the given end tag.
		Wickedly complex options, but it keeps the (sensitive) code in one place.
	*/
	static const char* ReadText(	const char* in,				// where to start
									CL_XML_STRING* text,			// the string read
									bool ignoreWhiteSpace,		// whether to keep the white space
									const char* endTag,			// what ends this text
									bool ignoreCase );			// whether to ignore case in the end tag
	virtual const char* Parse( const char* p ) = 0;

	// If an entity has been found, transform it into a character.
	static const char* GetEntity( const char* in, char* value );

	// Get a character, while interpreting entities.
	inline static const char* GetChar( const char* p, char* value )
	{
		assert( p );
		if ( *p == '&' )
		{
			return GetEntity( p, value );
		}
		else
		{
			*value = *p;
			return p+1;
		}
	}

	// Puts a string to a stream, expanding entities as it goes.
	// Note this should not contian the '<', '>', etc, or they will be transformed into entities!
	static void PutString( const CL_XML_STRING& str, CL_XML_OSTREAM* out );

	static void PutString( const CL_XML_STRING& str, CL_XML_STRING* out );

	// Return true if the next characters in the stream are any of the endTag sequences.
	bool static StringEqual(	const char* p,
								const char* endTag,
								bool ignoreCase );


	enum
	{
		CL_XML_NO_ERROR = 0,
		CL_XML_ERROR,
		CL_XML_ERROR_OPENING_FILE,
		CL_XML_ERROR_OUT_OF_MEMORY,
		CL_XML_ERROR_PARSING_ELEMENT,
		CL_XML_ERROR_FAILED_TO_READ_ELEMENT_NAME,
		CL_XML_ERROR_READING_ELEMENT_VALUE,
		CL_XML_ERROR_READING_ATTRIBUTES,
		CL_XML_ERROR_PARSING_EMPTY,
		CL_XML_ERROR_READING_END_TAG,
		CL_XML_ERROR_PARSING_UNKNOWN,
		CL_XML_ERROR_PARSING_COMMENT,
		CL_XML_ERROR_PARSING_DECLARATION,
		CL_XML_ERROR_DOCUMENT_EMPTY,

		CL_XML_ERROR_STRING_COUNT
	};
	static const char* errorString[ CL_XML_ERROR_STRING_COUNT ];

private:
	struct Entity
	{
		const char*     str;
		unsigned int	strLength;
		char		    chr;
	};
	enum
	{
		NUM_ENTITY = 5,
		MAX_ENTITY_LENGTH = 6

	};
	static Entity entity[ NUM_ENTITY ];
	static bool condenseWhiteSpace;
};


//: The parent class for everything in the Document Object Model. (Except for attributes, which are contained in elements.)
//- <p>Nodes have siblings, a parent, and children. A node can be
//-	in a document, or stand on its own. The type of a CL_XmlNode
//-	can be queried, and it can be cast to its more defined type.</p>
class CL_XmlNode : public CL_XmlBase
{
	friend class CL_XmlDocument;
	friend class CL_XmlElement;

//! Constructions:
public:
	virtual ~CL_XmlNode();

//! Attributes:
	//: The types of XML nodes supported by TinyXml.
	//- <p>All the unsupported types are picked up by UNKNOWN.</p>
	enum NodeType
	{
		DOCUMENT,
		ELEMENT,
		COMMENT,
		UNKNOWN,
		TEXT,
		DECLARATION,
		TYPECOUNT
	};

	//: One step up the DOM.
	CL_XmlNode* Parent() const					{ return parent; }

	//: The first child of this node.
	//- <p>Will be null if there are no children.</p>
	//!param value: The first child of this node with the matching 'value'. Will be null if none found.
	CL_XmlNode* FirstChild()	const	{ return firstChild; }
	CL_XmlNode* FirstChild( const char * value ) const;
	#ifdef CL_XML_USE_STL
	CL_XmlNode* FirstChild( const std::string& value ) const	{	return FirstChild (value.c_str ());	}
	#endif

	//- The last child of this node.
	//- <p>Will be null if there are no children.</p>
	//!param value: The last child of this node matching 'value'. Will be null if there are no children.
	CL_XmlNode* LastChild() const	{ return lastChild; }
	CL_XmlNode* LastChild( const char * value ) const;

	#ifdef CL_XML_USE_STL
	CL_XmlNode* LastChild( const std::string& value ) const	{	return LastChild (value.c_str ());	}
	#endif

	//: Returns the 'value' of the node. The meaning of 'value' changes for the specific type of CL_XmlNode.
	//- <pre>	Document:	filename of the xml file
	//-	Element:	name of the element
	//-	Comment:	the comment text
	//-	Unknown:	the tag contents
	//-	Text:		the text string
	//- </pre>
	//-
	//- <p>The subclasses will wrap this function.</p>
	const char * Value () const { return value.c_str (); }

	//: Query the type (as an enumerated value, above) of this node.
	virtual int Type() const	{ return type; }

	//: Return a pointer to the Document this node lives in.
	//- <p>Returns null if not in a document.</p>
	CL_XmlDocument* GetDocument() const;

	//: Returns true if this node has no children.
	bool NoChildren() const						{ return !firstChild; }

	//: Set user data.
	void  SetUserData( void* user )			{ userData = user; }
	
	//: Get user data.
	void* GetUserData()						{ return userData; }

//! Operations:
public:
	#ifdef CL_XML_USE_STL	

	    //: Input stream operator.
	    //- <p>An input stream operator, for every class. Tolerant of newlines and
	    //- formatting, but doesn't expect them.</p>
	    friend std::istream& operator >> (std::istream& in, CL_XmlNode& base);

	    //: Output stream operator.
	    //- <p>An output stream operator, for every class. Note that this outputs
	    //-	    without any newlines or formatting, as opposed to Print(), which
	    //-	    includes tabs and new lines.</p>
	    //-
	    //- <p>The operator<< and operator>> are not completely symmetric. Writing
	    //-	    a node to a stream is very well defined. You'll get a nice stream
	    //-	    of output, without any extra whitespace or newlines.</p>
	    //-
	    //- <p>But reading is not as well defined. (As it always is.) If you create
	    //-	    a CL_XmlElement (for example) and read that from an input stream,
	    //-	    the text needs to define an element or junk will result. This is
	    //-	    true of all input streams, but it's worth keeping in mind.</p>
	    //-
	    //- <p>A CL_XmlDocument will read nodes until it reads a root element.</p>
	    friend std::ostream & operator<< (std::ostream& out, const CL_XmlNode& base);

	#else
	    // Used internally, not part of the public API.
	    friend CL_XML_OSTREAM& operator<< (CL_XML_OSTREAM& out, const CL_XmlNode& base);
	#endif

	//: Changes the value of the node.
	//- <pre>Defined as:
	//-	Document:	filename of the xml file
	//-	Element:	name of the element
	//-	Comment:	the comment text
	//-	Unknown:	the tag contents
	//-	Text:		the text string
	//- </pre>
	void SetValue (const char * _value) { value = _value;}

    #ifdef CL_XML_USE_STL
	void SetValue( const std::string& value )    
	{	  
		StringToBuffer buf( value );
		SetValue( buf.buffer ? buf.buffer : "" );    	
	}	
	#endif

	//: Delete all the children of this node. Does not affect 'this'.
	void Clear();

	//: An alternate way to walk the children of a node.
	//- <p>One way to iterate over nodes is:</p>
	//- <pre>	for( child = parent->FirstChild(); child; child = child->NextSibling() )</pre>
	//-
	//- <p>IterateChildren does the same thing with the syntax:</p>
	//- <pre>	child = 0;
	//- while( child = parent->IterateChildren( child ) )</pre>
	//-
	//- <p>IterateChildren takes the previous child as input and finds
	//-	the next one. If the previous child is null, it returns the
	//-	first. IterateChildren will return null when done.</p>
	CL_XmlNode* IterateChildren( CL_XmlNode* previous ) const;

	//!param value: This flavor of IterateChildren searches for children with a particular 'value'.
	CL_XmlNode* IterateChildren( const char * value, CL_XmlNode* previous ) const;

	#ifdef CL_XML_USE_STL
	CL_XmlNode* IterateChildren( const std::string& value, CL_XmlNode* previous ) const	{	return IterateChildren (value.c_str (), previous);	}
	#endif

	//: Add a new node related to this. Adds a child past the LastChild.
	//- <p>Returns a pointer to the new object or NULL if an error occured.</p>
	CL_XmlNode* InsertEndChild( const CL_XmlNode& addThis );

	//: Add a new node related to this. Adds a child before the specified child.
	//- <p>Returns a pointer to the new object or NULL if an error occured.</p>
	CL_XmlNode* InsertBeforeChild( CL_XmlNode* beforeThis, const CL_XmlNode& addThis );

	//: Add a new node related to this. Adds a child after the specified child.
	//- <p>Returns a pointer to the new object or NULL if an error occured.</p>
	CL_XmlNode* InsertAfterChild(  CL_XmlNode* afterThis, const CL_XmlNode& addThis );

	//: Replace a child of this node.
	//- <p>Returns a pointer to the new object or NULL if an error occured.</p>
	CL_XmlNode* ReplaceChild( CL_XmlNode* replaceThis, const CL_XmlNode& withThis );

	//: Delete a child of this node.
	bool RemoveChild( CL_XmlNode* removeThis );

	//: Navigate to a sibling node.
	//!param value: Navigate to a sibling node with the given 'value'.
	CL_XmlNode* PreviousSibling() const			{ return prev; }

	CL_XmlNode* PreviousSibling( const char *value ) const;

	#ifdef CL_XML_USE_STL
	CL_XmlNode* PreviousSibling( const std::string& value ) const	{	return PreviousSibling (value.c_str ());	}
	CL_XmlNode* NextSibling( const std::string& value) const	{	return NextSibling (value.c_str ());	}
	#endif

	//: Navigate to a sibling node.
	//!param value: Navigate to a sibling node with the given 'value'.
	CL_XmlNode* NextSibling() const				{ return next; }

	CL_XmlNode* NextSibling( const char *value ) const;

	//: Convenience function to get through elements.
	//- <p>Calls NextSibling and ToElement. Will skip all non-Element
	//- nodes. Returns 0 if there is not another element.</p>
	CL_XmlElement* NextSiblingElement() const;

	CL_XmlElement* NextSiblingElement( const char * ) const;

	#ifdef CL_XML_USE_STL
	CL_XmlElement* NextSiblingElement( const std::string& value) const	{	return NextSiblingElement (value.c_str ());	}
	#endif

	//: Convenience function to get through elements.
	CL_XmlElement* FirstChildElement()	const;

	CL_XmlElement* FirstChildElement( const char * value ) const;

	#ifdef CL_XML_USE_STL
	CL_XmlElement* FirstChildElement( const std::string& value ) const	{	return FirstChildElement (value.c_str ());	}
	#endif

	//: Cast to a CL_XmlDocument type.
	//- <p>Will return null not of the requested type.</p>
	CL_XmlDocument* ToDocument()	const		{ return ( this && type == DOCUMENT ) ? (CL_XmlDocument*) this : 0; }
	
	//: Cast to a CL_XmlElement type.
	//- <p>Will return null not of the requested type.</p>
	CL_XmlElement*  ToElement() const		{ return ( this && type == ELEMENT  ) ? (CL_XmlElement*)  this : 0; }
	
	//: Cast to a CL_XmlComment type.
	//- <p>Will return null not of the requested type.</p>
	CL_XmlComment*  ToComment() const		{ return ( this && type == COMMENT  ) ? (CL_XmlComment*)  this : 0; }
	
	//: Cast to a CL_XmlUnknown type.
	//- <p>Will return null not of the requested type.</p>
	CL_XmlUnknown*  ToUnknown() const		{ return ( this && type == UNKNOWN  ) ? (CL_XmlUnknown*)  this : 0; }

	//: Cast to a CL_XmlText type.
	//- <p>Will return null not of the requested type.</p>
	CL_XmlText*	   ToText()    const		{ return ( this && type == TEXT     ) ? (CL_XmlText*)     this : 0; }
	
	//: Cast to a CL_XmlDeclaration type.
	//- <p>Will return null not of the requested type.</p>
	CL_XmlDeclaration* ToDeclaration() const	{ return ( this && type == DECLARATION ) ? (CL_XmlDeclaration*) this : 0; }

	//: Clone node.
	virtual CL_XmlNode* Clone() const = 0;

//! Implementation:
protected:
	CL_XmlNode( NodeType type );

	#ifdef CL_XML_USE_STL
	    // The real work of the input operator.
	    virtual void StreamIn( CL_XML_ISTREAM* in, CL_XML_STRING* tag ) = 0;
	#endif

	// The node is passed in by ownership. This object will delete it.
	CL_XmlNode* LinkEndChild( CL_XmlNode* addThis );

	// Figure out what is at *p, and parse it. Returns null if it is not an xml node.
	CL_XmlNode* Identify( const char* start );
	void CopyToClone( CL_XmlNode* target ) const	{ target->SetValue (value.c_str() );
												  target->userData = userData; }

	// Internal Value function returning a CL_XML_STRING
	CL_XML_STRING SValue() const	{ return value ; }

	CL_XmlNode*		parent;
	NodeType		type;

	CL_XmlNode*		firstChild;
	CL_XmlNode*		lastChild;

	CL_XML_STRING	value;

	CL_XmlNode*		prev;
	CL_XmlNode*		next;
	void*			userData;
};


//: XML attribute name-value pair.
//- <p>An attribute is a name-value pair. Elements have an arbitrary
//-	number of attributes, each with a unique name.</p>
//-
//- <p>Note: The attributes are not CL_XmlNodes, since they are not
//- part of the tinyXML document object model. There are other
//- suggested ways to look at this problem.</p>
//-
//- <p>Note: Attributes have a parent.</p>
class CL_XmlAttribute : public CL_XmlBase
{
	friend class CL_XmlAttributeSet;

//! Construction:
public:
	//: Constructs an attribute.
	//!param name: Attribute name.
	//!param value: Attibute value.
	CL_XmlAttribute() : prev( 0 ), next( 0 )	{}

	#ifdef CL_XML_USE_STL
	CL_XmlAttribute( const std::string& _name, const std::string& _value )
	{
		name = _name;
		value = _value;
	}
	#endif

	CL_XmlAttribute( const char * _name, const char * _value ): name( _name ), value( _value ), prev( 0 ), next( 0 ) {}
	
//! Attributes:
public:
	//: Return the name of this attribute.
	const char*		Name()  const		{ return name.c_str (); }
	
	//: Return the value of this attribute.
	const char*		Value() const		{ return value.c_str (); }
	
	//: Return the value of this attribute, converted to an integer.
	const int       IntValue() const;

	//: Return the value of this attribute, converted to a double.
	const double	DoubleValue() const;

	//: Set the name of this attribute.
	void SetName( const char* _name )	{ name = _name; }

	#ifdef CL_XML_USE_STL
	void SetName( const std::string& _name )	
	{	
		StringToBuffer buf( _name );
		SetName ( buf.buffer ? buf.buffer : "error" );	
	}
	#endif
	
	//: Set the value.
	void SetValue( const char* _value )	{ value = _value; }

	//: Set the value from an integer.
	void SetIntValue( int value );
	
	//: Set the value from a double.
	void SetDoubleValue( double value );

	#ifdef CL_XML_USE_STL
	void SetValue( const std::string& _value )	
	{	
		StringToBuffer buf( _value );
		SetValue( buf.buffer ? buf.buffer : "error" );	
	}
	#endif

	//: Get the next sibling attribute in the DOM. Returns null at end.
	CL_XmlAttribute* Next() const;
	
	//: Get the previous sibling attribute in the DOM. Returns null at beginning.
	CL_XmlAttribute* Previous() const;

//! Operations:
public:
	//: Compares the attribute names and returns true if equal.
	bool operator==( const CL_XmlAttribute& rhs ) const { return rhs.name == name; }

	//: Returns true if attribute name is lesser than the passed attribute.
	bool operator<( const CL_XmlAttribute& rhs )	 const { return name < rhs.name; }

	//: Returns true if attribute name is greater than the passed attribute.
	bool operator>( const CL_XmlAttribute& rhs )  const { return name > rhs.name; }

//! Implementation:
	/*	[internal use]
		Attribtue parsing starts: first letter of the name
						 returns: the next char after the value end quote
	*/
	virtual const char* Parse( const char* p );

	// [internal use]
	virtual void Print( FILE* cfile, int depth ) const;

	virtual void StreamOut( CL_XML_OSTREAM * out ) const;
	// [internal use]
	// Set the document pointer so the attribute can report errors.
	void SetDocument( CL_XmlDocument* doc )	{ document = doc; }

private:
	CL_XmlDocument*	document;	// A pointer back to a document, for error reporting.
	CL_XML_STRING name;
	CL_XML_STRING value;
	CL_XmlAttribute*	prev;
	CL_XmlAttribute*	next;
};


//: A class used to manage a group of attributes.
//- <p>It is only used internally, both by the ELEMENT and the DECLARATION.</p>
//-
//- <p>	The set can be changed transparent to the Element and Declaration
//-	classes that use it, but NOT transparent to the Attribute
//-	which has to implement a next() and previous() method. Which makes
//-	it a bit problematic and prevents the use of STL.</p>
//-
//- <p>	This version is implemented with circular lists because:</p>
//- <ul><li>I like circular lists</li>
//- <li>it demonstrates some independence from the (typical) doubly linked list.</li></ul>
class CL_XmlAttributeSet
{
//! Construction:
public:
	//: Constructs an attribute set.
	CL_XmlAttributeSet();

	~CL_XmlAttributeSet();

//! Operations:
public:
	void Add( CL_XmlAttribute* attribute );
	void Remove( CL_XmlAttribute* attribute );

	CL_XmlAttribute* First() const	{ return ( sentinel.next == &sentinel ) ? 0 : sentinel.next; }
	CL_XmlAttribute* Last()  const	{ return ( sentinel.prev == &sentinel ) ? 0 : sentinel.prev; }
	CL_XmlAttribute*	Find( const char * name ) const;

//! Implementation:
private:
	CL_XmlAttribute sentinel;
};


//: XML DOM Element.
//- <p>The element is a container class. It has a value, the element name,
//- and can contain other elements, text, comments, and unknowns.</p>
//- <p>Elements also contain an arbitrary number of attributes.</p>
class CL_XmlElement : public CL_XmlNode
{
//! Construction:
public:
	//: Construct an element.
	CL_XmlElement (const char * in_value);

	#ifdef CL_XML_USE_STL
	CL_XmlElement( const std::string& _value ) : 	CL_XmlNode( CL_XmlNode::ELEMENT )
	{
		firstChild = lastChild = 0;
		value = _value;
	}
	#endif

	virtual ~CL_XmlElement();

//! Attributes:
public:
	//: Given an attribute name, attribute returns the value for the attribute of that name, or null if none exists.
	const char* Attribute( const char* name ) const;

	//: Given an attribute name, attribute returns the value for the attribute of that name, or null if none exists.
	//- <p>If the attribute exists and can be converted to an integer,
	//-	the integer value will be put in the return 'i', if 'i'
	//-	is non-null.</p>
	const char* Attribute( const char* name, int* i ) const;

	#ifdef CL_XML_USE_STL
	const char* Attribute( const std::string& name ) const				{ return Attribute( name.c_str() ); }
	const char* Attribute( const std::string& name, int* i ) const		{ return Attribute( name.c_str(), i ); }
	#endif

	//: Sets an attribute of name to a given value.
	//- <p>The attribute will be created if it does not exist, or changed if it does.</p>
	void SetAttribute( const char* name, const char * value );

	#ifdef CL_XML_USE_STL
	void SetAttribute( const std::string& name, const std::string& value )	
	{	
		StringToBuffer n( name );
		StringToBuffer v( value );
		if ( n.buffer && v.buffer )
			SetAttribute (n.buffer, v.buffer );	
	}	
	void SetAttribute( const std::string& name, int value )	
	{	
		StringToBuffer n( name );
		if ( n.buffer )
			SetAttribute (n.buffer, value);	
	}	
	#endif

	void SetAttribute( const char * name, int value );

//! Operations:
public:
	//: Deletes an attribute with the given name.
	void RemoveAttribute( const char * name );
	
	#ifdef CL_XML_USE_STL
	void RemoveAttribute( const std::string& name )	{	RemoveAttribute (name.c_str ());	}
	#endif

	//: Access the first attribute in this element.
	CL_XmlAttribute* FirstAttribute() const	{ return attributeSet.First(); }
	
	//: Access the last attribute in this element.
	CL_XmlAttribute* LastAttribute() const 	{ return attributeSet.Last(); }

//! Implementation:
	// [internal use] Creates a new Element and returs it.
	virtual CL_XmlNode* Clone() const;
	// [internal use]

	virtual void Print( FILE* cfile, int depth ) const;

protected:

	// Used to be public [internal use]
	#ifdef CL_XML_USE_STL
	    virtual void StreamIn( CL_XML_ISTREAM * in, CL_XML_STRING * tag );
	#endif
	virtual void StreamOut( CL_XML_OSTREAM * out ) const;

	/*	[internal use]
		Attribtue parsing starts: next char past '<'
						 returns: next char past '>'
	*/
	virtual const char* Parse( const char* p );

	/*	[internal use]
		Reads the "value" of the element -- another element, or text.
		This should terminate with the current end tag.
	*/
	const char* ReadValue( const char* in );

private:
	CL_XmlAttributeSet attributeSet;
};


//: An XML DOM comment.
class CL_XmlComment : public CL_XmlNode
{
//! Construction:
public:
	//: Constructs an empty comment.
	CL_XmlComment() : CL_XmlNode( CL_XmlNode::COMMENT ) {}

	virtual ~CL_XmlComment()	{}

//! Implementation:
	// [internal use] Creates a new Element and returs it.
	virtual CL_XmlNode* Clone() const;
	// [internal use]
	virtual void Print( FILE* cfile, int depth ) const;
protected:
	// used to be public
	#ifdef CL_XML_USE_STL
	    virtual void StreamIn( CL_XML_ISTREAM * in, CL_XML_STRING * tag );
	#endif
	virtual void StreamOut( CL_XML_OSTREAM * out ) const;
	/*	[internal use]
		Attribtue parsing starts: at the ! of the !--
						 returns: next char past '>'
	*/
	virtual const char* Parse( const char* p );
};


//: XML DOM text. Contained in an element.
class CL_XmlText : public CL_XmlNode
{
	friend class CL_XmlElement;
//! Construction:
public:
	//: Constructor.
	//!param initValue: Initial value assigned to text node.
	CL_XmlText (const char * initValue) : CL_XmlNode (CL_XmlNode::TEXT)
	{
		SetValue( initValue );
	}

	virtual ~CL_XmlText() {}

	#ifdef CL_XML_USE_STL
	CL_XmlText( const std::string& initValue ) : CL_XmlNode (CL_XmlNode::TEXT)
	{
		SetValue( initValue );
	}
	#endif

//! Implementation:
protected :
	// [internal use] Creates a new Element and returns it.
	virtual CL_XmlNode* Clone() const;
	// [internal use]
	virtual void Print( FILE* cfile, int depth ) const;
	virtual void StreamOut ( CL_XML_OSTREAM * out ) const;
	// [internal use]
	bool Blank() const;	// returns true if all white space and new lines
	/*	[internal use]
			Attribtue parsing starts: First char of the text
							 returns: next char past '>'
		*/
	virtual const char* Parse( const char* p );
	// [internal use]
	#ifdef CL_XML_USE_STL
	    virtual void StreamIn( CL_XML_ISTREAM * in, CL_XML_STRING * tag );
	#endif
};


//: XML DOM Declaration node.
//- <p>In correct XML the declaration is the first entry in the file.</p>
//- <pre>	<?xml version="1.0" standalone="yes"?></pre>
//-
//- <p>tinyXml will happily read or write files without a declaration,
//- however. There are 3 possible attributes to the declaration:
//- version, encoding, and standalone.</p>
//-
//- <p>Note: In this version of the code, the attributes are
//- handled as special cases, not generic attributes, simply
//- because there can only be at most 3 and they are always the same.</p>
class CL_XmlDeclaration : public CL_XmlNode
{
//! Construction:
public:
	//: Constructs a declaration.
	CL_XmlDeclaration()   : CL_XmlNode( CL_XmlNode::DECLARATION ) {}

#ifdef CL_XML_USE_STL
	CL_XmlDeclaration(
						const std::string& _version,
						const std::string& _encoding,
						const std::string& _standalone )
					: CL_XmlNode( CL_XmlNode::DECLARATION )
	{
		version = _version;
		encoding = _encoding;
		standalone = _standalone;
	}
#endif

	CL_XmlDeclaration::CL_XmlDeclaration( const char * _version,
										const char * _encoding,
										const char * _standalone );

	virtual ~CL_XmlDeclaration()	{}

//! Attributes:
public:
	//: Version. Will return empty if none was found.
	const char * Version() const		{ return version.c_str (); }
	
	//: Encoding. Will return empty if none was found.
	const char * Encoding() const		{ return encoding.c_str (); }
	
	//: Is this a standalone document?
	const char * Standalone() const		{ return standalone.c_str (); }

//! Implementation:
	// [internal use] Creates a new Element and returs it.
	virtual CL_XmlNode* Clone() const;
	// [internal use]
	virtual void Print( FILE* cfile, int depth ) const;

protected:
	// used to be public
	#ifdef CL_XML_USE_STL
	    virtual void StreamIn( CL_XML_ISTREAM * in, CL_XML_STRING * tag );
	#endif
	virtual void StreamOut ( CL_XML_OSTREAM * out) const;
	//	[internal use]
	//	Attribtue parsing starts: next char past '<'
	//					 returns: next char past '>'

	virtual const char* Parse( const char* p );

private:
	CL_XML_STRING version;
	CL_XML_STRING encoding;
	CL_XML_STRING standalone;
};


//: XML DOM Unknown node.
//- <p>Any tag that tinyXml doesn't recognize is save as an
//- unknown. It is a tag of text, but should not be modified.
//- It will be written back to the XML, unchanged, when the file
//- is saved.</p>
class CL_XmlUnknown : public CL_XmlNode
{
//! Construction:
public:
	//: Constructor.
	CL_XmlUnknown() : CL_XmlNode( CL_XmlNode::UNKNOWN ) {}
	
	virtual ~CL_XmlUnknown() {}

//! Implementation:
	// [internal use]
	virtual CL_XmlNode* Clone() const;
	// [internal use]
	virtual void Print( FILE* cfile, int depth ) const;
protected:
	// used to be public
	#ifdef CL_XML_USE_STL
	    virtual void StreamIn( CL_XML_ISTREAM * in, CL_XML_STRING * tag );
	#endif
	virtual void StreamOut ( CL_XML_OSTREAM * out ) const;
	/*	[internal use]
		Attribute parsing starts: First char of the text
						 returns: next char past '>'
	*/
	virtual const char* Parse( const char* p );
};


//: XML DOM Document.
//- <p>Always the top level node. A document binds together all the
//- XML pieces. It can be saved, loaded, and printed to the screen.
//- The 'value' of a document node is the xml file name.</p>
class CL_XmlDocument : public CL_XmlNode
{
//! Construction:
public:
	//: Creates a document.
	//!param documentName: Create a document with a name. The name of the document is also the filename of the xml.
	CL_XmlDocument();
	
	CL_XmlDocument( const char * documentName );

	#ifdef CL_XML_USE_STL
	CL_XmlDocument( const std::string& documentName ) :
	    CL_XmlNode( CL_XmlNode::DOCUMENT )
	{
		value = documentName;
		error = false;
	}
	#endif

	virtual ~CL_XmlDocument() {}

//! Attributes:
public:
	//: Get the root element.
	//- <p>Get the root element -- the only top level element -- of the document.
	//- In well formed XML, there should only be one. TinyXml is tolerant of
	//- multiple elements at the document level.</p>
	CL_XmlElement* RootElement() const		{ return FirstChildElement(); }

	//: If, during parsing, a error occurs, Error will be set to true.
	bool Error() const						{ return error; }

	//: Contains a textual (english) description of the error if one occurs.
	const char * ErrorDesc() const	{ return errorDesc.c_str (); }

	//: Contains the error ID if an error occours.
	//- <p>Generally, you probably want the error string ( ErrorDesc() ). But if you
	//- prefer the ErrorId, this function will fetch it.</p>
	const int ErrorId()	const				{ return errorId; }

//! Operations:
public:
	//: Load a file using the current document value.
	//- <p>Returns true if successful. Will delete any existing
	//- document data before loading.</p>
	//!param filename: Load a file using the given filename.
	bool LoadFile();

	bool LoadFile( const char * filename );

	#ifdef CL_XML_USE_STL
	bool LoadFile( const std::string& filename )
	{
		StringToBuffer f( filename );
		return ( f.buffer && LoadFile( f.buffer ));
	}
	#endif
	
	//: Save a file using the current document value.
	//- <p>Returns true if successful.</p>
	//!param filename: Save a file using the given filename.
	bool SaveFile() const;
	
	bool SaveFile( const char * filename ) const;

	#ifdef CL_XML_USE_STL
	bool SaveFile( const std::string& filename ) const
	{
		StringToBuffer f( filename );
		return ( f.buffer && SaveFile( f.buffer ));
	}
	#endif

	//: Parse the given null terminated block of xml data.
	virtual const char* Parse( const char* p );

	//: If you have handled the error, it can be reset with this call.
	void ClearError()						{ error = false; errorId = 0; errorDesc = ""; }

	//: Dump the document to standard out.
	void Print() const						{ Print( stdout, 0 ); }

//! Implementation:
	// [internal use]
	virtual void Print( FILE* cfile, int depth = 0 ) const;
	// [internal use]
	void SetError( int err ) {		assert( err > 0 && err < CL_XML_ERROR_STRING_COUNT );
		error   = true;
		errorId = err;
	errorDesc = errorString[ errorId ]; }

protected :
	virtual void StreamOut ( CL_XML_OSTREAM * out) const;
	// [internal use]
	virtual CL_XmlNode* Clone() const;
	#ifdef CL_XML_USE_STL
	    virtual void StreamIn( CL_XML_ISTREAM * in, CL_XML_STRING * tag );
	#endif

private:
	bool error;
	int  errorId;
	CL_XML_STRING errorDesc;
};

#endif

