/*
	$Id: glyph_buffer.h,v 1.11 2003/02/05 04:38:24 plasmoid Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	See http://www.clanlib.org
	------------------------------------------------------------------------
*/

//! clanDisplay="Fonts"
//! header=display.h

#ifndef header_glyph_buffer
#define header_glyph_buffer

#ifdef WIN32
#pragma warning( disable : 4786)
#endif

#include <string>
#include <map>
#include <vector>

#include "font.h"
#include "../Core/Math/origin.h"
#include "color.h"
#include "sprite.h"

class CL_Point;
class CL_Rect;
class CL_GraphicContext;

//: Stores and draws glyphs and their positions in one or several fonts.
//- <p>A glyph buffer consists of a series of characters and positions,
//- and markers for font changes and various state changes. The internal data structures used
//- (six STL containers, one for the positions of characters, one for font markers, 
//- one for colors, two for scale, and one for angle) 
//- are made accessible and mutble externally. CL_Font
//- and CL_TextStyler both have draw_to_gb() methods, which allows them to
//- add data into a CL_GlyphBuffer rather than drawing to a CL_GraphicContext.</p>
//- <p>CL_GlyphBuffer is mostly useful for text special effects, or for any case
//- where you need access to the low-level character data. It is not intended
//- as a way of speeding up the rendering of static text messages. For that, just
//- have CL_Font or CL_TextStyler render into a CL_Canvas.</p>
//- <p>The effect maps for color, angle, and scale allow you to affect those properties
//- for individual glyphs during drawing and bounding rect calcluation. This is different
//- from the GlyphBuffer angle, color,
//- and scale properties; these apply to the entire CL_GlyphBuffer as a whole, while entries
//- in the effect maps affect individual glyphs seperately.</p>
//- <p>In functions that calculate the size or bounding rectangle of the GlyphBuffer
//- (such as bouding_rect(), get_size(), and the alignment calculations within draw),
//- effect properties do take effect on the final rectangle, but GlyphBuffer rotation
//- and scaling (set with CL_GlyphBuffer::set_angle(), and CL_GlyphBuffer::set_scale(),
//- respectively) do not. This mimics the behavior of CL_Sprite and CL_Surface; the content
//- affects the bounding rect calculation, but properties of the class object besides
//- the content itself only affect drawing.</p>
class CL_GlyphBuffer
{
//! Construction:
public:
	//: Constructs a glyph buffer.
	CL_GlyphBuffer();
	
	virtual ~CL_GlyphBuffer() {;}
	
//! Subclasses:
public:
	//: A character and position describing a glyph in the buffer, stored in the glyphs vector.
	struct Glyph
	{
		//:Constructs a new Glyph
		//p: The CL_Point that will be assigned to pos
		//c: The unsigned char that will be assigned to character
		Glyph() {;}
		Glyph(const CL_Point &p, unsigned char c) : pos(p), character(c) {;}
		
		//: The position where the glyph is drawn.
		CL_Point pos;
		
		//: The character which determines the glyph that's drawn.
		unsigned char character;
	};
	
//! Attributes:
public:
	//: Returns a reference to the glyphs in the buffer and their drawing positions.
	//- <p>See CL_GlyphBuffer::Glyph for details./p>
	std::vector<Glyph>& get_glyphs() {return glyphs;}
	const std::vector<Glyph>& get_glyphs() const {return glyphs;}
	
	//: Returns a reference to the font map.
	//- <p>Elements of the glyphs vector before the first font marker are ignored. Each element
	//- within this map causes a change in the font while drawing, where the key of the
	//- element is the index in the glyphs vector before which the font is changed.</p>
	std::map<int, CL_Font>& get_font_markers() {return font_markers;}
	const std::map<int, CL_Font>& get_font_markers() const {return font_markers;}
	
	//: Returns a reference to the color effect map.
	std::map<int, CL_Color>& get_color_effects() {return color_effects;}
	const std::map<int, CL_Color>& get_color_effects() const {return color_effects;}
	
	//: Returns a reference to the x scale effect map.
	//- <p> Scaling occurs away from the upper-left corner of each glyph. </p>
	std::map<int, float>& get_scale_x_effects() {return scale_x_effects;}
	const std::map<int, float>& get_scale_x_effects() const {return scale_x_effects;}
	
	//: Returns a reference to the y scale effect map.
	//- <p> Scaling occurs away from the upper-left corner of each glyph. </p>
	std::map<int, float>& get_scale_y_effects() {return scale_y_effects;}
	const std::map<int, float>& get_scale_y_effects() const {return scale_y_effects;}
	
	//: Returns a reference to the character angle effect map.
	std::map<int, float>& get_angle_effects() {return angle_effects;}
	const std::map<int, float>& get_angle_effects() const {return angle_effects;}

	//: Returns an iterator to the font_marker associated with a particular glyph.
	//- <p> If a negative index is supplied, assumes glyph 0. If an index
	//- past the last element is supplied, assumes the last element. If
	//- there are no CL_Fonts in the GlyphBuffer, or if there is no CL_Font
	//- associated with that glyph, returns the end iterator.</p>
	std::map<int, CL_Font>::iterator get_font_for(int glyph_index);
	std::map<int, CL_Font>::const_iterator get_font_for(int glyph_index) const;
	
	//: Returns the width of the entire buffer.
	//return: The width in pixels.
	int get_width() const
		{return get_size().width;}
	
	//: Returns the height of the entire buffer.
	//return: The height in pixels.
	int get_height() const
		{return get_size().height;}
	
	//: Returns the size of the entire buffer.
	//return: The size in pixels.
	CL_Size get_size() const
		{CL_Rect brect = internal_rect(); return brect.get_size();}
	
	//: Returns current angle.
	//- <p> This produces (mostly) equivalent results to the rotation property of CL_Font. It also has
	//- the same limitation; the font's char rotation origin must be origin_top_left:0:0 for it
	//- to produce legible results . See CL_Font::set_glyph_rotation_hotspot(). </p>
	float get_angle() const
		{return rot_angle;}
	
	//: Returns scale for x and y.
	//- <p> 1.0f is normal scale, 2.0f is twice the size, etc. </p>
	void get_scale(float &x, float &y) const
		{x = scale_x; y = scale_y;}
	
	//: Returns current alpha.
	//- <p> Alpha 0.0f is full transparency, and 1.0f is full visibility (solid). </p>
	float get_alpha() const
		{return a;}
	
	//: Returns current color.
	//- <p> Alpha 0.0f is full transparency, and 1.0f is full visibility (solid). </p>
	void get_color(float &red, float &green, float &blue, float &alpha) const
		{red = r; green = g; blue = b; alpha = a;}

	//: Returns current blending functions.
	//- <p>This is applied to the font before the effect blend function (if any) is applied to the font color, with
	//- the GlyphBuffer color as the source and the font color as the destination.</p>
	void get_blend_func(CL_BlendFunc &src, CL_BlendFunc &dest) const
		{src = blend_src; dest = blend_dest;}

	//: Returns current color effect blending functions.
	//- <p>This is used to blend color effects is applied to the font color. It is applied after the CL_GlyphBuffer's
	//- color is applied through the CL_GlyphBuffer's regular blending functions. The effect color is the source,
	//- while the font color is the destination.</p>
	void get_effect_blend_func(CL_BlendFunc &src, CL_BlendFunc &dest) const
		{src = effect_blend_src; dest = effect_blend_dest;}
	
	//: Returns translation hotspot for the entire glyph field.
	void get_alignment(CL_Origin &origin, int &x, int &y) const
		{origin = trans_origin; x = trans_x; y = trans_y;}
	
	//: Returns rotation hotspot for the entire glyph field.
	void get_rotation_hotspot(CL_Origin &origin, int &x, int &y) const
		{origin = rot_origin; x = rot_x; y = rot_y;}
	
	//: Returns the area that would be drawn to if a draw were performed.
	//return: The area into which any pixels would be drawn. If buffer is empty, returns a rectangle with zero size.
	//param x, y: Anchor position to start "drawing" at. Actual rendering position depends on the alignment mode.
	//param dest: Rectangle to render glyphs in. The glyphs will be positioned within the rectangle depending on the alignment mode.
	CL_Rect bounding_rect(int x = 0, int y = 0) const
		{return bounding_rect(CL_Rect(x, y, x, y));}
	
	CL_Rect bounding_rect(const CL_Rect &dest) const;
	
	//: Returns the area occupied by glyphs inside the glyph buffer's coordinate space.
	//return: The area occupied by all glyphs, inside the buffer. If buffer is empty, returns a rectangle with zero size.
	CL_Rect internal_rect() const;
	
//! Operations
public:
	//: Add elements onto the end of the glyphs vector.
	//return: A reference to this object, allowing you to chain multiple calls of add() together.
	//param x, y: The position to add the characters at (always upper-left positioning inside the buffer).
	//param font: The font to set the characters to.
	//param character: A single character to add without preprocessing.
	//param other: Another CL_GlyphBuffer to copy all glyphs and data from to the specified position. Its (0,0) is mapped to (x,y).
	//- <p> It's usually more useful and efficient to have ClanLib draw many glyphs into the GlyphBuffer at once using
	//- CL_Font::draw_to_gb() or CL_TextStyler::draw_to_gb() than to use this function or the glyphs vector
	//- to manually add glyphs.</p>
	CL_GlyphBuffer &add(
		int x,
		int y,
		unsigned char character)
		{glyphs.push_back(Glyph(CL_Point(x,y), character)); return *this;}
	
	CL_GlyphBuffer &add(
		const CL_GlyphBuffer &other,
		int x = 0,
		int y = 0);
	
	//: Adds a font change marker, set to occur before the next glyph/location you add is drawn.
	//return: A reference to this object, allowing you to chain change_font() into chains of calls to add().
	//param font: The font to change to.
	CL_GlyphBuffer &change_font(const CL_Font &font);
	
	//: Remove all the character position and font marker data in the buffer.
	void clear();
	
	//: Draw the glyphs in the buffer.
	//param x, y: Anchor position to draw at. Actual drawing position depends on the alignment mode.
	//param gc: Graphic context on which to render. If null, will use CL_Display's current graphic context.
	//param dest: Rectangle to draw glyphs in. The glyphs will be positioned within the rectangle depending on the alignment mode.
	void draw(
		int x = 0,
		int y = 0,
		CL_GraphicContext *gc = 0)
		{draw(CL_Rect(x, y, x, y), gc);}
	
	void draw(
		const CL_Rect &dest,
		CL_GraphicContext *gc = 0)
		{CL_Rect irect = internal_rect(); draw_glyphs(i2e_offset(irect, dest), irect, gc);}

	//: Draws the glyphs in the buffer without bounding rect alignment.
	//param x, y: Glyphs are drawn with the CL_GlyphBuffer's internal origin at this location.
	//param gc: Graphic context on which to render. If null, will use CL_Display's current graphic context.
	void fixed_draw(
		int x = 0,
		int y = 0,
		CL_GraphicContext *gc = 0)
		{draw_glyphs(CL_Point(x, y), CL_Rect(0,0,0,0), gc);}
	
	//: Set absolute rotation angle.
	//- <p> This produces (mostly) equivalent results to the rotation property of CL_Font. It also has
	//- the same limitation; the font's char rotation origin must be origin_top_left:0:0 for it
	//- to produce legible results . See CL_Font::set_glyph_rotation_hotspot(). </p>
	void set_angle(float angle);
	
	//: Add angle to current angle.
	//- <p> This produces (mostly) equivalent results to the rotation property of CL_Font. It also has
	//- the same limitation; the font's char rotation origin must be origin_top_left:0:0 for it
	//- to produce legible results . See CL_Font::set_glyph_rotation_hotspot(). </p>
	void rotate(float angle)
		{angle += rot_angle; set_angle(angle);}
	
	//: Set scale for x and y directions individually.
	//- <p> 1.0f is normal scale, 2.0f is twice the size, etc. </p>
	void set_scale(float x, float y)
		{scale_x = x; scale_y = y;}
	
	//: Sets transparency.
	//- <p> Alpha 0.0f is full transparency, and 1.0f is full visibility (solid). </p>
	void set_alpha(float alpha)
		{a = alpha;}
	
	//: Sets the color.
	//- <p> Alpha 0.0f is full transparency, and 1.0f is full visibility (solid). </p>
	void set_color(float red, float green, float blue, float alpha = 1.0f)
		{r = red; g = green; b = blue; a = alpha;}
	
	//: Sets blending functions.
	//- <p>This is applied to the font before the effect blend function (if any) is applied to the font color, with
	//- the GlyphBuffer color as the source and the font color as the destination.</p>
	void set_blend_func(CL_BlendFunc &src, CL_BlendFunc &dest)
		{blend_src = src; blend_dest = dest;}
	
	//: Set color effect blending functions.
	//- <p>This is used to blend color effects is applied to the font color. It is applied after the CL_GlyphBuffer's
	//- color is applied through the CL_GlyphBuffer's regular blending functions. The effect color is the source,
	//- while the font color is the destination.</p>
	void set_effect_blend_func(CL_BlendFunc &src, CL_BlendFunc &dest)
		{effect_blend_src = src; effect_blend_dest = dest;}
	
	//: Sets translation hotspot.
	void set_alignment(CL_Origin origin, int x = 0, int y = 0)
		{trans_origin = origin; trans_x = x; trans_y = y;}

	//: Sets rotation hotspot.
	void set_rotation_hotspot(CL_Origin origin, int x = 0, int y = 0)
		{rot_origin = origin; rot_x = x; rot_y = y;}
	
//! Implementation:
private:
	//: Used to iterate through the buffer, changing font and applying effects as needed
	struct Iter;
	friend struct Iter;
	struct Iter {
		private:
			//Pointer to the GlyphBuffer
			const CL_GlyphBuffer* pgb;
			
			//Various GB iterators
			std::map<int, CL_Color>::const_iterator color_it;
			std::map<int, float>::const_iterator scale_x_it;
			std::map<int, float>::const_iterator scale_y_it;
			std::map<int, float>::const_iterator angle_it;
			std::map<int, CL_Font>::const_iterator font_it;
			
			bool advance_iters(); //Increments the iterators to match glyph_num, returns true if glyph_num valid
			void apply_iters(); //Copies and initializes the font from the current iterators
			
			//Becomes true if advance_iters changed the font
			bool font_changed;

			//The pre-effect state of effect modified sprite properties
			//FIXME: Track pre-effect color here too
			float pe_angle, pe_scale_x, pe_scale_y;
			
			//Don't allow default construction
			Iter();
			
		public:
			Iter(const CL_GlyphBuffer& new_gb);
			
			//Each increment, the sprite is restored from the font and effecterized
			CL_Sprite sprite;
			
			//The current index in the vector
			int glyph_num;
			
			int get_glyph_index(unsigned char letter)
				{return font_it->second.get_glyph_index(letter);}
			
			Iter& operator++();
			bool at_end();
	};
	
	//: Returns the offset to align an internal bounding rectangle to an external bounding rectangle.
	CL_Point i2e_offset(const CL_Rect& src, const CL_Rect& dest) const;
	
	//: Returns the point to scale away from.
	//param irect: Internal rectangle containing all glyphs
	CL_Point get_scale_hotspot(const CL_Rect& irect) const;
	
	//: Returns the point to rotate around.
	//param irect: Internal rectangle containing all glyphs
	CL_Point get_rot_hotspot(const CL_Rect& irect) const;
	
	//: Draws all the glyphs in the buffer.
	//param offset: Added to each position right before drawing, to move to external coords
	//param irect: Result of internal_rect() passed in by caller, or CL_Rect(0,0,0,0)
	//param gc: The graphics context to draw to (default if 0)
	void draw_glyphs(
		const CL_Point& offset,
		CL_Rect irect,
		CL_GraphicContext* gc);
	
	//: Color.
	float r, g, b, a;
	
	//: Scale.
	float scale_x;
	float scale_y;
	
	//: Angle and rotation hotspot.
	float rot_angle;
	CL_Origin rot_origin;
	int rot_x;
	int rot_y;
	
	//: Alignment.
	CL_Origin trans_origin;
	int trans_x;
	int trans_y;
	
	//: The glyphs and their positions.
	std::vector<Glyph> glyphs;
	
	//: The font change markers.
	std::map<int, CL_Font> font_markers;
	
	//: Color, scale, and angle font effect maps.
	std::map<int, CL_Color> color_effects;
	std::map<int, float> scale_x_effects;
	std::map<int, float> scale_y_effects;
	std::map<int, float> angle_effects;

	//: The standard and effect color blending functions.
	CL_BlendFunc blend_src;
	CL_BlendFunc blend_dest;
	CL_BlendFunc effect_blend_src;
	CL_BlendFunc effect_blend_dest;
};

#endif
