/*
	$Id: text_styler.h,v 1.4 2003/02/05 04:38:24 plasmoid Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	See http://www.clanlib.org
	------------------------------------------------------------------------
*/

//! clanDisplay="Fonts"
//! header=display.h

#ifndef header_text_styler
#define header_text_styler

#ifdef WIN32
#pragma warning( disable : 4786)
#endif

#include <string>
#include <map>

#include "font.h"
#include "../Core/Math/origin.h"
#include "../Core/Math/rect.h"
#include "../Core/Math/point.h"

class CL_GraphicContext;
class CL_GlyphBuffer;

//: Draws specially formatted strings in one or several fonts.
//- <p> A CL_TextStyler consists of a collection of named CL_Fonts.
//- When passed a string to draw, it uses HTML tags to select 
//- which font to draw text in. For example:</p>
//- <code> Here's some text, {italic}isn't it {bold}amazing?{/bold}{/italic}</code>
//- <p> CL_TextStyler does not have any mechanisms on its own for bold
//- and italic text, so
//- in order for this to work, there must be fonts named 'bold' and
//- 'italic' that it knows of. Text outside font tags is
//- drawn in the font named 'default', if it exists. No blending 
//- or anything like that is performed with nested fonts; the innermost font
//- is the one used, and the nesting is for convienence. </p>
//- <p> The parser is pretty laid back. It doesn't mind if you don't put in closing tags
//- and let a tag pair run off the end of the string. References to fonts which are
//- not in the CL_TextStyler's font map will be ignored. Using bad nesting order is
//- also allowed (i.e. {a}{b}{/a}{/b}), as is using the 'default' font
//- manually if you want to. To draw a literal brace character, put two
//- {'s or }'s in a row. </p>
//- <p> In the future, other effects will be supported using [] tags.
//- Although none of this is implemented now, for forwards compatability
//- ]'s and ['s are treated as special characters and must be escaped the same way
//- as brace characters, by writing them twice in a row. </p>

class CL_TextStyler
{
//! Construction:
public:
	//: Constructs a text styler.
	CL_TextStyler();
	
	virtual ~CL_TextStyler() {;}

//! Subclasses:
public:
	//: The return type for test_draw(), draw(), and draw_to_gb()
	struct PrintInfo
	{
		//: A rectangle containing all characters that were drawn.
		//- <p> This rectangle compensates for total rotation, offsets,
		//- and scaling, but not for char rotation in the font. </p>
		CL_Rect bounds;
		
		//: The upper left corner of where the next glyph would've been drawn.
		CL_Point cursor;
		
		//: The last character in the string drawn, exclusive.
		//- <p>Formatting tags are considered to have been "drawn"
		//- the moment they're encountered, even though they only have
		//- an indirect effect on what's drawn.</p>
		std::string::const_iterator end;
	};
//! Attributes:
public:
	//: Returns a reference to the font map.
	std::map<std::string, CL_Font>& get_fonts()
		{return fonts;}
	
	const std::map<std::string, CL_Font>& get_fonts() const
		{return fonts;}
	
	//: Returns current angle in degrees.
	float get_angle() const
		{return rot_angle;}

	//: Returns scale for x and y.
	//- <p> 1.0f is normal scale, 2.0f is twice the size, etc. </p>
	void get_scale(float &x, float &y) const
		{x = scale_x, y = scale_y;}

	//: Returns current alpha.
	//- <p> Alpha 0.0f is full transparency, and 1.0f is full visibility (solid). </p>
	float get_alpha() const
		{return col_a;}

	//: Returns current color.
	//- <p> Alpha 0.0f is full transparency, and 1.0f is full visibility (solid). </p>
	void get_color(float &red, float &green, float &blue, float &alpha) const
		{red = col_a; green = col_g; blue = col_b; alpha = col_a;}
	
	//: Returns current blending functions.
	void get_blend_func(CL_BlendFunc &src, CL_BlendFunc &dest) const
		{src = blend_src; dest = blend_dest;}
	
	//: Returns translation hotspot.
	void get_alignment(CL_Origin &origin, int &x, int &y) const
		{origin = trans_origin; x = trans_x; y = trans_y;}

	//: Returns rotation hotspot.
	void get_rotation_hotspot(CL_Origin &origin, int &x, int &y) const
		{origin = rot_origin; x = rot_x; y = rot_y;}
	
	//: Returns the drawn height of some source text.
	//return: The height in pixels.
	//param str: The source string to parse and render.
	//param start: A starting iterator, inclusive.
	//param end: An ending iterator, exclusive.
	//param max_size: Acts like the size of a dest rectangle passed to draw(), for height truncating and word wrapping.
	//param delim: When word-wrapping, specifies characters (other than newlines) that act as word seperators.
	int CL_TextStyler::get_height(
		const std::string &str,
		const CL_Size &max_size = CL_Size(0,0),
		const std::string &delim = std::string(" ")) const
		{return get_size(str, max_size, delim).height;}
	
	int CL_TextStyler::get_height(
		const std::string::const_iterator& start,
		const std::string::const_iterator& end,
		const CL_Size &max_size = CL_Size(0,0),
		const std::string &delim = std::string(" ")) const
		{return get_size(start, end, max_size, delim).height;}
	
	//: Returns the drawn width of some source text.
	//return: The width in pixels.
	//param str: The source string to parse and render.
	//param start: A starting iterator, inclusive.
	//param end: An ending iterator, exclusive.
	//param max_size: Acts like the size of a dest rectangle passed to draw(), for height truncating and word wrapping.
	//param delim: When word-wrapping, specifies characters (other than newlines) that act as word seperators.
	int CL_TextStyler::get_width(
		const std::string &str,
		const CL_Size &max_size = CL_Size(0,0),
		const std::string &delim = std::string(" ")) const
		{return get_size(str, max_size, delim).width;}
	
	int CL_TextStyler::get_width(
		const std::string::const_iterator& start,
		const std::string::const_iterator& end,
		const CL_Size &max_size = CL_Size(0,0),
		const std::string &delim = std::string(" ")) const
		{return get_size(start, end, max_size, delim).width;}
	
	//: Returns the drawn size of some source text.
	//return: The size in pixels.
	//param str: The source string to parse and render.
	//param start: A starting iterator, inclusive.
	//param end: An ending iterator, exclusive.
	//param max_size: Acts like the size of a dest rectangle passed to draw(), for height truncating and word wrapping.
	//param delim: When word-wrapping, specifies characters (other than newlines) that act as word seperators.
	CL_Size CL_TextStyler::get_size(
		const std::string &str,
		const CL_Size &max_size = CL_Size(0,0),
		const std::string &delim = std::string(" ")) const
		{return get_size(str.begin(), str.end(), max_size, delim);}
	
	CL_Size CL_TextStyler::get_size(
		const std::string::const_iterator& start,
		const std::string::const_iterator& end,
		const CL_Size &max_size = CL_Size(0,0),
		const std::string &delim = std::string(" ")) const
		{return test_draw(CL_Rect(0, 0, max_size.width, max_size.height), start, end, delim).bounds.get_size();}
	
	//: Get the return value of a draw() without drawing anything.
	//return: A CL_TextStyler::PrintInfo, containing stats about the text that would've been drawn.
	//param x, y: Anchor position of where to render text. Actual rendering position depends on the alignment mode.
	//param dest: Rectangle to draw text in. The text will be word-wrapped against delimiters to fit inside the rectangle.
	//param str: The source string to parse and render.
	//param start: String position to begin rendering at, inclusive.
	//param end: String position to end rendering at, exclusive.
	//param delim: When word-wrapping, specifies characters (other than '\n') that act as word seperators.
	CL_TextStyler::PrintInfo test_draw(
		int x,
		int y,
		const std::string& str) const
		{return test_draw(CL_Rect(x, y, x, y), str.begin(), str.end());}
	
	CL_TextStyler::PrintInfo test_draw(
		const CL_Rect &dest,
		const std::string& str,
		const std::string& delim = std::string(" ")) const
		{return test_draw(dest, str.begin(), str.end(), delim);}
	
	CL_TextStyler::PrintInfo test_draw(
		int x,
		int y,
		const std::string::const_iterator& start,
		const std::string::const_iterator& end) const
		{return test_draw(CL_Rect(x, y, x, y), start, end);}
	
	CL_TextStyler::PrintInfo test_draw(
		const CL_Rect &dest,
		const std::string::const_iterator& start,
		const std::string::const_iterator& end,
		const std::string& delim = std::string(" ")) const;
//! Operations
	
	//: Copy assignment operator.
	CL_TextStyler &operator =(const CL_TextStyler &copy);
	
	//: Adds a new font to the map
	void add_font(std::string name, const CL_Font& font)
		{fonts.insert(std::pair<std::string, CL_Font>(name, font));}
	
	//: Draws text onto a graphics context.
	//return: A CL_TextStyler::PrintInfo, containing stats about the text that was drawn.
	//param x, y: Anchor position of where to draw at. Actual rendering position depends on the alignment mode.
	//param dest: Rectangle to draw text in. The text will be word-wrapped against delimiters to fit inside the rectangle.
	//param str: The source string to parse and render.
	//param start: String position to begin rendering at, inclusive.
	//param end: String position to end rendering at, exclusive.
	//param delim: When word-wrapping, specifies characters (other than '\n') that act as word seperators.
	//param gc: Graphic context on which to render upon. If null, will use CL_Display's current graphic context.
	CL_TextStyler::PrintInfo draw(
		int x,
		int y,
		const std::string& str,
		CL_GraphicContext *context = 0)
		{return draw(CL_Rect(x, y, x, y), str.begin(), str.end(), std::string(" "), context);}
	
	CL_TextStyler::PrintInfo draw(
		const CL_Rect& dest,
		const std::string& str,
		const std::string& delim = std::string(" "),
		CL_GraphicContext *context = 0)
		{return draw(dest, str.begin(), str.end(), delim, context);}
	
	CL_TextStyler::PrintInfo draw(
		int x,
		int y,
		const std::string::const_iterator& start,
		const std::string::const_iterator& end,
		CL_GraphicContext *context = 0)
		{return draw(CL_Rect(x, y, x, y), start, end, std::string(" "), context);}
	
	CL_TextStyler::PrintInfo draw(
		const CL_Rect& dest,
		const std::string::const_iterator& start,
		const std::string::const_iterator& end,
		const std::string& delim = std::string(" "),
		CL_GraphicContext *context = 0);
	
	//: Inserts data into a CL_GlyphBuffer.
	//return: A CL_TextStyler::PrintInfo, containing stats about the glyphs that were inserted.
	//param x, y: Anchor position of where to render text. Actual rendering position depends on the alignment mode.
	//param str: The source string to parse and render.
	//param start: String position to begin rendering at, inclusive.
	//param end: String position to end rendering at, exclusive.
	//param gb: Glyph buffer to insert characters/fonts into.
	//param delim: When word-wrapping, specifies characters (other than '\n') that act as word seperators.
	CL_TextStyler::PrintInfo draw_to_gb(
		int x,
		int y,
		const std::string& str,
		const CL_GlyphBuffer& gb) const
		{return draw_to_gb(CL_Rect(x, y, x, y), str.begin(), str.end(), gb);}
	
	CL_TextStyler::PrintInfo draw_to_gb(
		const CL_Rect &dest,
		const std::string& str,
		const CL_GlyphBuffer& gb,
		const std::string &delim = std::string(" ")) const
		{return draw_to_gb(dest, str.begin(), str.end(), gb, delim);}
	
	CL_TextStyler::PrintInfo draw_to_gb(
		int x,
		int y,
		const std::string::const_iterator &start,
		const std::string::const_iterator &end,
		const CL_GlyphBuffer& gb) const
		{return draw_to_gb(CL_Rect(x, y, x, y), start, end, gb);}
	
	CL_TextStyler::PrintInfo draw_to_gb(
		const CL_Rect &dest,
		const std::string::const_iterator &start,
		const std::string::const_iterator &end,
		const CL_GlyphBuffer& gb,
		const std::string &delim = std::string(" ")) const;
	
	//: Set absolute rotation angle in degrees.
	void set_angle(float angle);
	
	//: Add angle in degrees to current angle.
	void rotate(float angle)
		{angle += rot_angle; set_angle(angle);}
	
	//: Set scale for x and y directions individually.
	//- <p> 1.0f is normal scale, 2.0f is twice the size, etc. </p>
	void set_scale(float x, float y)
		{scale_x = x; scale_y = y;}
	
	//: Sets transparency.
	//- <p> Alpha 0.0f is full transparency, and 1.0f is full visibility (solid). </p>
	void set_alpha(float alpha)
		{col_a = alpha;}
	
	//: Sets the color.
	//- <p> Alpha 0.0f is full transparency, and 1.0f is full visibility (solid). </p>
	void set_color(float r, float g, float b, float a = 1.0f)
		{col_r = r; col_g = g; col_b = b; col_a = a;}
	
	//: Sets blending functions.
	void set_blend_func(CL_BlendFunc &src, CL_BlendFunc &dest)
		{blend_src = src; blend_dest = dest;}
	
	//: Sets translation hotspot.
	void set_alignment(CL_Origin origin, int x = 0, int y = 0)
		{trans_origin = origin; trans_x = x; trans_y = y;}
	
	//: Sets rotation hotspot.
	void set_rotation_hotspot(CL_Origin origin, int x = 0, int y = 0)
		{rot_origin = origin; rot_x = x; rot_y = y;}
	
// Implementation:
private:
	//: Angle and rotation hotspot.
	float rot_angle;
	CL_Origin rot_origin;
	int rot_x;
	int rot_y;
	
	//: Scale.
	float scale_x;
	float scale_y;
	
	//: Color.
	float col_r;
	float col_g;
	float col_b;
	float col_a;
	
	//: Color blending functions.
	CL_BlendFunc blend_src;
	CL_BlendFunc blend_dest;
	
	//: Alignment.
	CL_Origin trans_origin;
	int trans_x;
	int trans_y;
	
	//: The font map.
	std::map<std::string, CL_Font> fonts;
};

#endif
