/*
	$Id: soundbuffer.h,v 1.13 2002/07/15 16:49:50 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	See http://www.clanlib.org
	------------------------------------------------------------------------
*/

//! clanSound="Audio Mixing"
//! header=sound.h

#ifndef header_soundbuffer
#define header_soundbuffer

#include "soundbuffer_session.h"
#include "../Core/Resources/resource_manager.h"
#include "../Core/Resources/resource.h"

class CL_StaticSoundProvider;
class CL_StreamSoundProvider;
class CL_SoundCard;
class CL_SoundBuffer_Generic;

class CL_SoundBuffer
//: Sample interface in ClanLib.
//- <p>The CL_SoundBuffer class represents a sample in ClanLib. It can
//- either be static or streamed. The soundbuffer gets its sample data from
//- a soundprovider, that is passed during construction.</p>
//-
//- <p>A static sample are normally sound effects and other sounds, that do
//- not change. ClanLib will always load the entire sample into memory, and
//- possibly upload it to the soundcard.</p>
//-
//- <p>Streamed samples are either large sample that should be loaded a bit
//- at a time (music for instance), or sounds that change from playback to
//- playback (a microphone, speech over the net, etc).</p>
//-
//also: CL_StaticSoundProvider
//also: CL_StreamSoundProvider
//also: CL_SoundBuffer_Session
{
public:
//! Construction:
	//: Construct sound buffer.
	//- <p>A sound buffer can be constructed either as static or
	//- streamed. If the sound buffer is loaded from resources, the
	//- buffer type is determined by the resource option 'stream'
	//- associated with the resource.</p>
	//-
	//- <p>CL_SoundBuffer's internals are reference counted, so the copy
	//- constructor will create a new soundbuffer object which shares
	//- the same buffer as the original one. This means that if the copy
	//- is modified, the original is affected as well.</p>
	//-
	//- <p>If <i>delete_provider</i> is true, the provider will be
	//- deleted when the soundbuffer is deleted.</p>
	
	CL_SoundBuffer();
	
	CL_SoundBuffer(
		const std::string &res_id,
		CL_ResourceManager *manager);

	CL_SoundBuffer(
		CL_StaticSoundProvider *provider,
		bool delete_provider=false);
	
	CL_SoundBuffer(
		CL_StreamSoundProvider *provider,
		bool delete_provider=false);

	CL_SoundBuffer(const CL_SoundBuffer &copy);

	virtual ~CL_SoundBuffer();

//! Operators:
	//: Copy assignment.
	void operator =(const CL_SoundBuffer &copy);

//! Attributes:
	//: Returns the static soundprovider attached or NULL if it is a streamed
	//: sound buffer.
	//return: The static soundprovider or NULL if none.
	CL_StaticSoundProvider *get_static_provider() const;

	//: Return the streamed soundprovider attached or NULL if it is a static
	//: sound buffer.
	//return: The streamed soundprovider of NULL if none.
	CL_StreamSoundProvider *get_stream_provider() const;

	//: Returns the length of the soundbuffer. 
	//: That is num_samples * bytes_per_sample.
	//return: Length of soundbuffer.
	int get_length() const;
	
	//: Returns the number of samples in the soundbuffer.
	//return: Number of samples in soundbuffer.
	int get_num_samples() const;

	//: Returns the start frequency used when the buffer is played.
	//return: Default/start frequency.
	int get_frequency() const;
	
//! Operations:
	//: Sets the start frequency used when the buffer is played.
	//param new_freq: The new default frequency.
	//return: True if successful, false otherwise.
	bool set_frequency(int new_freq);

	//: Returns the start/default volume used when the buffer is played.
	//return: The volume.
	float get_volume() const;
	
	//: Sets the default volume used when the buffer is played.
	//param new_volume: The new default volume.
	//return: True if successful, false otherwise.
	bool set_volume(float new_volume);

	//: Returns the default panning position when the buffer is played.
	//return: The panning position.
	float get_pan() const;

	//: Sets the default panning position when the buffer is played.
	//param new_pan: The new default panning position.
	//return: True if successful, false otherwise.
	bool set_pan(float new_pan);

	//: Returns true if an instance of this soundbuffer is playing
	//param session: Pointer to session-pointer, set to point to the first session playing this soundbuffer.
	//return: Return value is only valid if function returns true
	bool is_playing(CL_SoundBuffer_Session **session=NULL, CL_SoundCard *card=NULL) const;

	//: Stops any sessions playing this soundbuffer
	void stop(CL_SoundCard *card=NULL);

	//: Plays the soundbuffer on the specified soundcard 
	//: and using the specified playback description.
	//param card: Soundcard to be used - NULL means use the current selected sound card (CL_Sound::get_selected_card().
	//return: The playback session.
	CL_SoundBuffer_Session play(bool looping=false, CL_SoundCard *card=NULL);

	//: Prepares the soundbuffer for playback on the specified soundcard.
	//param card: Soundcard to be used or NULL to use CL_Sound::get_selected_card().
	//return: The playback session.
	CL_SoundBuffer_Session prepare(bool looping=false, CL_SoundCard *card=NULL);

//! Implementation:
public:
	CL_SoundBuffer_Generic *impl;
	CL_Resource resource;
};

#endif
