/*
www.sourceforge.net/projects/tinyxml
Original code (2.0 and earlier )copyright (c) 2000-2002 Lee Thomason (www.grinninglizard.com)

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any
damages arising from the use of this software.

Permission is granted to anyone to use this software for any
purpose, including commercial applications, and to alter it and
redistribute it freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must
not claim that you wrote the original software. If you use this
software in a product, an acknowledgment in the product documentation
would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and
must not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source
distribution.
*/

#include "Core/precomp.h"
#include <ctype.h>
#include "API/Core/XML/tinyxml.h"

bool CL_XmlBase::condenseWhiteSpace = true;

void CL_XmlBase::PutString( const CL_XML_STRING& str, CL_XML_OSTREAM* stream )
{
	CL_XML_STRING buffer;
	PutString( str, &buffer );
	(*stream) << buffer;
}

void CL_XmlBase::PutString( const CL_XML_STRING& str, CL_XML_STRING* outString )
{
	int i=0;

	while( i<(int)str.length() )
	{
		int c = str[i];

		if (    c == '&' 
		     && i < ( (int)str.length() - 2 )
			 && str[i+1] == '#'
			 && str[i+2] == 'x' )
		{
			// Hexadecimal character reference.
			// Pass through unchanged.
			// &#xA9;	-- copyright symbol, for example.
			while ( i<(int)str.length() )
			{
				outString->append( str.c_str() + i, 1 );
				++i;
				if ( str[i] == ';' )
					break;
			}
		}
		else if ( c == '&' )
		{
			outString->append( entity[0].str, entity[0].strLength );
			++i;
		}
		else if ( c == '<' )
		{
			outString->append( entity[1].str, entity[1].strLength );
			++i;
		}
		else if ( c == '>' )
		{
			outString->append( entity[2].str, entity[2].strLength );
			++i;
		}
		else if ( c == '\"' )
		{
			outString->append( entity[3].str, entity[3].strLength );
			++i;
		}
		else if ( c == '\'' )
		{
			outString->append( entity[4].str, entity[4].strLength );
			++i;
		}
		else if ( c < 32 || c > 126 )
		{
			// Easy pass at non-alpha/numeric/symbol
			// 127 is the delete key. Below 32 is symbolic.
			char buf[ 32 ];
			sprintf( buf, "&#x%02X;", (unsigned) ( c & 0xff ) );
			outString->append( buf, strlen( buf ) );
			++i;
		}
		else
		{
			char realc = (char) c;
			outString->append( &realc, 1 );
			++i;
		}
	}
}


// <-- Strange class for a bug fix. Search for STL_STRING_BUG
CL_XmlBase::StringToBuffer::StringToBuffer( const CL_XML_STRING& str )
{
	buffer = new char[ str.length()+1 ];
	if ( buffer )
	{
		strcpy( buffer, str.c_str() );
	}
}


CL_XmlBase::StringToBuffer::~StringToBuffer()
{
	delete [] buffer;
}
// End strange bug fix. -->


CL_XmlNode::CL_XmlNode( NodeType _type )
{
	parent = 0;
	type = _type;
	firstChild = 0;
	lastChild = 0;
	prev = 0;
	next = 0;
	userData = 0;
}


CL_XmlNode::~CL_XmlNode()
{
	CL_XmlNode* node = firstChild;
	CL_XmlNode* temp = 0;

	while ( node )
	{
		temp = node;
		node = node->next;
		delete temp;
	}	
}


void CL_XmlNode::Clear()
{
	CL_XmlNode* node = firstChild;
	CL_XmlNode* temp = 0;

	while ( node )
	{
		temp = node;
		node = node->next;
		delete temp;
	}	

	firstChild = 0;
	lastChild = 0;
}


CL_XmlNode* CL_XmlNode::LinkEndChild( CL_XmlNode* node )
{
	node->parent = this;

	node->prev = lastChild;
	node->next = 0;

	if ( lastChild )
		lastChild->next = node;
	else
		firstChild = node;			// it was an empty list.

	lastChild = node;
	return node;
}


CL_XmlNode* CL_XmlNode::InsertEndChild( const CL_XmlNode& addThis )
{
	CL_XmlNode* node = addThis.Clone();
	if ( !node )
		return 0;

	return LinkEndChild( node );
}


CL_XmlNode* CL_XmlNode::InsertBeforeChild( CL_XmlNode* beforeThis, const CL_XmlNode& addThis )
{	
	if ( !beforeThis || beforeThis->parent != this )
		return 0;

	CL_XmlNode* node = addThis.Clone();
	if ( !node )
		return 0;
	node->parent = this;

	node->next = beforeThis;
	node->prev = beforeThis->prev;
	if ( beforeThis->prev )
	{
		beforeThis->prev->next = node;
	}
	else
	{
		assert( firstChild == beforeThis );
		firstChild = node;
	}
	beforeThis->prev = node;
	return node;
}


CL_XmlNode* CL_XmlNode::InsertAfterChild( CL_XmlNode* afterThis, const CL_XmlNode& addThis )
{
	if ( !afterThis || afterThis->parent != this )
		return 0;

	CL_XmlNode* node = addThis.Clone();
	if ( !node )
		return 0;
	node->parent = this;

	node->prev = afterThis;
	node->next = afterThis->next;
	if ( afterThis->next )
	{
		afterThis->next->prev = node;
	}
	else
	{
		assert( lastChild == afterThis );
		lastChild = node;
	}
	afterThis->next = node;
	return node;
}


CL_XmlNode* CL_XmlNode::ReplaceChild( CL_XmlNode* replaceThis, const CL_XmlNode& withThis )
{
	if ( replaceThis->parent != this )
		return 0;

	CL_XmlNode* node = withThis.Clone();
	if ( !node )
		return 0;

	node->next = replaceThis->next;
	node->prev = replaceThis->prev;

	if ( replaceThis->next )
		replaceThis->next->prev = node;
	else
		lastChild = node;

	if ( replaceThis->prev )
		replaceThis->prev->next = node;
	else
		firstChild = node;

	delete replaceThis;
	node->parent = this;
	return node;
}


bool CL_XmlNode::RemoveChild( CL_XmlNode* removeThis )
{
	if ( removeThis->parent != this )
	{	
		assert( 0 );
		return false;
	}

	if ( removeThis->next )
		removeThis->next->prev = removeThis->prev;
	else
		lastChild = removeThis->prev;

	if ( removeThis->prev )
		removeThis->prev->next = removeThis->next;
	else
		firstChild = removeThis->next;

	delete removeThis;
	return true;
}

CL_XmlNode* CL_XmlNode::FirstChild( const char * value ) const
{
	CL_XmlNode* node;
	for ( node = firstChild; node; node = node->next )
	{
		if ( node->SValue() == CL_XML_STRING( value ))
			return node;
	}
	return 0;
}

CL_XmlNode* CL_XmlNode::LastChild( const char * value ) const
{
	CL_XmlNode* node;
	for ( node = lastChild; node; node = node->prev )
	{
		if ( node->SValue() == CL_XML_STRING (value))
			return node;
	}
	return 0;
}

CL_XmlNode* CL_XmlNode::IterateChildren( CL_XmlNode* previous ) const
{
	if ( !previous )
	{
		return FirstChild();
	}
	else
	{
		assert( previous->parent == this );
		return previous->NextSibling();
	}
}

CL_XmlNode* CL_XmlNode::IterateChildren( const char * val, CL_XmlNode* previous ) const
{
	if ( !previous )
	{
		return FirstChild( val );
	}
	else
	{
		assert( previous->parent == this );
		return previous->NextSibling( val );
	}
}

CL_XmlNode* CL_XmlNode::NextSibling( const char * value ) const
{
	CL_XmlNode* node;
	for ( node = next; node; node = node->next )
	{
		if ( node->SValue() == CL_XML_STRING (value))
			return node;
	}
	return 0;
}


CL_XmlNode* CL_XmlNode::PreviousSibling( const char * value ) const
{
	CL_XmlNode* node;
	for ( node = prev; node; node = node->prev )
	{
		if ( node->SValue() == CL_XML_STRING (value))
			return node;
	}
	return 0;
}

void CL_XmlElement::RemoveAttribute( const char * name )
{
	CL_XmlAttribute* node = attributeSet.Find( name );
	if ( node )
	{
		attributeSet.Remove( node );
		delete node;
	}
}

CL_XmlElement* CL_XmlNode::FirstChildElement() const
{
	CL_XmlNode* node;

	for (	node = FirstChild();
	node;
	node = node->NextSibling() )
	{
		if ( node->ToElement() )
			return node->ToElement();
	}
	return 0;
}

CL_XmlElement* CL_XmlNode::FirstChildElement( const char * value ) const
{
	CL_XmlNode* node;

	for (	node = FirstChild( value );
	node;
	node = node->NextSibling( value ) )
	{
		if ( node->ToElement() )
			return node->ToElement();
	}
	return 0;
}


CL_XmlElement* CL_XmlNode::NextSiblingElement() const
{
	CL_XmlNode* node;

	for (	node = NextSibling();
	node;
	node = node->NextSibling() )
	{
		if ( node->ToElement() )
			return node->ToElement();
	}
	return 0;
}

CL_XmlElement* CL_XmlNode::NextSiblingElement( const char * value ) const
{
	CL_XmlNode* node;

	for (	node = NextSibling( value );
	node;
	node = node->NextSibling( value ) )
	{
		if ( node->ToElement() )
			return node->ToElement();
	}
	return 0;
}



CL_XmlDocument* CL_XmlNode::GetDocument() const
{
	const CL_XmlNode* node;

	for( node = this; node; node = node->parent )
	{
		if ( node->ToDocument() )
			return node->ToDocument();
	}
	return 0;
}


CL_XmlElement::CL_XmlElement (const char * _value)
: CL_XmlNode( CL_XmlNode::ELEMENT )
{
	firstChild = lastChild = 0;
	value = _value;
}

CL_XmlElement::~CL_XmlElement()
{
	while( attributeSet.First() )
	{
		CL_XmlAttribute* node = attributeSet.First();
		attributeSet.Remove( node );
		delete node;
	}
}

const char * CL_XmlElement::Attribute( const char * name ) const
{
	CL_XmlAttribute* node = attributeSet.Find( name );

	if ( node )
		return node->Value();

	return 0;
}


const char * CL_XmlElement::Attribute( const char * name, int* i ) const
{
	const char * s = Attribute( name );
	if ( i )
	{
		if ( s )
			*i = atoi( s );
		else
			*i = 0;
	}
	return s;
}


void CL_XmlElement::SetAttribute( const char * name, int val )
{	
	char buf[64];
	sprintf( buf, "%d", val );
	SetAttribute( name, buf );
}


void CL_XmlElement::SetAttribute( const char * name, const char * value )
{
	CL_XmlAttribute* node = attributeSet.Find( name );
	if ( node )
	{
		node->SetValue( value );
		return;
	}

	CL_XmlAttribute* attrib = new CL_XmlAttribute( name, value );
	if ( attrib )
	{
		attributeSet.Add( attrib );
	}
	else
	{
		CL_XmlDocument* document = GetDocument();
		if ( document ) document->SetError( CL_XML_ERROR_OUT_OF_MEMORY );
	}
}

void CL_XmlElement::Print( FILE* cfile, int depth ) const
{
	int i;
	for ( i=0; i<depth; i++ )
	{
		fprintf( cfile, "    " );
	}

	fprintf( cfile, "<%s", value.c_str() );

	CL_XmlAttribute* attrib;
	for ( attrib = attributeSet.First(); attrib; attrib = attrib->Next() )
	{
		fprintf( cfile, " " );
		attrib->Print( cfile, depth );
	}

	// There are 3 different formatting approaches:
	// 1) An element without children is printed as a <foo /> node
	// 2) An element with only a text child is printed as <foo> text </foo>
	// 3) An element with children is printed on multiple lines.
	CL_XmlNode* node;
	if ( !firstChild )
	{
		fprintf( cfile, " />" );
	}
	else if ( firstChild == lastChild && firstChild->ToText() )
	{
		fprintf( cfile, ">" );
		firstChild->Print( cfile, depth + 1 );
		fprintf( cfile, "</%s>", value.c_str() );
	}
	else
	{
		fprintf( cfile, ">" );

		for ( node = firstChild; node; node=node->NextSibling() )
		{
			if ( !node->ToText() )
			{
				fprintf( cfile, "\n" );
			}
			node->Print( cfile, depth+1 );
		}
		fprintf( cfile, "\n" );
		for( i=0; i<depth; ++i )
		fprintf( cfile, "    " );
		fprintf( cfile, "</%s>", value.c_str() );
	}
}

void CL_XmlElement::StreamOut( CL_XML_OSTREAM * stream ) const
{
	(*stream) << "<" << value;

	CL_XmlAttribute* attrib;
	for ( attrib = attributeSet.First(); attrib; attrib = attrib->Next() )
	{	
		(*stream) << " ";
		attrib->StreamOut( stream );
	}

	// If this node has children, give it a closing tag. Else
	// make it an empty tag.
	CL_XmlNode* node;
	if ( firstChild )
	{ 		
		(*stream) << ">";

		for ( node = firstChild; node; node=node->NextSibling() )
		{
			node->StreamOut( stream );
		}
		(*stream) << "</" << value << ">";
	}
	else
	{
		(*stream) << " />";
	}
}

CL_XmlNode* CL_XmlElement::Clone() const
{
	CL_XmlElement* clone = new CL_XmlElement( Value() );
	if ( !clone )
		return 0;

	CopyToClone( clone );

	// Clone the attributes, then clone the children.
	CL_XmlAttribute* attribute = 0;
	for(	attribute = attributeSet.First();
	attribute;
	attribute = attribute->Next() )
	{
		clone->SetAttribute( attribute->Name(), attribute->Value() );
	}

	CL_XmlNode* node = 0;
	for ( node = firstChild; node; node = node->NextSibling() )
	{
		clone->LinkEndChild( node->Clone() );
	}
	return clone;
}


CL_XmlDocument::CL_XmlDocument() : CL_XmlNode( CL_XmlNode::DOCUMENT )
{
	error = false;
	//	ignoreWhiteSpace = true;
}

CL_XmlDocument::CL_XmlDocument( const char * documentName ) : CL_XmlNode( CL_XmlNode::DOCUMENT )
{
	//	ignoreWhiteSpace = true;
	value = documentName;
	error = false;
}

bool CL_XmlDocument::LoadFile()
{
	// See STL_STRING_BUG below.
	StringToBuffer buf( value );

	if ( buf.buffer && LoadFile( buf.buffer ) )
		return true;

	return false;
}


bool CL_XmlDocument::SaveFile() const
{
	// See STL_STRING_BUG below.
	StringToBuffer buf( value );

	if ( buf.buffer && SaveFile( buf.buffer ) )
		return true;

	return false;
}

bool CL_XmlDocument::LoadFile( const char* filename )
{
	// Delete the existing data:
	Clear();

	// There was a really terrifying little bug here. The code:
	//		value = filename
	// in the STL case, cause the assignment method of the std::string to
	// be called. What is strange, is that the std::string had the same
	// address as it's c_str() method, and so bad things happen. Looks
	// like a bug in the Microsoft STL implementation.
	// See STL_STRING_BUG above.
	// Fixed with the StringToBuffer class.
	value = filename;

	FILE* file = fopen( value.c_str (), "r" );

	if ( file )
	{
		// Get the file size, so we can pre-allocate the string. HUGE speed impact.
		long length = 0;
		fseek( file, 0, SEEK_END );
		length = ftell( file );
		fseek( file, 0, SEEK_SET );

		// Strange case, but good to handle up front.
		if ( length == 0 )
		{
			fclose( file );
			return false;
		}

		// If we have a file, assume it is all one big XML file, and read it in.
		// The document parser may decide the document ends sooner than the entire file, however.
		CL_XML_STRING data;
		data.reserve( length );

		const int BUF_SIZE = 2048;
		char buf[BUF_SIZE];

		while( fgets( buf, BUF_SIZE, file ) )
		{
			data += buf;
		}
		fclose( file );

		Parse( data.c_str() );
		if (  !Error() )
		{
			return true;
		}
	}
	SetError( CL_XML_ERROR_OPENING_FILE );
	return false;
}

bool CL_XmlDocument::SaveFile( const char * filename ) const
{
	// The old c stuff lives on...
	FILE* fp = fopen( filename, "w" );
	if ( fp )
	{
		Print( fp, 0 );
		fclose( fp );
		return true;
	}
	return false;
}


CL_XmlNode* CL_XmlDocument::Clone() const
{
	CL_XmlDocument* clone = new CL_XmlDocument();
	if ( !clone )
		return 0;

	CopyToClone( clone );
	clone->error = error;
	clone->errorDesc = errorDesc.c_str ();

	CL_XmlNode* node = 0;
	for ( node = firstChild; node; node = node->NextSibling() )
	{
		clone->LinkEndChild( node->Clone() );
	}
	return clone;
}


void CL_XmlDocument::Print( FILE* cfile, int depth ) const
{
	CL_XmlNode* node;
	for ( node=FirstChild(); node; node=node->NextSibling() )
	{
		node->Print( cfile, depth );
		fprintf( cfile, "\n" );
	}
}

void CL_XmlDocument::StreamOut( CL_XML_OSTREAM * out ) const
{
	CL_XmlNode* node;
	for ( node=FirstChild(); node; node=node->NextSibling() )
	{
		node->StreamOut( out );

		// Special rule for streams: stop after the root element.
		// The stream in code will only read one element, so don't
		// write more than one.
		if ( node->ToElement() )
			break;
	}
}

CL_XmlAttribute* CL_XmlAttribute::Next() const
{
	// We are using knowledge of the sentinel. The sentinel
	// have a value or name.
	if ( next->value.empty() && next->name.empty() )
		return 0;
	return next;
}


CL_XmlAttribute* CL_XmlAttribute::Previous() const
{
	// We are using knowledge of the sentinel. The sentinel
	// have a value or name.
	if ( prev->value.empty() && prev->name.empty() )
		return 0;
	return prev;
}


void CL_XmlAttribute::Print( FILE* cfile, int /*depth*/ ) const
{
	CL_XML_STRING n, v;

	PutString( Name(), &n );
	PutString( Value(), &v );

	if (value.find ('\"') == CL_XML_STRING::npos)
		fprintf (cfile, "%s=\"%s\"", n.c_str(), v.c_str() );
	else
		fprintf (cfile, "%s='%s'", n.c_str(), v.c_str() );
}


void CL_XmlAttribute::StreamOut( CL_XML_OSTREAM * stream ) const
{
	if (value.find( '\"' ) != CL_XML_STRING::npos)
	{
		PutString( name, stream );
		(*stream) << "=" << "'";
		PutString( value, stream );
		(*stream) << "'";
	}
	else
	{
		PutString( name, stream );
		(*stream) << "=" << "\"";
		PutString( value, stream );
		(*stream) << "\"";
	}
}

void CL_XmlAttribute::SetIntValue( int value )
{
	char buf [64];
	sprintf (buf, "%d", value);
	SetValue (buf);
}

void CL_XmlAttribute::SetDoubleValue( double value )
{
	char buf [64];
	sprintf (buf, "%lf", value);
	SetValue (buf);
}

const int CL_XmlAttribute::IntValue() const
{
	return atoi (value.c_str ());
}

const double  CL_XmlAttribute::DoubleValue() const
{
	return atof (value.c_str ());
}

void CL_XmlComment::Print( FILE* cfile, int depth ) const
{
	for ( int i=0; i<depth; i++ )
	{
		fputs( "    ", cfile );
	}
	fprintf( cfile, "<!--%s-->", value.c_str() );
}

void CL_XmlComment::StreamOut( CL_XML_OSTREAM * stream ) const
{
	(*stream) << "<!--";
	PutString( value, stream );
	(*stream) << "-->";
}

CL_XmlNode* CL_XmlComment::Clone() const
{
	CL_XmlComment* clone = new CL_XmlComment();

	if ( !clone )
		return 0;

	CopyToClone( clone );
	return clone;
}


void CL_XmlText::Print( FILE* cfile, int /*depth*/ ) const
{
	CL_XML_STRING buffer;
	PutString( value, &buffer );
	fprintf( cfile, "%s", buffer.c_str() );
}


void CL_XmlText::StreamOut( CL_XML_OSTREAM * stream ) const
{
	PutString( value, stream );
}


CL_XmlNode* CL_XmlText::Clone() const
{	
	CL_XmlText* clone = 0;
	clone = new CL_XmlText( "" );

	if ( !clone )
		return 0;

	CopyToClone( clone );
	return clone;
}


CL_XmlDeclaration::CL_XmlDeclaration( const char * _version,
	const char * _encoding,
	const char * _standalone )
: CL_XmlNode( CL_XmlNode::DECLARATION )
{
	version = _version;
	encoding = _encoding;
	standalone = _standalone;
}


void CL_XmlDeclaration::Print( FILE* cfile, int /*depth*/ ) const
{
	fprintf (cfile, "<?xml ");

	if ( !version.empty() )
		fprintf (cfile, "version=\"%s\" ", version.c_str ());
	if ( !encoding.empty() )
		fprintf (cfile, "encoding=\"%s\" ", encoding.c_str ());
	if ( !standalone.empty() )
		fprintf (cfile, "standalone=\"%s\" ", standalone.c_str ());
	fprintf (cfile, "?>");
}

void CL_XmlDeclaration::StreamOut( CL_XML_OSTREAM * stream ) const
{
	(*stream) << "<?xml ";

	if ( !version.empty() )
	{
		(*stream) << "version=\"";
		PutString( version, stream );
		(*stream) << "\" ";
	}
	if ( !encoding.empty() )
	{
		(*stream) << "encoding=\"";
		PutString( encoding, stream );
		(*stream ) << "\" ";
	}
	if ( !standalone.empty() )
	{
		(*stream) << "standalone=\"";
		PutString( standalone, stream );
		(*stream) << "\" ";
	}
	(*stream) << "?>";
}

CL_XmlNode* CL_XmlDeclaration::Clone() const
{	
	CL_XmlDeclaration* clone = new CL_XmlDeclaration();

	if ( !clone )
		return 0;

	CopyToClone( clone );
	clone->version = version;
	clone->encoding = encoding;
	clone->standalone = standalone;
	return clone;
}


void CL_XmlUnknown::Print( FILE* cfile, int depth ) const
{
	for ( int i=0; i<depth; i++ )
		fprintf( cfile, "    " );
	fprintf( cfile, "%s", value.c_str() );
}

void CL_XmlUnknown::StreamOut( CL_XML_OSTREAM * stream ) const
{
	(*stream) << "<" << value << ">";		// Don't use entities hear! It is unknown.
}

CL_XmlNode* CL_XmlUnknown::Clone() const
{
	CL_XmlUnknown* clone = new CL_XmlUnknown();

	if ( !clone )
		return 0;

	CopyToClone( clone );
	return clone;
}


CL_XmlAttributeSet::CL_XmlAttributeSet()
{
	sentinel.next = &sentinel;
	sentinel.prev = &sentinel;
}


CL_XmlAttributeSet::~CL_XmlAttributeSet()
{
	assert( sentinel.next == &sentinel );
	assert( sentinel.prev == &sentinel );
}


void CL_XmlAttributeSet::Add( CL_XmlAttribute* addMe )
{
	assert( !Find( addMe->Name() ) );	// Shouldn't be multiply adding to the set.

	addMe->next = &sentinel;
	addMe->prev = sentinel.prev;

	sentinel.prev->next = addMe;
	sentinel.prev      = addMe;
}

void CL_XmlAttributeSet::Remove( CL_XmlAttribute* removeMe )
{
	CL_XmlAttribute* node;

	for( node = sentinel.next; node != &sentinel; node = node->next )
	{
		if ( node == removeMe )
		{
			node->prev->next = node->next;
			node->next->prev = node->prev;
			node->next = 0;
			node->prev = 0;
			return;
		}
	}
	assert( 0 );		// we tried to remove a non-linked attribute.
}

CL_XmlAttribute*	CL_XmlAttributeSet::Find( const char * name ) const
{
	CL_XmlAttribute* node;

	for( node = sentinel.next; node != &sentinel; node = node->next )
	{
		if ( node->name == name )
			return node;
	}
	return 0;
}


#ifdef CL_XML_USE_STL	
CL_XML_ISTREAM & operator >> (CL_XML_ISTREAM & in, CL_XmlNode & base)
{
	CL_XML_STRING tag;
	tag.reserve( 8 * 1000 );
	base.StreamIn( &in, &tag );

	base.Parse( tag.c_str() );
	return in;
}
#endif


CL_XML_OSTREAM & operator<< (CL_XML_OSTREAM & out, const CL_XmlNode & base)
{
	base.StreamOut (& out);
	return out;
}