
#include "Display/display_precomp.h"
#include "display_window_dx7.h"
#include "back_buffer_dx7.h"
#include "graphic_context_dx7.h"
#include "API/Core/System/error.h"
#include <iostream>

/////////////////////////////////////////////////////////////////////////////
// Construction:

CL_DisplayWindow_DX7::CL_DisplayWindow_DX7() :
	directdraw(0), primary_surface(0), backbuffer_surface(0),
	backbuffer_width(0), backbuffer_height(0), video_memory(true),
	can_flip(false)
{
	HRESULT result;

	result = DirectDrawCreate(0, &directdraw, 0); 
	if (result != DD_OK) throw CL_Error("DirectDrawCreate failed!");

	memset(&hel_caps, 0, sizeof(hel_caps));
	hel_caps.dwSize = sizeof(DDCAPS);
	result = directdraw->GetCaps(0, &hel_caps);	
	if (result != DD_OK)
	{
		directdraw->Release();
		throw CL_Error("Get HEL Caps failed!");
	}

	memset(&hal_caps, 0, sizeof(hel_caps));
	hal_caps.dwSize = sizeof(DDCAPS);
	result = directdraw->GetCaps(&hal_caps, NULL);
	if (result != DD_OK)
	{
		directdraw->Release();
		throw CL_Error("Get HAL Caps failed!");
	}

	pixel_buffers[0] = 0;
	pixel_buffers[1] = 0;

	gc = CL_GraphicContext(new CL_GraphicContext_DX7(this));
}

CL_DisplayWindow_DX7::~CL_DisplayWindow_DX7()
{
	if (directdraw)
	{
		directdraw->FlipToGDISurface();
		directdraw->Release();
	}
}

/////////////////////////////////////////////////////////////////////////////
// Attributes:

CL_GraphicContext *CL_DisplayWindow_DX7::get_gc()
{
	return &gc;
}

CL_PixelBuffer &CL_DisplayWindow_DX7::get_buffer(int i)
{
	return *pixel_buffers[i];
}

const CL_PixelBuffer &CL_DisplayWindow_DX7::get_buffer(int i) const
{
	return *pixel_buffers[i];
}

int CL_DisplayWindow_DX7::get_buffer_count() const
{
	return 2;
}

/////////////////////////////////////////////////////////////////////////////
// Operations:

void CL_DisplayWindow_DX7::create_window(const CL_DisplayWindowDescription &desc)
{
	HRESULT result;

	CL_DisplayWindow_Win32::create_window(desc);

	// Connect directdraw object to window and setup flipping system:
	try
	{
		video_memory = true;
		int flags = DDSCL_NORMAL;

		if (fullscreen)
		{
			flags = 
				DDSCL_EXCLUSIVE |
				DDSCL_FULLSCREEN |
				DDSCL_ALLOWREBOOT|
				DDSCL_ALLOWMODEX;
		}

		result = directdraw->SetCooperativeLevel(
			get_hwnd(),
			flags);

		if (result != DD_OK) throw CL_Error("DirectDraw SetCooperativeLevel failed!");

		// If fullscreen, do a mode switch:
		if (fullscreen)
		{
			result = directdraw->SetDisplayMode(desc.get_size().width, desc.get_size().height, 32/*desc.get_bpp()*/);
			if (result != DD_OK) throw CL_Error("Set fullscreen failed!");

			create_fullscreen_targets();
		}
		else
		{
			create_windowed_targets();
		}

		pixel_buffers[0] = new CL_PixelBuffer(new CL_BackBuffer_DX7(this));
		pixel_buffers[1] = new CL_PixelBuffer(new CL_BackBuffer_DX7(this));
	}
	catch (...)
	{
		destroy_window();
		throw;
	}
}

void CL_DisplayWindow_DX7::destroy_window()
{
	delete pixel_buffers[0];
	delete pixel_buffers[1];

	if (primary_surface) primary_surface->Release();
	if (backbuffer_surface && can_flip == false) backbuffer_surface->Release();

	primary_surface = 0;
	backbuffer_surface = 0;

	if (is_fullscreen()) directdraw->RestoreDisplayMode();

	CL_DisplayWindow_Win32::destroy_window();
}

void CL_DisplayWindow_DX7::flip()
{
	HRESULT result;

	// Surfaces are lost if the application is minimized, or if
	// some other ddraw application runs fullscreen.
	if (primary_surface->IsLost())
	{
		result = primary_surface->Restore();
		if (result != DD_OK) return;
	}

	if (backbuffer_surface->IsLost())
	{
		result = backbuffer_surface->Restore();
		if (result != DD_OK) return;
	}

	// Draw mouse cursor. :)
	CL_DisplayWindow_Win32::flip();

	if (can_flip)
	{
		DWORD flags = DDFLIP_WAIT;
		primary_surface->Flip(0, flags);
	}
	else
	{
		// Use surface Blt() because it can do some clipping for us,
		// and because manual copy is only faster if backbuffer is in sysmem.
		// (which isnn't nessesary the case here. it may just be a windows app)

		DDBLTFX ddbltfx;
		ddbltfx.dwSize = sizeof(ddbltfx);

		DWORD flags = DDBLT_WAIT;

		// The titlebar is part of the window.
		// That is why we start juggling with client rects and other stuff...
		RECT client_rect;
		RECT dest_rect;
		POINT Pt;

		GetClientRect(get_hwnd(), &client_rect);
		GetWindowRect(get_hwnd(), &dest_rect);

		Pt.x=0;
		Pt.y=0;
		ClientToScreen(get_hwnd(),&Pt);

		int width = dest_rect.right-dest_rect.left;
		int height = dest_rect.bottom-dest_rect.top;

		dest_rect.left = Pt.x;
		dest_rect.top = Pt.y;
		dest_rect.right = dest_rect.left + width;
		dest_rect.bottom = dest_rect.left + height;

		result = primary_surface->Blt(
			&dest_rect,
			backbuffer_surface,
			0,
			flags,
			&ddbltfx);
	}

	CL_DisplayWindow_Win32::flip();
}

/////////////////////////////////////////////////////////////////////////////
// Implementation:

void CL_DisplayWindow_DX7::create_fullscreen_targets()
{
	try // to create a surface flipping system with two surfaces.
	{
		// setup the primary surface (the flipping system):
		HRESULT err;

		DDSURFACEDESC surf_desc;
		memset(&surf_desc, 0, sizeof(surf_desc));
		surf_desc.dwSize = sizeof(DDSURFACEDESC);

		surf_desc.dwFlags =
			DDSD_CAPS |
			DDSD_BACKBUFFERCOUNT;

		surf_desc.ddsCaps.dwCaps =
			DDSCAPS_PRIMARYSURFACE |
			DDSCAPS_FLIP |
			DDSCAPS_COMPLEX;

		if (video_memory) surf_desc.ddsCaps.dwCaps |= DDSCAPS_VIDEOMEMORY;
		else surf_desc.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;

		// 3 buffers are more efficient, but at the price of memory.
		// We should make it possible for the application to specify the
		// buffer count.
		surf_desc.dwBackBufferCount = 2;

		err = directdraw->CreateSurface(&surf_desc, &primary_surface, NULL);
		if (err != DD_OK) throw CL_Error("Couldn't create flipping system.");

		// Get the backbuffer surface from the primary/front:
		DDSCAPS backbuffer_desc;
		backbuffer_desc.dwCaps = DDSCAPS_BACKBUFFER;

		err = primary_surface->GetAttachedSurface(
			&backbuffer_desc,
			&backbuffer_surface);

		if (err != DD_OK)
		{
			primary_surface->Release();
			primary_surface = 0;
			throw CL_Error("Couldn't get backbuffer in flipping system.");
		}

		can_flip = true;
	}
	catch (CL_Error err) // not enough memory, fallback to system memory backbuffer.
	{
		create_windowed_targets();
	}
}

void CL_DisplayWindow_DX7::create_windowed_targets()
{
	// setup the front buffer surface:
	HRESULT err;

	DDSURFACEDESC surf_desc;
	memset(&surf_desc, 0, sizeof(surf_desc));
	surf_desc.dwSize = sizeof(DDSURFACEDESC);
	surf_desc.dwFlags = DDSD_CAPS;
	surf_desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

	err = directdraw->CreateSurface(&surf_desc, &primary_surface, NULL);
	if (err != DD_OK) throw CL_Error("Couldn't create front buffer.");

	try
	{
		create_windowed_backbuffer();
	}
	catch (...)
	{
		primary_surface->Release();
		primary_surface = 0;
		throw;
	}

	// Clip blits to the window borders:
	LPDIRECTDRAWCLIPPER clipper;
	err = directdraw->CreateClipper(0, &clipper, NULL);
	cl_assert(err == DD_OK);
	clipper->SetHWnd(0, get_hwnd());

	// attach clipper to front buffer:
	primary_surface->SetClipper(clipper);
	clipper->Release();

	can_flip = false;
}

void CL_DisplayWindow_DX7::create_windowed_backbuffer()
{
	HRESULT err = 0;
	RECT client_rect;
	RECT dest_rect;
	POINT Pt;

	GetClientRect(get_hwnd(), &client_rect);
	GetWindowRect(get_hwnd(), &dest_rect);

	Pt.x=0;
	Pt.y=0;
	ClientToScreen(get_hwnd(),&Pt);

	int width = dest_rect.right-dest_rect.left;
	int height = dest_rect.bottom-dest_rect.top;

	if (width == 0 || height == 0) return; // minimized.

	// clean up previous backbuffer (if any):
	if (backbuffer_surface) backbuffer_surface->Release();
	backbuffer_surface = 0;

	// create back buffer surface: (trying in framebuffer first)
	DDSURFACEDESC surf_desc;
	memset(&surf_desc, 0, sizeof(surf_desc));
	surf_desc.dwSize = sizeof(DDSURFACEDESC);
	surf_desc.dwFlags=DDSD_CAPS|DDSD_HEIGHT|DDSD_WIDTH;
	surf_desc.dwWidth = width;
	surf_desc.dwHeight = height;
	surf_desc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	if (video_memory) surf_desc.ddsCaps.dwCaps |= DDSCAPS_VIDEOMEMORY;
	else surf_desc.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;

	err = directdraw->CreateSurface(&surf_desc, &backbuffer_surface, NULL);

	if (err != DD_OK) // damn. Not enough memory in video, fallback to system memory.
	{
		surf_desc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN|DDSCAPS_SYSTEMMEMORY;

		err = directdraw->CreateSurface(&surf_desc, &backbuffer_surface, NULL);
		if (err == DD_OK) throw CL_Error("Could not create backbuffer!");
	}

	backbuffer_width = width;
	backbuffer_height = height;
}
