#include "Display/display_precomp.h"
#include "API/Display/font.h"
#include "API/Display/display.h"
#include "API/Display/display_window.h"
#include "API/Display/graphic_context.h"
#include "API/Display/glyph_buffer.h"
#include "API/Display/color.h"
#include "font_generic.h"
#include "resourcedata_font.h"
#include "API/Core/Resources/resource_manager.h"
#include "API/Core/System/error.h"
#include "API/Core/System/clanstring.h"
#include <algorithm>
#include <cmath>
#include <memory>

#ifdef WIN32
#include "Win32/font_win32.h"
#define CL_Font_Platform CL_Font_Win32
#else
#include "X11/font_x11.h"
#define CL_Font_Platform CL_Font_X11
#endif

using namespace std;

/////////////////////////////////////////////////////////////////////////////
// CL_Font construction:

CL_Font::CL_Font()
: impl(0)
{}

CL_Font::CL_Font(const CL_Font &copy)
: impl(new CL_Font_Platform)
{
	if (copy.impl)
		*impl = *(copy.impl);
	
	resource = copy.resource;
	resource.load();
}

CL_Font::CL_Font(
	const string &resource_id,
	CL_ResourceManager *manager)
: impl(new CL_Font_Platform)
{
	CL_Resource &resource = manager->get_resource(resource_id);
	resource.load();
	
	CL_ResourceData_Font *data = (CL_ResourceData_Font *) resource.get_data("font");
	
	operator=(data->get_font());
}

CL_Font::CL_Font(
	const CL_Sprite &letters,
	const string &letter_chars,
	int space_width,
	int width_offset,
	int height_offset,
	int baseline)
: impl(new CL_Font_Platform)
{
	impl->create_font(
		letters,
		letter_chars,
		space_width,
		width_offset,
		height_offset,
		baseline);
}

CL_Font::CL_Font(
	const string &font_name,
	int height,
	int width,
	bool bold,
	bool italic,
	bool underline,
	bool strikeout)
: impl(new CL_Font_Platform)
{
	//First thirty two characters are undrawable in ASCII
	string letters;
	letters.reserve(256-32);
	for (int x = 32; x < 256; ++x)
		letters.append(1, (unsigned char)(x));

	impl->create_sysfont(
		font_name,
		letters,
		height,
		width,
		bold,
		italic,
		underline,
		strikeout);
}

CL_Font::CL_Font(
	const string &font_name,
	const string &letter_chars,
	int height,
	int width,
	bool bold,
	bool italic,
	bool underline,
	bool strikeout)
: impl(new CL_Font_Platform)
{
	impl->create_sysfont(
		font_name,
		letter_chars,
		height,
		width,
		bold,
		italic,
		underline,
		strikeout);
}

CL_Font::~CL_Font()
{
	resource.unload();
	delete impl; //Safe even if impl is 0
}

/////////////////////////////////////////////////////////////////////////////
// CL_Font attributes:

int CL_Font::get_width_offset() const
{
	return impl->width_offset;
}

int CL_Font::get_height_offset() const
{
	return impl->height_offset;
}

int CL_Font::get_baseline() const
{
	return impl->baseline;
}

void CL_Font::get_scale(float &x, float &y) const
{
	impl->glyphs.get_scale(x, y);
}

float CL_Font::get_alpha() const
{
	return impl->glyphs.get_alpha();
}

void CL_Font::get_color(float &red, float &green, float &blue, float &alpha) const
{
	impl->glyphs.get_color(red, green, blue, alpha);
}

void CL_Font::get_blend_func(CL_BlendFunc &src, CL_BlendFunc &dest) const
{
	impl->glyphs.get_blend_func(src, dest);
}

float CL_Font::get_glyph_angle() const
{
	return impl->glyphs.get_angle();
}

void CL_Font::get_glyph_rotation_hotspot(CL_Origin &origin, int &x, int &y) const
{
	impl->glyphs.get_rotation_hotspot(origin, x, y);
}

float CL_Font::get_angle() const
{
	return impl->angle;
}

void CL_Font::get_rotation_hotspot(CL_Origin &origin, int &x, int &y) const
{
	origin = impl->rot_origin;
	x = impl->rot_x;
	y = impl->rot_y;
}

void CL_Font::get_alignment(CL_Origin &origin, int &x, int &y) const
{
	origin = impl->trans_origin;
	x = impl->trans_x;
	y = impl->trans_y;
}

int CL_Font::get_height() const
{
	float scale_x, scale_y;
	impl->glyphs.get_scale(scale_x, scale_y);
	
	return (int)((impl->height + impl->height_offset) * scale_y);
}
	
int CL_Font::get_height(
	const string &str,
	const CL_Size &max_size,
	const string &delim) const
{
	return impl->get_size(str.begin(), str.end(), max_size.width, max_size.height, delim).height;
}
	
int CL_Font::get_height(
	const string::const_iterator &start,
	const string::const_iterator &end,
	const CL_Size &max_size,
	const string &delim) const
{
	return impl->get_size(start, end, max_size.width, max_size.height, delim).height;
}

int CL_Font::get_width(unsigned char letter) const
{
	float scale_x, scale_y;
	impl->glyphs.get_scale(scale_x, scale_y);
	
	const short glyph = impl->ctg(letter);

	if (glyph != -1)
		return (int)((impl->gwidth(glyph) + impl->width_offset ) * scale_x);
	else if (letter == '\n')
		return 0;
	else
		return (int)((impl->spacelen + impl->width_offset ) * scale_x);
}

int CL_Font::get_width(
	const string &str,
	const CL_Size &max_size,
	const string &delim) const
{
	return impl->get_size(str.begin(), str.end(), max_size.width, max_size.height, delim).width;
}
	
int CL_Font::get_width(
	const string::const_iterator &start,
	const string::const_iterator &end,
	const CL_Size &max_size,
	const string &delim) const
{
	return impl->get_size(start, end, max_size.width, max_size.height, delim).width;
}

CL_Size CL_Font::get_size(unsigned char letter) const
{
	return CL_Size(get_width(letter), get_height());
}

CL_Size CL_Font::get_size(
	const string &str,
	const CL_Size &max_size,
	const string &delim) const
{
	return impl->get_size(str.begin(), str.end(), max_size.width, max_size.height, delim);
}

CL_Size CL_Font::get_size(
	const string::const_iterator &start,
	const string::const_iterator &end,
	const CL_Size &max_size,
	const string &delim) const
{
	return impl->get_size(start, end, max_size.width, max_size.height, delim);
}

const CL_Sprite &CL_Font::get_glyphs() const
{
	return impl->glyphs;
}

int CL_Font::get_glyph_index(unsigned char letter) const
{
	return impl->ctg(letter);
}

CL_Font::PrintInfo CL_Font::test_draw(
	int x,
	int y,
	const string &str) const
{
	return test_draw(CL_Rect(x, y, x, y), str.begin(), str.end(), string());
}

CL_Font::PrintInfo CL_Font::test_draw(
	const CL_Rect &dest,
	const string &str,
	const string &delim) const
{
	return test_draw(dest, str.begin(), str.end(), delim);
}

CL_Font::PrintInfo CL_Font::test_draw(
	int x,
	int y,
	const string::const_iterator &start,
	const string::const_iterator &end) const
{
	return test_draw(CL_Rect(x, y, x, y), start, end, string());
}

CL_Font::PrintInfo CL_Font::test_draw(
	const CL_Rect &dest,
	const string::const_iterator &start,
	const string::const_iterator &end,
	const string &delim) const
{
	CL_Font_Generic::NullCharDest null_dest;
	return PrintInfo(impl->calc_char_positions(start, end, null_dest, dest, delim));
}

/////////////////////////////////////////////////////////////////////////////
// CL_Font operations:

CL_Font &CL_Font::operator =(const CL_Font &copy)
{
	resource.unload();
	
	if (copy.impl) {
		if (impl == 0)
			impl = new CL_Font_Platform(*(CL_Font_Platform*)(copy.impl)); //This should be safe
		else
			*impl = *copy.impl;
	}
	
	resource = copy.resource;
	resource.load();

	return *this;
}

CL_Font::PrintInfo CL_Font::draw(
	int x,
	int y,
	const string &str,
	CL_GraphicContext *context)
{
	return draw(CL_Rect(x, y, x, y), str.begin(), str.end(), string(), context);
}

CL_Font::PrintInfo CL_Font::draw(
	const CL_Rect &dest,
	const string &str,
	const string &delim,
	CL_GraphicContext *context)
{
	return draw(dest, str.begin(), str.end(), delim, context);
}

CL_Font::PrintInfo CL_Font::draw(
	int x,
	int y,
	const string::const_iterator &start,
	const string::const_iterator &end,
	CL_GraphicContext *context)
{
	return draw(CL_Rect(x, y, x, y), start, end, string(), context);
}

CL_Font::PrintInfo CL_Font::draw(
	const CL_Rect &dest,
	const string::const_iterator &start,
	const string::const_iterator &end,
	const string &delim,
	CL_GraphicContext *context)
{
	if (context == 0)
		context = CL_Display::get_current_window()->get_gc();
	
	CL_Font_Generic::DrawCharDest char_dest(context, impl);
	return PrintInfo(impl->calc_char_positions(start, end, char_dest, dest, delim));
}

CL_Font::PrintInfo CL_Font::draw_to_gb(
	int x,
	int y,
	const string &str,
	CL_GlyphBuffer &gb) const
{
	return draw_to_gb(CL_Rect(x, y, x, y), str.begin(), str.end(), gb, string());
}

CL_Font::PrintInfo CL_Font::draw_to_gb(
	const CL_Rect &dest,
	const string &str,
	CL_GlyphBuffer &gb,
	const string &delim) const
{
	return draw_to_gb(dest, str.begin(), str.end(), gb, delim);
}

CL_Font::PrintInfo CL_Font::draw_to_gb(
	int x,
	int y,
	const string::const_iterator &start,
	const string::const_iterator &end,
	CL_GlyphBuffer &gb) const
{
	return draw_to_gb(CL_Rect(x, y, x, y), start, end, gb, string());
}

CL_Font::PrintInfo CL_Font::draw_to_gb(
	const CL_Rect &dest,
	const string::const_iterator &start,
	const string::const_iterator &end,
	CL_GlyphBuffer &gb,
	const string &delim) const
{
	gb.change_font(*this);
	
	//If no word wrapping or height truncating, we know exactly how many characters we're about to add
	if ((dest.get_width() == 0) && (dest.get_height() == 0))
		gb.get_glyphs().reserve(gb.get_glyphs().size() + (end - start));

	CL_Font_Generic::GlyphBufferDest char_dest(&gb, impl);
	
	return PrintInfo(impl->calc_char_positions(start, end, char_dest, dest, delim));
}

void CL_Font::set_width_offset(int offset)
{
	impl->width_offset = offset;
}

void CL_Font::set_height_offset(int offset)
{
	impl->height_offset = offset;
}

void CL_Font::set_baseline(int baseline)
{
	impl->baseline = baseline;
}

void CL_Font::set_scale(float x, float y)
{
	impl->glyphs.set_scale(x, y);
}

void CL_Font::set_alpha(float alpha)
{
	impl->glyphs.set_alpha(alpha);
}

void CL_Font::set_color(float r, float g, float b, float a)
{
	impl->glyphs.set_color(r, g, b, a);
}

void CL_Font::set_blend_func(CL_BlendFunc src, CL_BlendFunc dest)
{
	impl->glyphs.set_blend_func(src, dest);
}

void CL_Font::set_glyph_angle(float angle)
{
	impl->glyphs.set_angle(angle);
}

void CL_Font::set_glyph_rotation_hotspot(CL_Origin origin, int x, int y)
{
	impl->glyphs.set_rotation_hotspot(origin, x, y);
}

void CL_Font::glyph_rotate(float angle)
{
	impl->glyphs.rotate(angle);
}

void CL_Font::set_angle(float angle)
{
	while (angle >= 360.0)
		angle -= 360.0;
	while (angle < 0.0)
		angle += 360.0;
	
	impl->angle = angle;
}

void CL_Font::rotate(float angle)
{
	angle += impl->angle;
	set_angle(angle);
}

void CL_Font::set_rotation_hotspot(CL_Origin origin, int x, int y)
{
	impl->rot_origin = origin;
	impl->rot_x = x;
	impl->rot_y = y;
}

void CL_Font::set_alignment(CL_Origin origin, int x, int y)
{
	impl->trans_origin = origin;
	impl->trans_x = x;
	impl->trans_y = y;
}
