#include "Display/display_precomp.h"
#include "font_generic.h"
#include "API/Core/Math/size.h"
#include "API/Core/Math/rect.h"
#include "API/Core/System/error.h"
#include "API/Core/System/clanstring.h"
#include "API/Display/sprite.h"
#include "API/Display/glyph_buffer.h"
#include <cmath>
#include <algorithm>
#include <vector>

using namespace std;

/////////////////////////////////////////////////////////////////////////////
// CL_Font_Generic::CanonCharDest:

void CL_Font_Generic::CanonCharDest::operator()(
	const CL_Point &point,
	unsigned char character,
	short glyph)
{
	if (ang < 0.05)
		drawChar(point, character, glyph);
	else
		drawChar(point.rotate(hotspot, ang), character, glyph);
}

/////////////////////////////////////////////////////////////////////////////
// CL_Font_Generic::DrawCharDest:

void CL_Font_Generic::DrawCharDest::drawChar(
	const CL_Point& point,
	unsigned char character,
	short glyph)
{
	if (glyph != -1) {
		fg->glyphs.set_frame(glyph);
		fg->glyphs.draw(point.x, point.y, gc);
	}
}

/////////////////////////////////////////////////////////////////////////////
// CL_Font_Generic::GlyphBufferDest:

void CL_Font_Generic::GlyphBufferDest::drawChar(
	const CL_Point& point,
	unsigned char character,
	short glyph)
{
	gb->get_glyphs().push_back(CL_GlyphBuffer::Glyph(point, character));
}

/////////////////////////////////////////////////////////////////////////////
// CL_Font_Generic construction:

CL_Font_Generic::CL_Font_Generic()
	: height(0),
	baseline(0),
	width_offset(0),
	height_offset(0),
	spacelen(0),
	angle(0.0),
	rot_origin(origin_center),
	rot_x(0),
	rot_y(0),
	trans_origin(origin_top_left),
	trans_x(0),
	trans_y(0)
{
	set_letters("");
}

CL_Font_Generic::CL_Font_Generic(const CL_Font_Generic &copy)
	: height(copy.height),
	baseline(copy.baseline),
	width_offset(copy.width_offset),
	height_offset(copy.height_offset),
	spacelen(copy.spacelen),
	angle(copy.angle),
	rot_origin(copy.rot_origin),
	rot_x(copy.rot_x),
	rot_y(copy.rot_y),
	trans_origin(copy.trans_origin),
	trans_x(copy.trans_x),
	trans_y(copy.trans_y),
	glyphs(copy.glyphs)
{
	cur_ctg = copy.cur_ctg;
	++(cur_ctg->second.ref_count);
}

CL_Font_Generic::~CL_Font_Generic()
{
	decrement();
}

/////////////////////////////////////////////////////////////////////////////
// CL_Font_Generic attributes:

//Static char-to-glyph-map cache
CL_Font_Generic::CTGVector_Cache CL_Font_Generic::ctg_vector_cache;

CL_Size CL_Font_Generic::get_size(
	const string::const_iterator &start,
	const string::const_iterator &end,
	int max_width,
	int max_height,
	const string &delim,
	float scale_x,
	int offset_x,
	int space_width,
	int line_height)
{
	CL_Size retVal(0,0);
	
	//Width of the current line
	int current_line = 0;
	
	//End of the current line
	string::const_iterator lend;
	
	for(
	string::const_iterator curchar = start;
	((max_height < 1) || (retVal.height + line_height <= max_height)) && (curchar != end);
	curchar = lend)
	{
		lend = wrap_pos(curchar, end, max_width, delim, scale_x, offset_x, space_width);
		
		//For each character in the line
		for (; curchar != lend; ++curchar)
		{
			const unsigned char c = *curchar;
			const short glyph = ctg(c);
			if (glyph != -1)
				current_line += (int)((gwidth(glyph) + offset_x) * scale_x);
			else if (c != '\n')
				current_line += space_width;
		}
		
		if (current_line > retVal.width)
			retVal.width = current_line;
		
		retVal.height += line_height;
		current_line = 0;
	}
	
	return retVal;
}
	
CL_Size CL_Font_Generic::get_size(
	const string::const_iterator &start,
	const string::const_iterator &end,
	int max_width,
	int max_height,
	const string &delim)
{
	float scale_x, scale_y;
	glyphs.get_scale(scale_x, scale_y);
	
	const int offset_x = width_offset;
	const int line_height = (int)((height + height_offset) * scale_y);
	const int space_width = (int)((spacelen + offset_x) * scale_x);
	
	return get_size(start, end, max_width, max_height, delim, scale_x, offset_x, space_width, line_height);
}

/////////////////////////////////////////////////////////////////////////////
// CL_Font_Generic operations:

void CL_Font_Generic::decrement()
{
	//Do not erase the blank cache entry, since it's used by default CL_Fonts
	if ((--(cur_ctg->second.ref_count) <= 0) && (cur_ctg->first != ""))
		ctg_vector_cache.cache.erase(cur_ctg);
}

CL_Font_Generic& CL_Font_Generic::operator =(const CL_Font_Generic &copy)
{
	decrement();
	cur_ctg = copy.cur_ctg;
	++(cur_ctg->second.ref_count);
	
	height = copy.height;
	baseline = copy.baseline;
	width_offset = copy.width_offset;
	height_offset = copy.height_offset;
	spacelen = copy.spacelen;
	angle = copy.angle;
	rot_origin = copy.rot_origin;
	rot_x = copy.rot_x;
	rot_y = copy.rot_y;
	trans_origin = copy.trans_origin;
	trans_x = copy.trans_x;
	trans_y = copy.trans_y;
	glyphs = copy.glyphs;
	
	return *this;
}

void CL_Font_Generic::set_letters(const string &letters)
{
	if ((cur_ctg = ctg_vector_cache.cache.find(letters)) == ctg_vector_cache.cache.end())
	{
		cur_ctg = ctg_vector_cache.cache.insert(pair<string, CTGVector>(letters, CTGVector())).first;
		
		string::const_iterator it = letters.begin();
		for (short i=0; i < letters.length(); ++i)
		{
			cur_ctg->second.char_to_glyph_map[*it] = i;
			++it;
		}
	}
	else
		++(cur_ctg->second.ref_count);
}

void CL_Font_Generic::create_font(
	const CL_Sprite &new_glyphs,
	const string &letters,
	int new_spacelen,
	int new_width_offset,
	int new_height_offset,
	int new_baseline)
{
	//Size of char limit on most platforms is 256
	const int length = min((int)(letters.size()), 256);
	
	if ((length > new_glyphs.get_frame_count()) || (length == 0))
	{
		throw CL_Error(CL_String::format(
			"Font error: Letter characters: %1, Available font glyphs: %2", 
			length,
			glyphs.get_frame_count()
		));
	}
	
	if (new_spacelen < 0)
	{
		string::size_type space_pos = letters.find(' ');
		
		if (space_pos != string::npos) {
			//If there is a character for space, then use it
			new_spacelen = new_glyphs.get_frame_size((int)space_pos).width;
		} else {
			//Make the space size the average of all character sizes
			new_spacelen = 0;
			
			for (int pos = 0; pos < length; ++pos) {
				new_spacelen += new_glyphs.get_frame_size((int)(pos)).width;
			}
			
			new_spacelen /= length;
		}
	}
	
	for (short i=0; i < length; ++i)
	{
		int char_height = new_glyphs.get_frame_size(i).height;
		if (char_height > height)
			height = char_height;
	}
	
	decrement();
	set_letters(letters);
	glyphs = new_glyphs;
	glyphs.set_alignment(origin_top_left, 0, 0);
	glyphs.set_rotation_hotspot(origin_top_left, 0, 0); //Important for rotation
	width_offset = new_width_offset;
	height_offset = new_height_offset;
	baseline = new_baseline;
	spacelen = new_spacelen;
}

CL_Font::PrintInfo CL_Font_Generic::calc_char_positions(
	const string::const_iterator &start,
	const string::const_iterator &end,
	CharDest &func,
	const CL_Rect &dest,
	const string &delim)
{	
	///////
	// Determine the rotation hotspot and drawing position, and other misc values
	///////
	
	float scale_x, scale_y;
	glyphs.get_scale(scale_x, scale_y);
	
	const int offset_x = width_offset;
	const int line_height = (int)((height + height_offset) * scale_y); //Distance to move at carriage feed
	const int space_width = (int)((spacelen + offset_x) * scale_x); //Distance to move at unrecognized character
	
	//The position to draw to, which is relative to justification.
	//That is, if justif is justif_right, this is the right side, and so on.
	CL_Point initial_pos(dest.left, dest.top);

	//Adjust for the size of the dest rectangle
	//We adjust for the size of the image rectangle (the rectangle containing the text) later
	//Actually, we try to avoid having to calculate that rectangle at all, if we can
	initial_pos += calc_origin(trans_origin, CL_Size(dest.get_width(), dest.get_height()));
	
	//The total size of the text we're about to pipe through the destination
	//Try to get the initial_pos/rot_hotspot without calculating this in advance
	CL_Size img_rect_size(0, 0);
	
	//The justification mode (determines which side of dest we hang to, and indiv. line positions)
	Justif justif;
	
	//The rotation hotspot (will be changed later if necesssary)
	CL_Point rot_hotspot(dest.left, dest.top);
	
	//Determine the justification based on alignment
	if (
	(trans_origin == origin_top_left) ||
	(trans_origin == origin_center_left) ||
	(trans_origin == origin_bottom_left))
		justif = justif_left;
	else if(
	(trans_origin == origin_top_center) ||
	(trans_origin == origin_center) ||
	(trans_origin == origin_bottom_center))
		justif = justif_center;
	else if(
	(trans_origin == origin_top_right) ||
	(trans_origin == origin_center_right) ||
	(trans_origin == origin_bottom_right))
		justif = justif_right;
	
	//Since x position of initial_pos is relative to justif...
	//We only have to adjust initial_pos by the height of the image rectangle, not the width
	if (
	(trans_origin == origin_center_left) ||
	(trans_origin == origin_center) ||
	(trans_origin == origin_center_right))
	{
		img_rect_size = get_size(
			start,
			end,
			dest.get_width(),
			dest.get_height(),
			delim,
			scale_x,
			offset_x,
			space_width,
			line_height);
		initial_pos.y -= img_rect_size.height/2;
	}
	else if(
	(trans_origin == origin_bottom_left) ||
	(trans_origin == origin_bottom_center) ||
	(trans_origin == origin_bottom_right))
	{
		img_rect_size = get_size(
			start,
			end,
			dest.get_width(),
			dest.get_height(),
			delim,
			scale_x,
			offset_x,
			space_width,
			line_height);
		initial_pos.y -= img_rect_size.height;
	}
	
	//If we're rotating, we need to calculate the position of the hotspot and send it to the destination
	if (angle > 0.05)
	{
		//The rotation hotspot is the upper left corner of the dest rectangle, by default
		//If both origins are the same, we dont need to calculate the size of the image rectangle
		//But if either one isnt, we need the size of the image rectangle to know how far to offset
		if (trans_origin != rot_origin)
		{
			//Only recalculate img_rect_size if we didn't calculate it already above
			if (img_rect_size.width == 0 && img_rect_size.height == 0)
				img_rect_size = get_size(
					start,
					end,
					dest.get_width(),
					dest.get_height(),
					delim,
					scale_x,
					offset_x,
					space_width,
					line_height);
			
			//Align the image and destination rectangle along trans_origin
			CL_Point trans_offset = calc_origin(
				trans_origin,
				CL_Size(dest.get_width() - img_rect_size.width, dest.get_height() - img_rect_size.height));
			CL_Rect image_rect(
				dest.left + trans_offset.x,
				dest.top + trans_offset.y,
				dest.left + img_rect_size.width + trans_offset.x,
				dest.top + img_rect_size.height + trans_offset.y);
			
			//Calculate the origin based on that and the rotation origin
			rot_hotspot = CL_Point(image_rect.left, image_rect.top);
			rot_hotspot += calc_origin(rot_origin, CL_Size(image_rect.get_width(), image_rect.get_height()));
		}
		else
		{
			rot_hotspot += calc_origin(rot_origin, CL_Size(dest.get_width(), dest.get_height()));
		}
		
		rot_hotspot.x += rot_x;
		rot_hotspot.y += rot_y;
	}
	
	initial_pos.x -= trans_x;
	initial_pos.y -= trans_y;	
	
	///////
	// Prepare for character piping
	///////
		
	//Ready the destination for (potentially) rotated characters
	func.set_rot(angle, rot_hotspot);
	
	CL_Font::PrintInfo retVal;
	retVal.bounds = CL_Rect(initial_pos.x, initial_pos.y, initial_pos.x, initial_pos.y);
	retVal.end = start;
	retVal.cursor = initial_pos;
	
	string::const_iterator lend; //End of the current line, exclusive
	CL_Point pos(initial_pos); //Drawing position, upper left of character
	
	///////
	// Pipe the character positions
	///////
	
	int current_line = 0; //Length of the current line
	int longest_line = 0; //Length of the longest line drawn so far
	
	//For each line (exclusive end of this line becomes inclusive start of next line)
	for (
	string::const_iterator curchar = start;
	((dest.get_height() < 1) || (pos.y + line_height <= initial_pos.y + dest.get_height())) && (curchar != end);
	curchar = lend)
	{
		//Find the character ending this line
		lend = wrap_pos(curchar, end, dest.get_width(), delim, scale_x, offset_x, space_width);
			
		//If the first character of a line was wider than dest.get_width(), stop the presses
		if (lend == curchar)
			break;
		
		if (justif == justif_left)
		{
			//Calculate current_line based on difference in pos
			current_line = pos.x;
			
			//For each character in the line
			for (; curchar != lend; ++curchar)
			{
				const unsigned char character = *curchar;
				const short glyph = ctg(character);
				
				func(pos, character, glyph);
				
				if (glyph != -1)
					pos.x += (int)((gwidth(glyph) + offset_x) * scale_x);
				else if (character != '\n') //Newlines wrap to the beginning of a new line
					pos.x += space_width;
			}
			
			//If the last character is a space, and space is a delimiter, and word wrapping is on, then remove it
			//This helps word wrapped text to sync properly
			if ((*(curchar-1) == ' ') && (dest.get_width() != 0) && (delim.find(' ') != string::npos))
				pos.x -= space_width;
			
			current_line = pos.x - current_line;
		}
		else if (justif == justif_center)
		{
			//Offset x, then use regular left-to-right drawing
			current_line = get_size(curchar, lend, 0, 0, string(), scale_x, offset_x, space_width, line_height).width;
			pos.x -= current_line/2;
			
			//If the last character is a space, and space is a delimiter, and word wrapping is on, then remove it
			//This helps word wrapped text to sync properly
			if ((*(lend-1) == ' ') && (dest.get_width() != 0) && (delim.find(' ') != string::npos))
			{
				pos.x += space_width/2;
				current_line -= space_width;
			}
			
			//For each character in the line
			for (; curchar != lend; ++curchar)
			{
				const unsigned char character = *curchar;
				const short glyph = ctg(character);
				
				func(pos, character, glyph);
				
				if (glyph != -1)
					pos.x += (int)((gwidth(glyph) + offset_x) * scale_x);
				else if (character != '\n') //Newlines wrap to the beginning of a new line
					pos.x += space_width;
			}
		}
		else if (justif == justif_right)
		{
			//Calculate current_line based on difference in pos
			current_line = pos.x;
			
			//If the last character is a space, and space is a delimiter, and word wrapping is on, then remove it
			//This helps word wrapped text to sync properly
			if ((*(lend-1) == ' ') && (dest.get_width() != 0) && (delim.find(' ') != string::npos))
			{
				pos.x += space_width;
			}
			
			//Draw backwards, right-to-left
			string::const_reverse_iterator rcurchar(curchar);
			string::const_reverse_iterator rit(lend);
			while (rit != rcurchar)
			{
				const unsigned char character = *rit;
				const short glyph = ctg(character);
				
				if (glyph != -1)
					pos.x -= (int)((gwidth(glyph) + offset_x) * scale_x);
				else if (character != '\n') //Newlines wrap to the beginning of a new line
					pos.x -= space_width;
				
				func(pos, character, glyph);
				
				++rit;
			}
			
			current_line -= pos.x;
			pos.x = initial_pos.x;
		}
		
		//Keep track of the longest line as we draw
		if (current_line > longest_line)
			longest_line = current_line;
		
		//Load the final cursor position into the return value
		retVal.cursor = pos;
		
		//Carraige return
		pos.y += line_height;
		pos.x = initial_pos.x;
	}
	
	///////
	// Finish calculating the return value
	///////
	
	retVal.end = lend;
		
	//We don't need to add line_height again here to get the bottom of the bottom line
	//It already did that at the end of the final line loop iteration
	retVal.bounds.top = initial_pos.y;
	retVal.bounds.bottom = pos.y;
	
	//Determine the sides of the bounding rect based on the justification mode
	if (justif == justif_left)
	{
		retVal.bounds.left = initial_pos.x;
		retVal.bounds.right = initial_pos.x + longest_line;
	}
	else if (justif == justif_center)
	{
		retVal.bounds.left = initial_pos.x - longest_line/2;
		retVal.bounds.right = initial_pos.x + longest_line/2;
	}
	else if (justif == justif_right)
	{
		retVal.bounds.left = initial_pos.x - longest_line;
		retVal.bounds.right = initial_pos.x;
	}
	
	return retVal;
}

string::const_iterator CL_Font_Generic::wrap_pos(
	const string::const_iterator &start,
	const string::const_iterator &end,
	int max_width,
	const string &delim,
	float scale_x,
	int offset_x,
	int space_width)
{
	//To keep things simple, we break on newlines when given a non-valid max_width.
	//This way, ccp can call wrap_pos for every line, and word wrapping comes naturally.
	if (max_width < 1)
	{
		string::const_iterator ret = find(start, end, '\n');
		
		if (ret == end)
			return end;
		else
			return ret + 1;
	}
	
	int width = 0;
	
	for (string::const_iterator it = start; it != end; ++it) 
	{
		const unsigned char character = *it;
		const short glyph = ctg(character);
		
		//We can still have manual breaks if we're word wrapping
		if (character == '\n')
			return it + 1;
		else if (glyph != -1)
			width += (int)((gwidth(glyph) + offset_x) * scale_x);
		else
			width += space_width;
		
		if (width > max_width)
		{	
			//If space is a delimiter, and the last entry is a space, tack it to the end of the line
			//Since spaces should be blank, this helps us wrap cleanly, and avoid unintentional indentation
			//Other delimiters might not be blank, and we dont want them sticking over the edge of max_width...
			if ((delim.find(' ') != string::npos) && (*it == ' '))
				return it + 1;
			
			//Try to break the last line after a delimiter
			for (string::const_iterator fdelim = it-1; fdelim >= start; --fdelim) {	
				//Include the breaking delimiter on the first line
				//Otherwise we end up indenting the next line
				if (delim.find(*fdelim) != string::npos)
					return fdelim + 1;
			}
			
			//If we reach the beginning again, no delims on this line 
			return it;
		}
	}
	
	//Whole input fits inside max_width
	return end;
}
