#ifndef header_font_generic
#define header_font_generic

#include "API/Display/sprite.h"
#include "API/Core/Math/origin.h"
#include "API/Display/font.h"
#include "API/Core/Math/point.h"
#include <string>
#include <map>
#include <utility> //For std::pair

class CL_GraphicContext;
class CL_GlyphBuffer;
class CL_Size;
class CL_Rect;

class CL_Font_Generic
{
//! Construction:
public:
	CL_Font_Generic(); 
	CL_Font_Generic(const CL_Font_Generic &copy);
	
	virtual ~CL_Font_Generic();

//! Subclasses:
public:
	//: Abstract for a functor that can be passed to calc_char_positions
	struct CharDest
	{
		virtual ~CharDest() {;}
		
		//: To send a char to the destination
		//param pos: Upper left corner of the position to draw to
		//param character: Character to draw
		//param glyph: The result of a call to ctg(character)
		virtual void operator ()(const CL_Point& pos, unsigned char character, short glyph) = 0;
		
		//: Called exactly once at the beginning of calc_char_positions, sets the angle and hotspot
		virtual void set_rot(float new_ang, const CL_Point &new_hotspot) = 0;
	};
	
	//: Canonical stuff used for most char dests
	struct CanonCharDest : public CharDest
	{
		//: The font implementation, angle of rotation, and rotation hotspot
		//Don't change these under the noses of derivatives: just set them in the ctor or with set_rot
		CL_Font_Generic* fg;
		float ang;
		CL_Point hotspot;
		
		CanonCharDest() : fg(NULL), ang(0.0), hotspot(CL_Point(0, 0)) {;}
		CanonCharDest(CL_Font_Generic* new_fg) : fg(new_fg), ang(0.0), hotspot(CL_Point(0, 0)) {;}
		
		void operator ()(const CL_Point& pos, unsigned char character, short glyph);
		void set_rot(float new_ang, const CL_Point &new_hotspot)
			{ang = new_ang; hotspot = new_hotspot;}
		
		//: To actually draw the character: called by op()
		virtual void drawChar(const CL_Point& pos, unsigned char character, short glyph) = 0;
	};
	
	//: Functor to draw characters to a graphic context
	struct DrawCharDest : public CanonCharDest
	{
		CL_GraphicContext* gc;
		
		DrawCharDest() : gc(NULL) {;}
		
		DrawCharDest(CL_GraphicContext* new_gc, CL_Font_Generic* new_fg)
			: CanonCharDest(new_fg), gc(new_gc) {;}
		
		//Undo the effect of set_rot on destruction (assumes set_rot is called exactly once)
		~DrawCharDest()
			{fg->glyphs.rotate(-ang);}
		
		void set_rot(float new_ang, const CL_Point &new_hotspot)
			{CanonCharDest::set_rot(new_ang, new_hotspot); fg->glyphs.rotate(ang);}
		
		void drawChar(const CL_Point &pos, unsigned char character, short glyph);
	};
	
	//: Functor to place character coordinates into a CL_GlyphBuffer
	struct GlyphBufferDest : public CanonCharDest
	{
		//: Glyph buffer to draw onto (assumes correct font is already marked)
		CL_GlyphBuffer* gb;
		
		GlyphBufferDest(
			CL_GlyphBuffer* new_gb,
			CL_Font_Generic* new_fg)
			: CanonCharDest(new_fg), gb(new_gb) {;} 
		
		void drawChar(const CL_Point& point, unsigned char character, short glyph);
	};
	
	//: Character destination that doesn't do anything (used for bounding rect)
	struct NullCharDest : public CharDest
	{
		void operator ()(const CL_Point&, unsigned char, short) {;}
		void set_rot(float, const CL_Point&) {;}
	};

//! Enums:
public:
	enum Justif
	{
		justif_left,
		justif_center,
		justif_right
	};
	
//! Attributes:
public:
	//: Glyphs of the actual characters.
	CL_Sprite glyphs;
	
	//: Type for a glyph map vector and associated reference count
	//: Several are contained by a single static instance of CTGVector_Cache
	struct CTGVector
	{
		std::vector<short> char_to_glyph_map;
		unsigned int ref_count;
		CTGVector() : ref_count(1)
		{
			//Fill the vector with -1's, symbolizing unimplemented characters
			char_to_glyph_map.reserve(256);
			for (short i = 0; i < 256; ++i)
				char_to_glyph_map.push_back(-1);
		}
	};
	
	//: To map characters to frames of the glyphs sprite: 
	//- <p> Key is the 'letters' string given to create_font),
	//- Value is the map from characters to glyphs, and the reference count. 
	//- We make a seperate wrapper class so that we can initialize the
	//- map with an entry for "" the first time its used.</p>
	struct CTGVector_Cache
	{
		std::map<std::string, CTGVector> cache;
		CTGVector_Cache()
		{
			//Initialize the empty ctg map used for default CL_Fonts
			cache.insert(std::pair<std::string, CTGVector>("", CTGVector()));
		}
	};
	static CTGVector_Cache ctg_vector_cache;
	
	//: Current ctg map we're using, which matches our current letters string
	std::map<std::string, CTGVector>::iterator cur_ctg;
	
	//: Height of font (i.e. height of the tallest character, before scaling/offsets)
	int height;
	
	//: Current baseline distance.
	int baseline;
	
	//: Current width and height offsets.
	int width_offset;
	int height_offset;
	
	//: Space length (distance to go at any unrecognized character)
	int spacelen;
	
	//: Angle and rotation hotspot
	float angle;
	CL_Origin rot_origin;
	int rot_x;
	int rot_y;
	
	//: Alignment
	CL_Origin trans_origin;
	int trans_x;
	int trans_y;
	
	//: Gets the width and height of a string.
	//param max_width: If more than 0 wordwraps to this width
	//param max_height: If more than 0 truncates to this height
	CL_Size get_size(
		const std::string::const_iterator &start,
		const std::string::const_iterator &end,
		int max_width,
		int max_height,
		const std::string &delim,
		float scale_x,
		int offset_x,
		int space_width,
		int line_height);
	
	CL_Size get_size(
		const std::string::const_iterator &start,
		const std::string::const_iterator &end,
		int max_width = 0,
		int max_height = 0,
		const std::string &delim = std::string(" "));
	
	//: Gets the width of a single glyph directly from the map
	//- <p> CL_Font::get_width(unsigned char) does a more complete job.
	//- Don't rely on this except internally.</p>
	int gwidth(short glyph)
		{return glyphs.get_frame_size(glyph).width;}
	
	//: Converts a character into a glyph number
	short ctg(unsigned char character)
		{return cur_ctg->second.char_to_glyph_map[character];}
	
//! Operations:
public:
	//: Copy assignment
	CL_Font_Generic &operator =(const CL_Font_Generic &copy);
	
	//: Lowers reference count on current ctg map in the cache
	void decrement();
	
	//: Sets the ctg map to a given letters string, creates if it neccessary, increments ref count
	void set_letters(const std::string &letters);
	
	//: Loads a system font. Should be called exactly once, immediately after default construction.
	virtual void create_sysfont(
		const std::string &font_name,
		const std::string &letter_chars,
		int height,
		int width,
		bool bold,
		bool italic,
		bool underline,
		bool strikeout) = 0;
	
	//: Loads a regular font. Should be called exactly once, immediately after default construction.
	void create_font(
		const CL_Sprite &letters,
		const std::string &letter_chars,
		int space_width,
		int width_offset,
		int height_offset,
		int baseline);
	
	//: Calculates and pipes the positioning of glyphs under a given justification.
	//return: Bounding rectangle, position after final glyph drawn, rot hotspot, last character drawn exclusive. 
	//param start: The position to begin processing at, inclusive
	//param end: The position to end processing at, exclusive
	//param func: The functor to pipe output through.
	//param dest: The rectangle to draw into (see below)
	//param delim: If max_width is given, these are the characters to seperate words by (besides '\n')
	//- <p> Dest is used along with the current print rotation to determine justification
	//- and wrapping. If the rectangle has a width of zero, then no word wrapping is performed.
	//- If it has a height of zero, then no height trunication is performed. This means you can
	//- turn a CL_Point(x, y) into a CL_Rect(x, y, x, y) and get the expected behavior. </p>
	CL_Font::PrintInfo calc_char_positions(
		const std::string::const_iterator &start,
		const std::string::const_iterator &end,
		CharDest &func,
		const CL_Rect &dest,
		const std::string &delim = std::string(" "));
	
//! Implementation
private:
	//: Finds the last character of this line, exclusive.
	std::string::const_iterator wrap_pos(
		const std::string::const_iterator &start,
		const std::string::const_iterator &end,
		int max_width,
		const std::string &delim,
		float scale_x,
		int offset_x,
		int space_width);
};

#endif
