#include "Display/display_precomp.h"
#include "API/Core/Math/point.h"
#include "API/Core/Math/rect.h"
#include "API/Core/System/error.h"
#include "API/Display/glyph_buffer.h"
#include "API/Display/graphic_context.h"
#include "API/Display/sprite.h"
#include "API/Display/display.h"
#include "API/Display/display_window.h"

#include <cmath>
#include <map>
#include <climits>

using namespace std;

CL_GlyphBuffer::CL_GlyphBuffer() :
	r(1.0f), g(1.0f), b(1.0f), a(1.0f),
	scale_x(1.0f), scale_y(1.0f),
	rot_angle(0.0f), rot_origin(origin_center),
	rot_x(0), rot_y(0), trans_origin(origin_top_left),
	trans_x(0), trans_y(0), blend_src(blend_src_alpha),
	blend_dest(blend_one_minus_src_alpha),
	effect_blend_src(blend_src_alpha),
	effect_blend_dest(blend_one_minus_src_alpha) {}

map<int, CL_Font>::iterator CL_GlyphBuffer::get_font_for(int glyph_index)
{
	if (font_markers.size() == 0)
		return font_markers.end();
	
	if (glyph_index < 0)
		glyph_index = 0;
	
	if (glyph_index > (glyphs.size() - 1))
		glyph_index = glyphs.size() - 1;
	
	//Find a font_marker associated with an index above the given one, or the end
	map<int, CL_Font>::iterator it = font_markers.begin();
	for (; it != font_markers.end(); ++it)
		if (it->first > glyph_index)
			break;
	
	if (it == font_markers.begin())
		return font_markers.end();
	
	return --it;
}

map<int, CL_Font>::const_iterator CL_GlyphBuffer::get_font_for(int glyph_index) const
{
	if (font_markers.size() == 0)
		return font_markers.end();
	
	if (glyph_index < 0)
		glyph_index = 0;
	
	if (glyph_index > (glyphs.size() - 1))
		glyph_index = glyphs.size() - 1;
	
	//Find a font_marker associated with an index above the given one, or the end
	map<int, CL_Font>::const_iterator it = font_markers.begin();
	for (; it != font_markers.end(); ++it)
		if (it->first > glyph_index)
			break;
	
	if (it == font_markers.begin())
		return font_markers.end();
	
	return --it;
}

CL_Rect CL_GlyphBuffer::bounding_rect(const CL_Rect &dest) const
{
	CL_Rect retVal = internal_rect();
	CL_Point offset = i2e_offset(retVal, dest);
	
	retVal.left += offset.x;
	retVal.top += offset.y;
	retVal.right += offset.x;
	retVal.bottom += offset.y;
	
	return retVal;
}

CL_Rect CL_GlyphBuffer::internal_rect() const
{
	//Get a local copy of the GB angle, and the x and y scales
	float angle = rot_angle;
	float x_scale = scale_x; //These names are too similar, but they're the same thing anyways
	float y_scale = scale_y;
	
	//True if we're doing total GlyphBuffer rotation
	bool gb_rotate = false;
	if (angle > 0.05)
		gb_rotate = true;
	
	//True if we're doing total GlyphBuffer scaling
	bool gb_scale = false;
	if ((x_scale < 0.995) || (x_scale > 1.005) || (y_scale < 0.995) || (y_scale > 1.005))
		gb_scale = true;
	
	//Use a weird (and invalid) rectangle to start out with
	//Since we get the bounding rect by just comparing the side of each glyph with a side of the rect-in-progress...
	//Using this rect means the first glyph we compare against determines the initial rectangle
	CL_Rect retVal = CL_Rect(INT_MAX, INT_MAX, INT_MIN, INT_MIN);
	
	//For every glyph in the buffer...
	for (Iter it(*this); !it.at_end(); ++it)
	{
		vector<Glyph>::const_iterator glyph = glyphs.begin() + it.glyph_num; //Reference to the glyph vector element
		
		//Skip glyphs with no frame in the sprite
		short frame_idx = it.get_glyph_index(glyph->character);
		if (frame_idx != -1)
		{
			CL_Size gs = it.sprite.get_frame_size(frame_idx); //Size of the glyph
			CL_Rect gr; //Rectangle containing the glyph
			
			//Set the upper-left coords (ignore the lower-right till after GB rotation)
			gr.left = glyph->pos.x;
			gr.top = glyph->pos.y;
			
			//Get the lower-right coord of the glyph based on upper-left and size
			gr.right = gr.left + gs.width;
			gr.bottom = gr.top + gs.height;
			
			//Find the pre-GB-scaling scale of this glyph
			float spr_scale_x, spr_scale_y;
			it.sprite.get_scale(spr_scale_x, spr_scale_y);
			if (gb_scale) //If GB scaling is enabled, than counter its effect on the sprite scale
			{
				spr_scale_x /= x_scale;
				spr_scale_y /= y_scale;
			}
			
			//Compensate for sprite scaling effects
			if ((spr_scale_x > 1.005) || (spr_scale_x < 0.995) || (spr_scale_y > 1.005) || (spr_scale_x < 0.995))
			{
				//Rescale the rectangle's bounding box away from the upper left corner
				gr.right = gr.left + (int)(gr.get_width() * spr_scale_x);
				gr.bottom = gr.top + (int)(gr.get_height() * spr_scale_y);
			}
			
			//Find the pre-GB-rotation angle of this glyph
			float sprite_angle = it.sprite.get_angle();
			if (gb_rotate) //If GB rotation is enabled, then counter its sprite rotation
				sprite_angle -= angle;
			while(sprite_angle < 0.0)
				sprite_angle += 360;
			while(sprite_angle > 360.0)
				sprite_angle -= 360;
			
			//Compensate for sprite rotation effects and font rotation
			if (sprite_angle > 0.05)
			{
				CL_Origin grot_hotspot;
				int grot_x, grot_y;
				it.sprite.get_rotation_hotspot(grot_hotspot, grot_x, grot_y);
				
				//Rotate the corners of the rectange around the hotspot
				gr = gr.get_rot_bounds(grot_hotspot, grot_x, grot_y, sprite_angle);
			}
			
			//Alter the bounding rect
			retVal.left = min(gr.left, retVal.left);
			retVal.right = max(gr.right, retVal.right);
			retVal.top = min(gr.top, retVal.top);
			retVal.bottom = max(gr.bottom, retVal.bottom);
		}
	}
	
	if (retVal == CL_Rect(INT_MAX, INT_MAX, INT_MIN, INT_MIN))
		retVal = CL_Rect(0,0,0,0);
	
	return retVal;
}

CL_GlyphBuffer &CL_GlyphBuffer::add(
	const CL_GlyphBuffer &other,
	int x,
	int y)
{
	//Offset to make copied font markers/effects sane in the new buffer
	const int offset = glyphs.size();

	glyphs.reserve(glyphs.size() + other.glyphs.size());
	
	if ((x == 0) && (y == 0))
	{
		for (vector<Glyph>::const_iterator it = other.glyphs.begin(); it != other.glyphs.end(); ++it)
			glyphs.push_back(*it);
	}
	else
	{
		for (vector<Glyph>::const_iterator it = other.glyphs.begin(); it != other.glyphs.end(); ++it)
			glyphs.push_back(Glyph(CL_Point(it->pos.x + x, it->pos.y + y), it->character));
	}

	map<int, CL_Font>::const_iterator it_font;
	for (it_font = other.font_markers.begin(); it_font != other.font_markers.end(); ++it_font)
		font_markers.insert(pair<int, CL_Font>(it_font->first + offset, it_font->second));
	
	map<int, CL_Color>::const_iterator it_color;
	for (it_color = other.color_effects.begin(); it_color != other.color_effects.end(); ++it_color)
		color_effects.insert(pair<int, CL_Color>(it_color->first + offset, it_color->second));
	
	map<int, float>::const_iterator it_float;
	for (it_float = other.scale_x_effects.begin(); it_float != other.scale_x_effects.end(); ++it_float)
		scale_x_effects.insert(pair<int, float>(it_float->first + offset, it_float->second));
	for (it_float = other.scale_y_effects.begin(); it_float != other.scale_y_effects.end(); ++it_float)
		scale_y_effects.insert(pair<int, float>(it_float->first + offset, it_float->second));
	for (it_float = other.angle_effects.begin(); it_float != other.angle_effects.end(); ++it_float)
		angle_effects.insert(pair<int, float>(it_float->first + offset, it_float->second));
	
	return *this;
}

CL_GlyphBuffer &CL_GlyphBuffer::change_font(const CL_Font &font)
{
	font_markers.insert(pair<int, CL_Font>(glyphs.size(), font));
	return *this;
}

void CL_GlyphBuffer::clear()
{
	glyphs.clear();
	font_markers.clear();
	angle_effects.clear();
	scale_x_effects.clear();
	scale_y_effects.clear();
	color_effects.clear();
}

void CL_GlyphBuffer::set_angle(float angle)
{
	while(angle >= 360.0f)
		angle -= 360.0f;
	while(angle < 0.0f)
		angle += 360.0f;
	
	rot_angle = angle;
}

CL_GlyphBuffer::Iter::Iter(const CL_GlyphBuffer& new_gb)
	: pgb(&new_gb),
	color_it(new_gb.get_color_effects().begin()),
	scale_x_it(new_gb.get_scale_x_effects().begin()),
	scale_y_it(new_gb.get_scale_y_effects().begin()),
	angle_it(new_gb.get_angle_effects().begin()),
	font_it(new_gb.get_font_markers().begin()),
	font_changed(true) //True so that apply_iters works properly the first time
{
	//It's an error if there are no font markers at all
	if (
		(font_it != new_gb.get_font_markers().end()) &&
		(font_it->first < new_gb.get_glyphs().size()) &&
		(font_it->first >= 0))
	{
		glyph_num = font_it->first;
		apply_iters();
	}
	else
	{
		glyph_num = -1;
	}
}

bool CL_GlyphBuffer::Iter::advance_iters()
{
	//Glyph_num is set to -1 by the ctor if the GB is invalid (empty glyphs or font_markers)
	if (glyph_num == -1) return false;
	
	++glyph_num;
	
	if (glyph_num == pgb->glyphs.size())
	{
		return false;
	}
	else
	{
		//Advance the font marker till it matches glyph_num
		//Font markers affect characters from their index up to the next font marker, or the end
		//So we check to see if the next font marker includes glyph_num, and advance if it does
		const map<int, CL_Font>& font_markers = pgb->get_font_markers();
		map<int, CL_Font>::const_iterator next_font = font_it;
		if (
			(font_it != font_markers.end()) && 
			((++next_font) != font_markers.end()) && //This line modifies our local next_font iter
			(next_font->first <= glyph_num)
		)
		{
			++font_it;
			font_changed = true;
		}
		
		//Advance the effect iterators till they match glyph_num
		//The effect iterators much match glyph_num perfectly to take effect
		while ((color_it != pgb->get_color_effects().end()) && (color_it->first < glyph_num)) ++color_it;
		while ((scale_x_it != pgb->get_scale_x_effects().end()) && (scale_x_it->first < glyph_num)) ++scale_x_it;
		while ((scale_y_it != pgb->get_scale_y_effects().end()) && (scale_y_it->first < glyph_num)) ++scale_y_it;
		while ((angle_it != pgb->get_angle_effects().end()) && (angle_it->first < glyph_num)) ++angle_it;
		
		return true;
	}
}

void CL_GlyphBuffer::Iter::apply_iters()
{
	if (font_changed)
	{
		//Reload the local copy of the sprite (which is modified to match GB properties)
		sprite = font_it->second.get_glyphs();
		
		//Calculate and apply the pre-effect scale
		//This doesnt finish the job, we still need to move the glyphs away from each other
		//But thats done outside Iter
		sprite.get_scale(pe_scale_x, pe_scale_y);
		if ((pgb->scale_x > 1.005) || (pgb->scale_x < 0.995) || (pgb->scale_y > 1.005) || (pgb->scale_x < 0.995))
		{
			pe_scale_x *= pgb->scale_x;
			pe_scale_y *= pgb->scale_y;
			sprite.set_scale(pe_scale_x, pe_scale_y);
		}
		
		//Calculate and apply the pre-effect angle (we also need to apply font rotation to the sprite here)
		//Again, this doesnt finish the job, we still need to rotate the positions of the glyphs for the gb angle
		//But again, that's done outside Iter
		pe_angle = sprite.get_angle();
		if ((pgb->rot_angle > 0.05) || (font_it->second.get_angle() > 0.05))
		{
			if (pgb->rot_angle > 0.05)
				pe_angle += pgb->rot_angle;
			
			if (font_it->second.get_angle() > 0.05)
				pe_angle += font_it->second.get_angle();
			
			while(pe_angle < 0.0)
				pe_angle += 360;
			while(pe_angle > 360.0)
				pe_angle -= 360;
			
			sprite.set_angle(pe_angle);
		}
		
		//FIXME: Calculate and apply the pre-effect GlyphBuffer color property here
		
		font_changed = false;
	}
	
	//For each effect, if there was no effect for this glyph, reload from the original state
	
	//Apply any angle effects
	if ((angle_it != pgb->get_angle_effects().end()) && (angle_it->first == glyph_num))
		sprite.set_angle(angle_it->second + pe_angle);
	else
		sprite.set_angle(pe_angle);
	
	//Apply any scale effects
	if ((scale_x_it != pgb->get_scale_x_effects().end()) || (scale_y_it != pgb->get_scale_y_effects().end()))
	{
		float x = 1.0;
		float y = 1.0;
		
		if ((scale_x_it != pgb->get_scale_x_effects().end()) && (scale_x_it->first == glyph_num))
			x = scale_x_it->second;
		
		if ((scale_y_it != pgb->get_scale_y_effects().end()) && (scale_y_it->first == glyph_num))
			y = scale_y_it->second;
		
		sprite.set_scale(pe_scale_x * x, pe_scale_y * y);
	}
	else
		sprite.set_scale(pe_scale_x, pe_scale_y);
	
	//FIXME: Apply any color effects to the sprite color here
}

CL_GlyphBuffer::Iter& CL_GlyphBuffer::Iter::operator++()
{
	if (advance_iters())
		apply_iters();

	return *this;
}

bool CL_GlyphBuffer::Iter::at_end()
{
	if (glyph_num == -1)
		return true;
	if (glyph_num == pgb->glyphs.size())
		return true;
	
	return false;
}	

CL_Point CL_GlyphBuffer::i2e_offset(const CL_Rect& src, const CL_Rect& dest) const
{
	//Align the source rectangle to the origin
	CL_Point retVal(-src.left, -src.top);
	retVal -= calc_origin(trans_origin, CL_Size(src.get_width(), src.get_height()));
	
	//Align the origin to the dest rectangle
	retVal.x += dest.left;
	retVal.y += dest.top;
	retVal += calc_origin(trans_origin, CL_Size(dest.get_width(), dest.get_height()));
	
	//Apply the translation offsets
	retVal.x -= trans_x;
	retVal.y -= trans_y;
	
	return retVal;
}

CL_Point CL_GlyphBuffer::get_scale_hotspot(const CL_Rect& irect) const
{
	CL_Point scale_hotspot(irect.left, irect.top);
	scale_hotspot += calc_origin(trans_origin, irect.get_size());
	scale_hotspot.x += trans_x;
	scale_hotspot.y += trans_y;
	
	return scale_hotspot;
}

CL_Point CL_GlyphBuffer::get_rot_hotspot(const CL_Rect& irect) const
{
	CL_Point rot_hotspot(irect.left, irect.top);
	rot_hotspot += calc_origin(rot_origin, irect.get_size());
	rot_hotspot.x += rot_x;
	rot_hotspot.y += rot_y;
	
	return rot_hotspot;
}

void CL_GlyphBuffer::draw_glyphs(
	const CL_Point& offset,
	CL_Rect irect,
	CL_GraphicContext* gc)
{
	if (gc == 0)
		gc = CL_Display::get_current_window()->get_gc();
	
	//Get a local copy of the angle, for speed's sake
	float angle = rot_angle;
	float x_scale = scale_x; //These names are too similar, but they're the same thing anyways
	float y_scale = scale_y;
	
	CL_Point rot_hotspot;
	CL_Point scale_hotspot;
	
	//True if we're doing total GlyphBuffer rotation
	bool rotate = false;
	if (angle > 0.05)
	{
		rotate = true;
		
		if (irect == CL_Rect(0,0,0,0))
			irect = internal_rect();
		
		rot_hotspot = get_rot_hotspot(irect);
	}
	
	//True if we're doing total GlyphBuffer scaling
	bool scale = false;
	if ((x_scale < 0.995) || (x_scale > 1.005) || (y_scale < 0.995) || (y_scale > 1.005))
	{
		scale = true;
		
		if (irect == CL_Rect(0,0,0,0))
			irect = internal_rect();
		
		scale_hotspot = get_scale_hotspot(irect);
	}
	
	//For every glyph in the buffer...
	for (Iter it(*this); !it.at_end(); ++it)
	{
		vector<Glyph>::const_iterator glyph = glyphs.begin() + it.glyph_num; //Reference to the glyph vector element
		CL_Point draw_pos(glyph->pos);
		
		//Calculate the scaling if enabled
		if (scale)
		{
			//The offset from scale_hotspot to the new drawing position
			CL_Point scale_offset = draw_pos - scale_hotspot;
			scale_offset.x = (int)(scale_offset.x*scale_x);
			scale_offset.y = (int)(scale_offset.y*scale_y);
			draw_pos = scale_hotspot + scale_offset;
		}
		
		//Calcluate the rotation if enabled
		if (rotate)
			draw_pos = draw_pos.rotate(rot_hotspot, angle);
		
		//Move it into the dest rectangle
		draw_pos += offset;
		
		//Draw it
		short frame_idx = it.get_glyph_index(glyph->character);
		if (frame_idx != -1)
		{
			it.sprite.set_frame(frame_idx);
			it.sprite.draw(draw_pos.x, draw_pos.y);
		}
	}
}
