
#include "Display/display_precomp.h"
#include "resourcedata_sprite.h"
#include "API/Display/sprite_description.h"
#include "API/Display/pixel_buffer.h"
#include "API/Core/System/clanstring.h"
#include "API/Core/System/error.h"
#include "API/Core/Resources/resourceoptions.h"
#include "API/Core/Resources/resource_manager.h"
#include "API/Display/Providers/provider_factory.h"

//////////////////////////////////////////////////////////////////////////////
// CL_ResourceData_Sprite construction:

CL_ResourceData_Sprite::CL_ResourceData_Sprite(CL_Resource &resource)
: CL_ResourceData_SpriteDescription(resource, "sprite")
{
	slots.connect(resource.sig_load_file(), this, &CL_ResourceData_Sprite::on_load_file);
	slots.connect(resource.sig_load_datafile(), this, &CL_ResourceData_Sprite::on_load_datafile);
	slots.connect(resource.sig_save_datafile(), this, &CL_ResourceData_Sprite::on_save_datafile);
	slots.connect(resource.sig_unload(), this, &CL_ResourceData_Sprite::on_unload);
}
	
CL_ResourceData_Sprite::~CL_ResourceData_Sprite()
{
}
	
//////////////////////////////////////////////////////////////////////////////
// CL_ResourceData_Sprite attributes:


//////////////////////////////////////////////////////////////////////////////
// CL_ResourceData_Sprite implementation:

void CL_ResourceData_Sprite::on_load_file()
{
	CL_Resource resource = get_resource();
	CL_ResourceOptions &options = resource.get_options();
	
	CL_SpriteDescription desc = load_description();
	
	bool pack_texture = true;
	if (options.exists("pack_texture"))
	{
		pack_texture = options.get_option("pack_texture").get_value() == "yes";
	}

	sprite = CL_Sprite(desc, pack_texture);
	sprite.resource = resource;

	// Load play options	
	bool play_pingpong = false;
	bool play_loop = true;
	bool play_backward = false;
	if (options.exists("play_options"))
	{
		std::vector<std::string> &play_options = options.get_option("play_options").get_values();
		for(unsigned int i=0; i<play_options.size(); ++i)
		{
			if(play_options[i] == "loop")
				play_loop = true;
			else if(play_options[i] == "noloop")
				play_loop = false;
			else if(play_options[i] == "pingpong")
				play_pingpong = true;
			else if(play_options[i] == "nopingpong")
				play_pingpong = false;
			else if(play_options[i] == "backward")
				play_backward = true;
			else if(play_options[i] == "forward")
				play_backward = false;
		}
	}
	sprite.set_play_backward(play_backward);
	sprite.set_play_loop(play_loop);
	sprite.set_play_pingpong(play_pingpong);

	// Load play default speed
	float play_default_speed = 0.05f;
	if (options.exists("play_speed"))
		play_default_speed = (float)atof(options.get_option("play_speed").get_value().c_str()) / 1000.0f;
	if (options.exists("play_default_speed"))
		play_default_speed = (float)atof(options.get_option("play_default_speed").get_value().c_str()) / 1000.0f;
	int frame_count = sprite.get_frame_count();
	for(int i=0; i<frame_count; ++i)
		sprite.set_frame_delay(i, play_default_speed);

	// Load show on finish
	CL_Sprite::ShowOnFinish show_on_finish = CL_Sprite::show_blank;
	if (options.exists("show_on_finish"))
	{
		std::string str = options.get_option("show_on_finish").get_value().c_str();
		if(str == "blank")
			show_on_finish = CL_Sprite::show_blank;
		else if(str == "last_frame")
			show_on_finish = CL_Sprite::show_last_frame;
		else if(str == "first_frame")
			show_on_finish = CL_Sprite::show_first_frame;
	}
	sprite.set_show_on_finish(show_on_finish);

	// Load base angle
	float base_angle = 0;
	if (options.exists("base_angle"))
		base_angle = (float)atof(options.get_option("base_angle").get_value().c_str());
	sprite.set_base_angle(base_angle);

	// Load rotation hotspot
	CL_Origin origin = origin_center;
	int xoffset = 0;
	int yoffset = 0;
	if (options.exists("rotation_hotspot"))
	{
		int count = options.get_option("rotation_hotspot").get_value_count();
		std::string str = options.get_option("rotation_hotspot").get_value(0).c_str();
		if(str == "top_left")
			origin = origin_top_left;
		else if(str == "top_center")
			origin = origin_top_center;
		else if(str == "top_right")
			origin = origin_top_right;
		else if(str == "center_left")
			origin = origin_center_left;
		else if(str == "center")
			origin = origin_center;
		else if(str == "center_right")
			origin = origin_center_right;
		else if(str == "bottom_left")
			origin = origin_bottom_left;
		else if(str == "bottom_center")
			origin = origin_bottom_center;
		else if(str == "bottom_right")
			origin = origin_bottom_right;
		if(count > 1)
			xoffset = atoi(options.get_option("rotation_hotspot").get_value(1).c_str());
		if(count > 2)
			yoffset = atoi(options.get_option("rotation_hotspot").get_value(2).c_str());
	}
	sprite.set_rotation_hotspot(origin, xoffset, yoffset);

	// Load translation hotspot
	origin = origin_top_left;
	xoffset = 0;
	yoffset = 0;
	if (options.exists("translation_hotspot"))
	{
		int count = options.get_option("translation_hotspot").get_value_count();
		std::string str = options.get_option("translation_hotspot").get_value(0).c_str();
		if(str == "top_left")
			origin = origin_top_left;
		else if(str == "top_center")
			origin = origin_top_center;
		else if(str == "top_right")
			origin = origin_top_right;
		else if(str == "center_left")
			origin = origin_center_left;
		else if(str == "center")
			origin = origin_center;
		else if(str == "center_right")
			origin = origin_center_right;
		else if(str == "bottom_left")
			origin = origin_bottom_left;
		else if(str == "bottom_center")
			origin = origin_bottom_center;
		else if(str == "bottom_right")
			origin = origin_bottom_right;
		if(count > 1)
			xoffset = atoi(options.get_option("translation_hotspot").get_value(1).c_str());
		if(count > 2)
			yoffset = atoi(options.get_option("translation_hotspot").get_value(2).c_str());
	}
	sprite.set_alignment(origin, xoffset, yoffset);

	// Load scale
	float scalex = 1.0f;
	float scaley = 1.0f;
	if (options.exists("scale"))
	{
		scalex = (float)atof(options.get_option("scale").get_value(0).c_str());
		scaley = (float)atof(options.get_option("scale").get_value(1).c_str());
	}
	sprite.set_scale(scalex, scaley);

	// Load color
	float r = 1.0f;
	float g = 1.0f;
	float b = 1.0f;
	float a = 1.0f;
	if (options.exists("color"))
	{
		int count = options.get_option("color").get_value_count();
		r = (float)atof(options.get_option("color").get_value(0).c_str());
		if(count > 1)
			g = (float)atof(options.get_option("color").get_value(1).c_str());
		if(count > 2)
			b = (float)atof(options.get_option("color").get_value(2).c_str());
		if(count > 3)
			a = (float)atof(options.get_option("color").get_value(3).c_str());
	}
	sprite.set_color(r, g, b, a);

	// Load alpha
	if (options.exists("alpha"))
	{
		float alpha = (float)atof(options.get_option("alpha").get_value().c_str());
		sprite.set_alpha(alpha);
	}

	// Load id
	if (options.exists("id"))
	{
		sprite.set_id(atoi(options.get_option("id").get_value().c_str()));
	}

	// Load frameX_speed and frameX_offset
	for(int f=0; f<frame_count; ++f)
	{
		std::string frame_var = "frame";
		frame_var += (f + 1);

		std::string str = frame_var + "_speed";
		if (options.exists(str))
		{
			float speed = (float)atof(options.get_option(str).get_value().c_str()) / 1000.0f;
			sprite.set_frame_delay(f, speed);
		}

		str = frame_var + "_offset";
		if (options.exists(str))
		{
			xoffset = atoi(options.get_option(str).get_value(0).c_str());
			yoffset = atoi(options.get_option(str).get_value(1).c_str());
			sprite.set_frame_offset(f, CL_Point(xoffset, yoffset));
		}
	}
}
	
void CL_ResourceData_Sprite::on_load_datafile(CL_InputSourceProvider *)
{
}
	
void CL_ResourceData_Sprite::on_save_datafile(CL_OutputSourceProvider *)
{
}
	
void CL_ResourceData_Sprite::on_unload()
{
	sprite = CL_Sprite();
}
