
#include "Display/display_precomp.h"
#include "API/Display/display.h"
#include "API/Display/display_window.h"
#include "API/Display/surface.h"
#include "API/Display/graphic_context.h"
#include "API/Display/Providers/provider_factory.h"
#include "API/Core/System/error.h"
#include "API/Core/Resources/resource_manager.h"
#include "surface_generic.h"
#include "surface_target.h"
#include "display_target.h"
#include "resourcedata_surface.h"
#include "graphic_context_generic.h"

/////////////////////////////////////////////////////////////////////////////
// CL_Surface construction:

CL_Surface::CL_Surface()
: impl(new CL_Surface_Generic(0))
{
}

CL_Surface::CL_Surface(
	CL_PixelBuffer *provider,
	bool delete_provider,
	Hint hint)
: impl(0)
{
	impl = new CL_Surface_Generic(
		CL_DisplayTarget::current->create_surface(provider, delete_provider, hint));
}

CL_Surface::CL_Surface(
	const std::string &resource_id,
	CL_ResourceManager *manager)
: impl(new CL_Surface_Generic(0))
{
	CL_Resource &resource = manager->get_resource(resource_id);
	resource.load();

	CL_ResourceData_Surface *data =
		(CL_ResourceData_Surface *) resource.get_data("surface");

	*this = data->get_surface();
}

CL_Surface::CL_Surface(const std::string &filename, Hint hint)
: impl(new CL_Surface_Generic(0))
{
	*this = CL_Surface(CL_ProviderFactory::load(filename), true, hint);
}

CL_Surface::CL_Surface(const CL_Surface &copy)
: impl(new CL_Surface_Generic(0))
{
	*impl = *copy.impl;
	resource = copy.resource;
	resource.load();
}

CL_Surface::~CL_Surface()
{
	delete impl;
	resource.unload();
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface attributes:

int CL_Surface::get_width() const
{
	return impl->target->get_width();
}
	
int CL_Surface::get_height() const
{
	return impl->target->get_height();
}

float CL_Surface::get_angle() const
{
	return impl->angle;
}

void CL_Surface::get_scale(float &x, float &y) const
{
	x = impl->scale_x;
	y = impl->scale_y;
}

float CL_Surface::get_alpha() const
{
	return impl->alpha;
}

void CL_Surface::get_color(float &red, float &green, float &blue, float &alpha)
{
	red = impl->red;
	green = impl->green;
	blue = impl->blue;
	alpha = impl->alpha;
}

void CL_Surface::get_blend_func(CL_BlendFunc &src, CL_BlendFunc &dest) const
{
	src = impl->blend_src;
	dest = impl->blend_dest;
}

void CL_Surface::get_alignment(CL_Origin &origin, int &x, int &y) const
{
	x = (int)impl->translation_hotspot.x;
	y = (int)impl->translation_hotspot.y;
	origin = impl->translation_origin;
}

void CL_Surface::get_rotation_hotspot(CL_Origin &origin, int &x, int &y) const
{
	x = (int)impl->rotation_hotspot.x;
	y = (int)impl->rotation_hotspot.y;
	origin = impl->rotation_origin;
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface operations:

CL_Surface &CL_Surface::operator =(const CL_Surface &copy)
{
	resource.unload();
	*impl = *copy.impl;
	resource = copy.resource;
	resource.load();
	return *this;
}

void CL_Surface::draw(
	int x,
	int y,
	CL_GraphicContext *gc)
{
	if (gc == 0) gc = CL_Display::get_current_window()->get_gc();
	gc->impl->select_gc();
	impl->target->draw(impl, x, y, gc);
}

void CL_Surface::draw(
	const CL_Rect &dest,
	CL_GraphicContext *gc)
{
	if (gc == 0) gc = CL_Display::get_current_window()->get_gc();
	gc->impl->select_gc();
	impl->target->draw(impl, dest, gc);
}

void CL_Surface::draw(
	const CL_Rect &src,
	const CL_Rect &dest,
	CL_GraphicContext *gc)
{
	if (gc == 0) gc = CL_Display::get_current_window()->get_gc();
	gc->impl->select_gc();
	impl->target->draw(impl, src, dest, gc);
}

void CL_Surface::set_angle(float angle)
{
	impl->angle = angle;

	while(impl->angle >= 360.0f)
		impl->angle -= 360.0f;
	while(impl->angle < 0.0f)
		impl->angle += 360.0f;
}

void CL_Surface::rotate(float angle)
{
	impl->angle += angle;

	while(impl->angle >= 360.0f)
		impl->angle -= 360.0f;
	while(impl->angle < 0.0f)
		impl->angle += 360.0f;
}

void CL_Surface::set_scale(float x, float y)
{
	impl->scale_x = x;
	impl->scale_y = y;
}

void CL_Surface::set_alpha(float alpha)
{
	impl->alpha = alpha;
}

void CL_Surface::set_color(float r, float g, float b, float a)
{
	impl->red = r;
	impl->green = g;
	impl->blue = b;
	impl->alpha = a;
}

void CL_Surface::set_blend_func(CL_BlendFunc src, CL_BlendFunc dest)
{
	impl->blend_src = src;
	impl->blend_dest = dest;
}

void CL_Surface::set_alignment(CL_Origin origin, int x, int y)
{
	impl->translation_hotspot.x = (float)x;
	impl->translation_hotspot.y = (float)y;
	impl->translation_origin = origin;
}

void CL_Surface::set_rotation_hotspot(CL_Origin origin, int x, int y)
{
	impl->rotation_hotspot.x = (float)x;
	impl->rotation_hotspot.y = (float)y;
	impl->rotation_origin = origin;
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface implementation:
