
#include <cstdio>
#include <iostream>

#include "display_window_opengl.h"
#include "Display/display_precomp.h"
#include "Display/display_window_generic.h"
#include "API/Core/System/keep_alive.h"
#include "graphic_context_opengl.h"
#include "GL/pixel_buffer_opengl_frame.h"
#include "API/Core/System/error.h"
#include "GL/gl_state_checker.h"
#include "Display/X11/input_device_x11keyboard.h"
#include "Display/X11/input_device_x11mouse.h"

GLXContext CL_DisplayWindow_OpenGL::share_context = 0;

/////////////////////////////////////////////////////////////////////////////
// Construction:

CL_DisplayWindow_OpenGL::CL_DisplayWindow_OpenGL() :
	input_context(1, 1, 0),
	fullscreen(false), fullscreen_width(0), fullscreen_height(0),
	saved_position(0, 0, 0, 0),
	left_ctrl_down(false), left_alt_down(false), left_shift_down(false),
	right_ctrl_down(false), right_alt_down(false), right_shift_down(false),
	context(0),
	disp(0),
	window(0)
{
	gc = CL_GraphicContext(new CL_GraphicContext_OpenGL(this));
}

CL_DisplayWindow_OpenGL::~CL_DisplayWindow_OpenGL()
{
	if(fullscreen)
		set_windowed();
	destroy_window();
}

/////////////////////////////////////////////////////////////////////////////
// Attributes:

int CL_DisplayWindow_OpenGL::get_width() const
{
	XWindowAttributes attr;
	XGetWindowAttributes(disp, window, &attr);

	return attr.width;
}

int CL_DisplayWindow_OpenGL::get_height() const
{
	XWindowAttributes attr;
	XGetWindowAttributes(disp, window, &attr);

	return attr.height;
}

bool CL_DisplayWindow_OpenGL::is_fullscreen() const
{
	return fullscreen;
}

bool CL_DisplayWindow_OpenGL::has_focus() const
{
	return focus;
}

CL_PixelBuffer &CL_DisplayWindow_OpenGL::get_buffer(int i)
{
	if (i == 0) return buffer_front;
	else return buffer_back;
}

const CL_PixelBuffer &CL_DisplayWindow_OpenGL::get_buffer(int i) const
{
	if (i == 0) return buffer_front;
	else return buffer_back;
}

int CL_DisplayWindow_OpenGL::get_buffer_count() const
{
	return 2;
}

CL_GraphicContext *CL_DisplayWindow_OpenGL::get_gc()
{
	return &gc;
}

/////////////////////////////////////////////////////////////////////////////
// Operations:

void CL_DisplayWindow_OpenGL::create_window(const CL_DisplayWindowDescription &desc)
{
	fullscreen_width = desc.get_size().width;
	fullscreen_height = desc.get_size().height;

	XVisualInfo *vi;
	Colormap cmap;

	disp = XOpenDisplay(0);
	if (disp == 0) throw CL_Error("Could not open X11 display!");

	static int gl_attribs_single[] =
	{
		GLX_RGBA,
		GLX_RED_SIZE, 4,
		GLX_GREEN_SIZE, 4,
		GLX_BLUE_SIZE, 4,
		GLX_DEPTH_SIZE, 16,
		None
	};
	
	static int gl_attribs_double[] =
	{
		GLX_RGBA,
		GLX_DOUBLEBUFFER,
		GLX_RED_SIZE, 4,
		GLX_GREEN_SIZE, 4,
		GLX_BLUE_SIZE, 4,
		GLX_DEPTH_SIZE, 16,
		None 
	};

	// get an appropriate visual
	vi = glXChooseVisual(disp, DefaultScreen(disp), gl_attribs_double);
	
	if (vi == NULL)
	{
		vi = glXChooseVisual(disp, window, gl_attribs_single);
		printf("Only Singlebuffered Visual!\n");
	}
	else
	{
		printf("Got Doublebuffered Visual!\n");
	}

	// create a GLX context
	context = glXCreateContext(disp, vi, share_context, GL_TRUE);

	if( share_context == NULL )
		share_context = context;

	glXGetConfig(disp, vi, GLX_BUFFER_SIZE, &glx_bpp);

	// create a color map
	cmap = XCreateColormap( disp, RootWindow(disp, vi->screen),  vi->visual, AllocNone);

	attributes.colormap = cmap;
	attributes.border_pixel = 0;
	attributes.override_redirect = False;

	// create a window in window mode
	attributes.event_mask =
		ExposureMask |
		KeyPressMask |
		KeyReleaseMask |
		ButtonPressMask |
		ButtonReleaseMask |
		StructureNotifyMask |
		PointerMotionMask |
		EnterWindowMask |
		LeaveWindowMask |
		FocusChangeMask;

	window = XCreateWindow(disp, RootWindow(disp, vi->screen),
		0, 0, desc.get_size().width, desc.get_size().height, 0, vi->depth, InputOutput, vi->visual,
		CWBorderPixel | CWColormap | CWOverrideRedirect | CWEventMask, &attributes);
	
	XSelectInput(disp, window,
		FocusChangeMask | KeyPressMask | KeyReleaseMask
		| PropertyChangeMask | StructureNotifyMask |
		KeymapStateMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask);

	// set title of window:
	set_title(desc.get_title());

	// handle wm_delete_events if in windowed mode:
	Atom wm_delete = XInternAtom(disp, "WM_DELETE_WINDOW", True);
	XSetWMProtocols(disp, window, &wm_delete, 1);

	// make window visible:
	XMapRaised(disp, window);

	select_context();
	gc.impl->setup_2d();

	if (glXIsDirect(disp, context))
		printf("Congrats, you have Direct Rendering!\n");
	else
		printf("Sorry, no Direct Rendering possible!\n");

	// Create input devices for window:
	keyboard = CL_InputDevice(new CL_InputDevice_X11Keyboard(this));
	mouse = CL_InputDevice(new CL_InputDevice_X11Mouse(this));
	get_ic()->get_keyboard(0) = keyboard;
	get_ic()->get_mouse(0) = mouse;

	buffer_front = CL_PixelBuffer(new CL_PixelBuffer_OpenGL_Frame(GL_FRONT, gc));
	buffer_back = CL_PixelBuffer(new CL_PixelBuffer_OpenGL_Frame(GL_BACK, gc));
	
	XSync(disp, True);
	
	focus = true;
	
	system_cursor = XCreateFontCursor(disp, XC_left_ptr);
	char *data = (char*)malloc(64); // 8x8
	memset(data, 0, 64);

	XColor black_color;
	memset(&black_color, 0, sizeof(black_color));

	cursor_bitmap = XCreateBitmapFromData(disp, window, data, 8, 8);
	hidden_cursor = XCreatePixmapCursor(disp, cursor_bitmap, cursor_bitmap, &black_color, &black_color, 0,0 );
	
	if (desc.is_fullscreen())
		set_fullscreen(desc.get_size().width, desc.get_size().height, desc.get_bpp());

}

void CL_DisplayWindow_OpenGL::destroy_window()
{
	set_windowed();

	if (context)
	{
		if (!glXMakeCurrent(disp, None, NULL))
		{
			printf("Could not release drawing context.\n");
		}

		glXDestroyContext(disp, context);
		context = NULL;
	}

	keyboard = CL_InputDevice();
	mouse = CL_InputDevice();
	get_ic()->get_keyboard(0) = keyboard;
	get_ic()->get_mouse(0) = mouse;

	if (window) XDestroyWindow(disp, window);
	window = 0;
	
	if (disp) XCloseDisplay(disp);
	disp = 0;
}

void CL_DisplayWindow_OpenGL::on_xevent(XEvent &event)
{
	//Disabled for now
}

void CL_DisplayWindow_OpenGL::flip()
{
	CL_GLStateChecker::flush();
	XSync(disp,False);
	glXSwapBuffers(disp, window);
}

void CL_DisplayWindow_OpenGL::select_context()
{
	CL_GLStateChecker::set_context(disp, window, context);
/*
	bool retval = glXMakeCurrent(
		disp,
		window,
		context);

	if (retval != true) throw CL_Error("Could not make window current OpenGL rendering context!");
*/
}

void CL_DisplayWindow_OpenGL::set_fullscreen(int width, int height, int bpp)
{

	if(fullscreen)
		return;

	//Hide Window
	XUnmapWindow(disp, window);
	XSync(disp, True);


	// Turn off WM control
	attributes.override_redirect = True;
	XChangeWindowAttributes(disp, window, CWBorderPixel | CWColormap | CWOverrideRedirect, &attributes);

	XMoveResizeWindow(disp, window, 0, 0, width, height);


	
	
	// Re-appear window
	XMapRaised(disp, window);
	XSync(disp, True);

	// Get input focus
	XSetInputFocus (disp,window, RevertToNone, CurrentTime);

	// Vid-mode Switching
	int num_modes;
	XF86VidModeModeInfo **modes;
	XF86VidModeModeLine cur_mode;
	std::list<XF86VidModeModeInfo *> usable_modes;

	XF86VidModeGetModeLine(disp, 0, &dotclock, &cur_mode);

	old_mode.dotclock = dotclock;
	old_mode.hdisplay = cur_mode.hdisplay;
	old_mode.hsyncstart = cur_mode.hsyncstart;
	old_mode.hsyncend = cur_mode.hsyncend;
	old_mode.htotal = cur_mode.htotal;
	old_mode.vdisplay = cur_mode.vdisplay;
	old_mode.vsyncstart = cur_mode.vsyncstart;
	old_mode.vsyncend = cur_mode.vsyncend;
	old_mode.vtotal = cur_mode.vtotal;
	old_mode.flags = cur_mode.flags;
	old_mode.privsize = 0;

	XF86VidModeGetAllModeLines(disp, 0, &num_modes, &modes);

	for(int i = 0; i < num_modes; i++)
	{
		if(modes[i]->hdisplay == width && modes[i]->vdisplay == height) {
			std::cout << "Mode: "<< width << "x" << height << std::endl;
			usable_modes.push_back(modes[i]);
		}
	}

	int x,y;

	set_position(0,0);

	// Move the mouse and switch moves
	XWarpPointer(disp, None, window, 0, 0, 0, 0, 0, 0);
	XF86VidModeSwitchToMode(disp, 0, *(usable_modes.begin()));
	XWarpPointer(disp, None, window, 0, 0, 0, 0, 0, 0);

	
	XGrabPointer(disp, window, True, 0, GrabModeAsync, GrabModeAsync, window, None, CurrentTime);
	
	XF86VidModeGetViewPort(disp, 0, &x, &y);

	set_position(x, y);
	
	old_x = x;
	old_y = y;

	std::cout << "Viewport is: " << x << " " << y << std::endl;

	XSync(disp, True);

	fullscreen = true;

}

void CL_DisplayWindow_OpenGL::set_windowed()
{
	if(!fullscreen)
		return;

	// Hide window
	XUnmapWindow(disp, window);
	XSync(disp, True);

	// Turn on WM control
	attributes.override_redirect = False;
	XChangeWindowAttributes(disp, window, CWBorderPixel | CWColormap | CWOverrideRedirect, &attributes);

	XMoveResizeWindow(disp, window, 0, 0, fullscreen_width, fullscreen_height);

	XMapRaised(disp, window);
	XSync(disp, True);

	fullscreen = false;

		
	XF86VidModeSwitchToMode(disp, 0, &old_mode);
	XF86VidModeSetViewPort(disp, 0, old_x, old_y);

	XUngrabPointer(disp, CurrentTime);

	XSync(disp, True);

}

void CL_DisplayWindow_OpenGL::set_title(const std::string &title)
{
	XSetStandardProperties(disp, window, title.c_str(), title.c_str(), None, NULL, 0, NULL);
}

void CL_DisplayWindow_OpenGL::set_position(const CL_Rect &pos)
{
	set_position(pos.left, pos.top);
	XResizeWindow(disp, window, pos.get_width(), pos.get_height());
	set_size(pos.get_width(), pos.get_height());
}

void CL_DisplayWindow_OpenGL::set_position(int x, int y)
{
	XMoveWindow(disp, window, x, y);
}

void CL_DisplayWindow_OpenGL::set_size(int width, int height)
{
	fullscreen_width = width;
	fullscreen_height = height;
	sig_resize(width, height);
}

void CL_DisplayWindow_OpenGL::set_buffer_count(int flipping_buffers)
{
// Not implemented.
// We can't change how many buffer we use in OpenGL, the drivers pick for us
#ifdef DEBUG
	std::cout << "CL_DisplayWindow_OpenGL::set_buffer_count is not implemented" << std::endl;
#endif
}

void CL_DisplayWindow_OpenGL::update(const CL_Rect &rect)
{
	CL_GLStateChecker::flush();
	select_context();

	glReadBuffer(GL_BACK);

	if(glGetError() == GL_NO_ERROR)		// Check if we got more than 1 buffer.
	{
		glDisable(GL_TEXTURE_2D);

		float raster_pos[4];

		glGetFloatv(GL_CURRENT_RASTER_POSITION, raster_pos);

		glDrawBuffer(GL_FRONT);
		glRasterPos3f(rect.left, rect.bottom, raster_pos[2]);

		glCopyPixels(rect.top, rect.left, rect.get_width(), rect.get_height(), GL_COLOR);

		// Return to drawing to the back buffer
		glDrawBuffer(GL_BACK);
		
		glEnable(GL_TEXTURE_2D);
	}
}

void CL_DisplayWindow_OpenGL::show_system_cursor()
{
	XDefineCursor(disp, window, system_cursor);
}

void CL_DisplayWindow_OpenGL::hide_system_cursor()
{
	XDefineCursor(disp, window, hidden_cursor);
}

void CL_DisplayWindow_OpenGL::cursor_changed()
{
// We have to extract the data from the cursor in order to display it

// At this point nothing more than hiding/unhiding the system cursor is supported
// I'm waiting for the release of XFree86 4.3, which has the Xcursor extension library.
// This will greatly simplify this, so I don't want to start implenting this.
// --MSR, Nov. 13, 2002
}

void CL_DisplayWindow_OpenGL::keep_alive()
{
	XEvent event;
	CL_Rect *rect;
	
	for (int i=XPending(disp); i>0; i--)
	{
		XNextEvent(disp, &event);
		switch(event.type)
		{
			//Resize or Move
			case ConfigureNotify:
				#ifdef DEBUG
					std::cout << "ConfigureNotify Event received" << std::endl;
				#endif
				set_size(event.xconfigure.width,event.xconfigure.height);
				break;
			case ClientMessage:
				#ifdef DEBUG
					std::cout << "Received ClientMessage, sending close signal" << std::endl;
				#endif
				sig_window_close();
				break;
			case Expose:
				// Repaint notification
				// Could be more efficient if we checked ahead for other
				// repaint notifications
				#ifdef DEBUG
					std::cout << "Expose Event received" << std::endl;
				#endif
				rect = new CL_Rect(event.xexpose.x, event.xexpose.y,
					event.xexpose.x + event.xexpose.width, event.xexpose.y + event.xexpose.height);
				sig_paint(*rect);
				delete rect;
				break;
			case FocusIn:
				focus = true;
				sig_got_focus();
				break;
			case FocusOut:
				focus = false;
				sig_lost_focus();
				break;
			case KeyPress:
			case KeyRelease:
			case ButtonPress:
			case ButtonRelease:
			case MotionNotify:
				sig_xevent(event);
				break;
			#ifdef DEBUG
			case EnterNotify:
			case LeaveNotify:
				std::cout << "The mouse has left or entered the window" << std::endl;
				break;
			default:
				std::cout << "Unhandled event type: " << event.type << std::endl;
			#endif
		}
	}
}
