
#include "Display/display_precomp.h"
#include "canvas_opengl.h"
#include "GL/gl.h"
#include "pixel_buffer_opengl_texture.h"
#include "API/Display/display.h"
#include "API/Display/display_window.h"
#include "Display/graphic_context_generic.h"
#include "graphic_context_opengl.h"
#include "gl_state_checker.h"

#ifdef WIN32
#pragma warning(disable : 4355) // 'this' : used in base member initializer list
#endif

/////////////////////////////////////////////////////////////////////////////
// CL_Canvas_OpenGL construction:

CL_Canvas_OpenGL::CL_Canvas_OpenGL(
	const CL_Size &size,
	CL_Canvas::Format canvas_format,
	bool use_colorkey,
	unsigned int colorkey)
{
	canvas_size = size;
	
	texture_size.width = 1;
	texture_size.height = 1;
	while (texture_size.width < canvas_size.width) texture_size.width *= 2;
	while (texture_size.height < canvas_size.height) texture_size.height *= 2;

	// Convert provider to OpenGL texture format ABGR:
	CL_PixelFormat abgr_format;
	abgr_format.set_depth(32);
	abgr_format.enable_colorkey(false);
	abgr_format.set_alpha_mask(0xFF000000);
	abgr_format.set_blue_mask (0x00FF0000);
	abgr_format.set_green_mask(0x0000FF00);
	abgr_format.set_red_mask  (0x000000FF);

	CL_PixelBuffer *buffer = new CL_PixelBuffer(
		texture_size.width,
		texture_size.height,
		texture_size.width * 4,
		abgr_format);

	buffer->lock();
	memset(buffer->get_data(), 0, buffer->get_pitch()*buffer->get_height());

	// Upload to OpenGL:
	glGenTextures(1, &texture_handle);
	glBindTexture(GL_TEXTURE_2D, texture_handle);

	glTexImage2D(
		GL_TEXTURE_2D,          // target
		0,                      // level
		GL_RGBA,                // internalformat
		texture_size.width,		// width
		texture_size.height,	// height
		0,                      // border
		GL_RGBA,                // format
		GL_UNSIGNED_BYTE,       // type
		buffer->get_data());    // texels

	// Texture mode parameters must be setup before creating textures!
	// Dunno if this is a driver bug in both nvidia and my laptops s3.
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);

	buffer->unlock();
	delete buffer;

	pixelbuffer = CL_PixelBuffer(
		new CL_PixelBuffer_OpenGL_Texture(
			0, 0,
			canvas_size.width, canvas_size.height,
			texture_handle,
			CL_Display::get_current_window()->get_gc()));

	gc = new CL_GraphicContext(new CL_GraphicContext_OpenGL(this));
}

CL_Canvas_OpenGL::CL_Canvas_OpenGL(
	CL_PixelBuffer *provider,
	bool delete_provider,
	CL_Canvas::Format canvas_format,
	bool use_colorkey,
	unsigned int colorkey)
{
	// Convert provider to OpenGL texture format ABGR:
	CL_PixelFormat abgr_format;
	abgr_format.set_depth(32);
	abgr_format.enable_colorkey(false);
	abgr_format.set_alpha_mask(0xFF000000);
	abgr_format.set_blue_mask (0x00FF0000);
	abgr_format.set_green_mask(0x0000FF00);
	abgr_format.set_red_mask  (0x000000FF);

	provider->lock();

	canvas_size.width = provider->get_width();
	canvas_size.height = provider->get_height();
	
	texture_size.width = 1;
	texture_size.height = 1;
	while (texture_size.width < canvas_size.width) texture_size.width *= 2;
	while (texture_size.height < canvas_size.height) texture_size.height *= 2;

	CL_PixelBuffer *buffer = new CL_PixelBuffer(
		texture_size.width,
		texture_size.height,
		texture_size.width * 4,
		abgr_format);

	provider->convert(
		buffer->get_data(), abgr_format,
		buffer->get_pitch(),
		CL_Rect(0, 0, canvas_size.width, canvas_size.height),
		CL_Rect(0, 0, canvas_size.width, canvas_size.height));

	provider->unlock();
	if (delete_provider) delete provider;

	buffer->lock();

	// Upload to OpenGL:
	glGenTextures(1, &texture_handle);
	glBindTexture(GL_TEXTURE_2D, texture_handle);

	glTexImage2D(
		GL_TEXTURE_2D,          // target
		0,                      // level
		GL_RGBA,                // internalformat
		texture_size.width,		// width
		texture_size.height,	// height
		0,                      // border
		GL_RGBA,                // format
		GL_UNSIGNED_BYTE,       // type
		buffer->get_data());    // texels

	// Texture mode parameters must be setup before creating textures!
	// Dunno if this is a driver bug in both nvidia and my laptops s3.
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);

	buffer->unlock();
	delete buffer;

	pixelbuffer = CL_PixelBuffer(
		new CL_PixelBuffer_OpenGL_Texture(
			0, 0,
			canvas_size.width, canvas_size.height,
			texture_handle,
			CL_Display::get_current_window()->get_gc()));

	gc = new CL_GraphicContext(new CL_GraphicContext_OpenGL(this));
}

CL_Canvas_OpenGL::~CL_Canvas_OpenGL()
{
	delete gc;
}

/////////////////////////////////////////////////////////////////////////////
// CL_Canvas_OpenGL attributes:

CL_PixelBuffer &CL_Canvas_OpenGL::get_pixelbuffer()
{
	return pixelbuffer;
}

CL_GraphicContext *CL_Canvas_OpenGL::get_gc()
{
	return gc;
}

/////////////////////////////////////////////////////////////////////////////
// CL_Canvas_OpenGL operations:

void CL_Canvas_OpenGL::draw(int x, int y, CL_GraphicContext *gc)
{
	gc->impl->select_gc();

	float w = texture_size.width;
	float h = texture_size.height;
	float sw = canvas_size.width;
	float sh = canvas_size.height;

	CL_Rectf pos(x, y, x + sw, y + sh);
	CL_Rectf tex_coords(0, 0, sw / w, sh / h);

	CL_GLState state;
	state.blendfunc[0] = blend_src_alpha; // attributes->blend_src;
	state.blendfunc[1] = blend_one_minus_src_alpha; // attributes->blend_dest;
	state.color[0] = 1.0f; // attributes->red;
	state.color[1] = 1.0f; // attributes->green;
	state.color[2] = 1.0f; // attributes->blue;
	state.color[3] = 1.0f; // attributes->alpha;
	state.texture_handle = texture_handle;
	state.rotation_hotspot = CL_Pointf(0,0);
	state.rotate_angle = 0;

	CL_GLStateChecker::add_quad(state, pos, tex_coords);

/*
	float w = (float) texture_size.width;
	float h = (float) texture_size.height;
	float sw = (float) canvas_size.width;
	float sh = (float) canvas_size.height;

	glBindTexture(GL_TEXTURE_2D, texture_handle);

	// Draw surface
	glBegin(GL_TRIANGLE_STRIP);
		glTexCoord2f(0.0f, 0.0f);     glVertex2f(0, 0);
		glTexCoord2f(sw / w, 0.0f);   glVertex2f(sw, 0);
		glTexCoord2f(0.0f, sh / h);   glVertex2f(0, sh);
		glTexCoord2f(sw / w, sh / h); glVertex2f(sw, sh);
	glEnd();
*/
}

void CL_Canvas_OpenGL::select_context()
{
}

/////////////////////////////////////////////////////////////////////////////
// CL_Canvas_OpenGL implementation:
