
#include "Display/display_precomp.h"
#include "GL/gl_state.h"
#include "GL/gl_state_checker.h"
#include "API/Display/blend_func.h"
#include "API/Display/color.h"
#include "API/Core/Math/rect.h"
#include "API/Core/Math/point.h"

/////////////////////////////////////////////////////////////////////////////
// CL_GLStateChecker construction:

CL_GLStateChecker::CL_GLStateChecker()
{
	internal_state.blendfunc[0] = blend_src_alpha;
	internal_state.blendfunc[1] = blend_one_minus_src_alpha;
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	internal_state.color[0] = 255;
	internal_state.color[1] = 255;
	internal_state.color[2] = 255;
	internal_state.color[3] = 255;
	glColor4fv(internal_state.color);

	internal_state.texture_handle = -1;
	glDisable(GL_TEXTURE_2D);

	internal_state.rotate_angle = 0.0f;
	
	begin_primitive = 0;
}

/////////////////////////////////////////////////////////////////////////////
// CL_GLStateChecker operations:

void CL_GLStateChecker::add_clear(
	const CL_Color &color)
{
	glClearColor(color.get_red() / 255.0f, color.get_green() / 255.0f, color.get_blue() / 255.0f, color.get_alpha() / 255.0f);
	glClear(GL_COLOR_BUFFER_BIT);
}

void CL_GLStateChecker::add_pixel(
	const CL_GLState &state,
	const CL_Pointf &position)
{
	set_state(state, GL_POINTS + 1);

	glVertex2f(position.x + 0.375f, position.y + 0.375f);
}

void CL_GLStateChecker::add_line(
	const CL_GLState &state,
	const CL_Rectf &position)
{
	set_state(state, GL_LINES + 1);

	glVertex2f(position.left + 0.375f, position.top + 0.375f);
	glVertex2f(position.right + 0.375f, position.bottom + 0.375f);
}

void CL_GLStateChecker::add_gradient(
	const CL_GLState &state,
	const CL_Rectf &position,
	const CL_Gradient &gradient)
{
	set_state(state, GL_QUADS + 1);

	float px1 = position.left;
	float py1 = position.top;
	float px2 = position.right;
	float py2 = position.bottom;

	// Top left vertex
	glVertex2f(px1, py1);

	// Bottom left vertex
	glColor4f(
		gradient.bottom_left.get_red() / 255.0f,
		gradient.bottom_left.get_green() / 255.0f,
		gradient.bottom_left.get_blue() / 255.0f,
		gradient.bottom_left.get_alpha() / 255.0f);
	glVertex2f(px1, py2);

	// Bottom right vertex
	glColor4f(
		gradient.bottom_right.get_red() / 255.0f,
		gradient.bottom_right.get_green() / 255.0f,
		gradient.bottom_right.get_blue() / 255.0f,
		gradient.bottom_right.get_alpha() / 255.0f);
	glVertex2f(px2, py2);

	// Top right vertex
	internal_state.color[0] = gradient.top_right.get_red() / 255.0f;
	internal_state.color[1] = gradient.top_right.get_green() / 255.0f;
	internal_state.color[2] = gradient.top_right.get_blue() / 255.0f;
	internal_state.color[3] = gradient.top_right.get_alpha() / 255.0f;
	glColor4fv(internal_state.color);
	glVertex2f(px2, py1);
}

void CL_GLStateChecker::add_quad(
	const CL_GLState &state,
	const CL_Rectf &position,
	const CL_Rectf &tex_coords)
{
	set_state(state, GL_QUADS + 1);

	float px1 = position.left;
	float py1 = position.top;
	float px2 = position.right;
	float py2 = position.bottom;

	float tx1 = tex_coords.left;
	float ty1 = tex_coords.top;
	float tx2 = tex_coords.right;
	float ty2 = tex_coords.bottom;

	if (state.rotate_angle == 0.0f)
	{
		glTexCoord2f(tx1, ty2); glVertex2f(px1, py2);
		glTexCoord2f(tx2, ty2); glVertex2f(px2, py2);
		glTexCoord2f(tx2, ty1); glVertex2f(px2, py1);
		glTexCoord2f(tx1, ty1); glVertex2f(px1, py1);
	}
	else
	{
		glTexCoord2f(tx1, ty2); glVertex2f(calc_rotate_x(px1, py2), calc_rotate_y(px1, py2));
		glTexCoord2f(tx2, ty2); glVertex2f(calc_rotate_x(px2, py2), calc_rotate_y(px2, py2));
		glTexCoord2f(tx2, ty1); glVertex2f(calc_rotate_x(px2, py1), calc_rotate_y(px2, py1));
		glTexCoord2f(tx1, ty1); glVertex2f(calc_rotate_x(px1, py1), calc_rotate_y(px1, py1));
	}
}

void CL_GLStateChecker::flush()
{
	if (begin_primitive)
	{
		glEnd();
		begin_primitive = 0;
	}
}

#ifdef WIN32

void CL_GLStateChecker::set_context(HDC hdc, HGLRC context)
{
	if (cur_drawable != hdc && cur_context != context)
	{
		flush();
		wglMakeCurrent(hdc, context);
		cur_drawable = hdc;
		cur_context = context;
	}
}

#else

void CL_GLStateChecker::set_context(Display *disp, GLXDrawable drawable, GLXContext context)
{
	if (cur_drawable != drawable && cur_context != context)
	{
		flush();
		glXMakeCurrent(disp, drawable, context);
		cur_drawable = drawable;
		cur_context = context;
	}
}

#endif

/////////////////////////////////////////////////////////////////////////////
// CL_GLStateTree implementation:

void CL_GLStateChecker::set_state(const CL_GLState &state, GLenum primitive)
{
	if (internal_state.texture_handle != state.texture_handle)
	{
		internal_state.texture_handle = state.texture_handle;
		
		if (begin_primitive)
		{
			glEnd();
			begin_primitive = 0;
		}
		
		if(state.texture_handle == -1)
		{
			glDisable(GL_TEXTURE_2D);
		}
		else
		{
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, (GLuint)state.texture_handle);
		}
	}
	
	if (internal_state.blendfunc[0] != state.blendfunc[0] || 
		internal_state.blendfunc[1] != state.blendfunc[1])
	{
		if (begin_primitive)
		{
			glEnd();
			begin_primitive = 0;
		}

		internal_state.blendfunc[0] = state.blendfunc[0];
		internal_state.blendfunc[1] = state.blendfunc[1];
		
		GLenum blendfunc[2];
		for(int i = 0; i < 2; ++i)
		{			
			switch (state.blendfunc[i])
			{
			case blend_zero:
				blendfunc[i] = GL_ZERO;
				break;
			case blend_one:
				blendfunc[i] = GL_ONE;
				break;
			case blend_dest_color:
				blendfunc[i] = GL_DST_COLOR;
				break;
			case blend_src_color:
				blendfunc[i] = GL_SRC_COLOR;
				break;
			case blend_one_minus_dest_color:
				blendfunc[i] = GL_ONE_MINUS_DST_COLOR;
				break;
			case blend_one_minus_src_color:
				blendfunc[i] = GL_ONE_MINUS_SRC_COLOR;
				break;
			case blend_src_alpha:
				blendfunc[i] = GL_SRC_ALPHA;
				break;
			case blend_one_minus_src_alpha:
				blendfunc[i] = GL_ONE_MINUS_SRC_ALPHA;
				break;
			case blend_dst_alpha:
				blendfunc[i] = GL_DST_ALPHA;
				break;
			case blend_one_minus_dest_alpha:
				blendfunc[i] = GL_ONE_MINUS_DST_ALPHA;
				break;
			case blend_src_alpha_saturate:
				blendfunc[i] = GL_SRC_ALPHA_SATURATE;
				break;
			}
		}
		glBlendFunc(blendfunc[0], blendfunc[1]);
	}

	if (internal_state.color[0] != state.color[0] ||
		internal_state.color[1] != state.color[1] ||
		internal_state.color[2] != state.color[2] ||
		internal_state.color[3] != state.color[3])
	{
		internal_state.color[0] = state.color[0];
		internal_state.color[1] = state.color[1];
		internal_state.color[2] = state.color[2];
		internal_state.color[3] = state.color[3];

		glColor4fv(state.color);
	}

	if (primitive != begin_primitive)
	{
		if (begin_primitive)
		{
			glEnd();
			begin_primitive = 0;
		}

		glBegin(primitive - 1);
		begin_primitive = primitive;
	}

	internal_state.rotation_hotspot = state.rotation_hotspot;

	if (internal_state.rotate_angle != state.rotate_angle)
	{
		internal_state.rotate_angle = state.rotate_angle;

		double angle_rad = 3.1415926f*state.rotate_angle/180;
		rotate_x.x = cos(angle_rad);
		rotate_x.y = sin(angle_rad);
		rotate_y.x = cos(3.1415926f/2+angle_rad);
		rotate_y.y = sin(3.1415926f/2+angle_rad);
	}
}

CL_Pointf CL_GLStateChecker::rotate_x(1.0f, 0.0f);

CL_Pointf CL_GLStateChecker::rotate_y(0.0f, 1.0f);

CL_GLStateChecker gl_state_tree;

CL_GLState CL_GLStateChecker::internal_state;

GLenum CL_GLStateChecker::begin_primitive;

#ifdef WIN32
	HDC CL_GLStateChecker::cur_drawable = 0;

	HGLRC CL_GLStateChecker::cur_context = 0;
#else
	GLXDrawable CL_GLStateChecker::cur_drawable = 0;

	GLXContext CL_GLStateChecker::cur_context = 0;
#endif
