
#include "Display/display_precomp.h"
#include <iostream>
#include "Display/display_window_generic.h"
#include "graphic_context_opengl.h"
#include "canvas_opengl.h"
#include "API/Display/color.h"
#include "GL/gl_state_checker.h"
#include <GL/gl.h>
#include <GL/glu.h>

#ifndef GL_CLAMP_TO_EDGE_EXT
#define GL_CLAMP_TO_EDGE_EXT 0x812F
#endif

/////////////////////////////////////////////////////////////////////////////
// CL_GraphicContext_OpenGL construction:

CL_GraphicContext_OpenGL::CL_GraphicContext_OpenGL(CL_DisplayWindow_Generic *window)
: window(window), canvas(0), in_2d(false)
{
	slots.connect(window->sig_resize, this, &CL_GraphicContext_OpenGL::on_resize);
}

CL_GraphicContext_OpenGL::CL_GraphicContext_OpenGL(CL_Canvas_OpenGL *canvas)
: window(0), canvas(canvas), in_2d(false)
{
}

CL_GraphicContext_OpenGL::~CL_GraphicContext_OpenGL()
{
}

/////////////////////////////////////////////////////////////////////////////
// CL_GraphicContext_OpenGL attributes:

int CL_GraphicContext_OpenGL::get_width() const
{
	if (canvas) return canvas->get_pixelbuffer().get_width();
	return window->get_width();
}

int CL_GraphicContext_OpenGL::get_height() const
{
	if (canvas) return canvas->get_pixelbuffer().get_height();
	return window->get_height();
}

/////////////////////////////////////////////////////////////////////////////
// CL_GraphicContext_OpenGL operations:

void CL_GraphicContext_OpenGL::draw_pixel(int x, int y, const CL_Color &color)
{
	select_gc();

	CL_GLState state;
	state.texture_handle = -1; // disable texture
	state.color[0] = color.get_red()/255.0f;
	state.color[1] = color.get_green()/255.0f;
	state.color[2] = color.get_blue()/255.0f;
	state.color[3] = color.get_alpha()/255.0f;
	state.blendfunc[0] = blend_src_alpha;
	state.blendfunc[1] = blend_one_minus_src_alpha;
	state.rotate_angle = 0.0f;
	
	CL_Pointf pos((float)x, (float)y);

	CL_GLStateChecker::add_pixel(state, pos);
}

void CL_GraphicContext_OpenGL::draw_line(int x1, int y1, int x2, int y2, const CL_Color &color)
{
	select_gc();

	CL_GLState state;
	state.texture_handle = -1; // disable texture
	state.color[0] = color.get_red()/255.0f;
	state.color[1] = color.get_green()/255.0f;
	state.color[2] = color.get_blue()/255.0f;
	state.color[3] = color.get_alpha()/255.0f;
	state.blendfunc[0] = blend_src_alpha;
	state.blendfunc[1] = blend_one_minus_src_alpha;
	state.rotate_angle = 0.0f;

	CL_Rectf pos((float)x1, (float)y1, (float)x2, (float)y2);

	CL_GLStateChecker::add_line(state, pos);
}

void CL_GraphicContext_OpenGL::draw_rect(const CL_Rect &rect, const CL_Color &color)
{
	select_gc();

	int x1 = rect.left;
	int y1 = rect.top;
	int x2 = rect.right;
	int y2 = rect.bottom;

	CL_GLState state;
	state.texture_handle = -1; // disable texture
	state.color[0] = color.get_red()/255.0f;
	state.color[1] = color.get_green()/255.0f;
	state.color[2] = color.get_blue()/255.0f;
	state.color[3] = color.get_alpha()/255.0f;
	state.blendfunc[0] = blend_src_alpha;
	state.blendfunc[1] = blend_one_minus_src_alpha;
	state.rotate_angle = 0.0f;

	// Because of buggy OpenGL drivers, I couldn't use LINE_LOOP.

	CL_GLStateChecker::add_line(state, CL_Rectf(x1, y1, x2 - 1, y1));
	CL_GLStateChecker::add_line(state, CL_Rectf(x2 - 1, y1, x2 - 1, y2 - 1));
	CL_GLStateChecker::add_line(state, CL_Rectf(x1 + 1, y2 - 1, x2, y2 - 1));
	CL_GLStateChecker::add_line(state, CL_Rectf(x1, y1 + 1, x1, y2));
}

void CL_GraphicContext_OpenGL::fill_rect(const CL_Rect &rect, const CL_Color &color)
{
	select_gc();

	CL_GLState state;
	state.texture_handle = -1;
	state.color[0] = color.get_red()/255.0f;
	state.color[1] = color.get_green()/255.0f;
	state.color[2] = color.get_blue()/255.0f;
	state.color[3] = color.get_alpha()/255.0f;
	state.blendfunc[0] = blend_src_alpha;
	state.blendfunc[1] = blend_one_minus_src_alpha;
	state.rotate_angle = 0.0f;

	static CL_Rectf tex_coords(0, 0, 1.0f, 1.0f);
	CL_Rectf pos(rect.left, rect.top, rect.right, rect.bottom);
	CL_GLStateChecker::add_quad(state, pos, tex_coords);
}

void CL_GraphicContext_OpenGL::fill_rect(const CL_Rect &rect, const CL_Gradient &gradient)
{
	select_gc();

	CL_GLState state;
	state.texture_handle = -1;
	state.color[0] = gradient.top_left.get_red()/255.0f;
	state.color[1] = gradient.top_left.get_green()/255.0f;
	state.color[2] = gradient.top_left.get_blue()/255.0f;
	state.color[3] = gradient.top_left.get_alpha()/255.0f;
	state.blendfunc[0] = blend_src_alpha;
	state.blendfunc[1] = blend_one_minus_src_alpha;
	state.rotate_angle = 0.0f;

	CL_Rectf pos(rect.left, rect.top, rect.right, rect.bottom);
	CL_GLStateChecker::add_gradient(state, pos, gradient);
}

void CL_GraphicContext_OpenGL::clear(const CL_Color &color)
{
	select_gc();
	CL_GLStateChecker::add_clear(color);
}

/////////////////////////////////////////////////////////////////////////////
// CL_GraphicContext_OpenGL implementation:

void CL_GraphicContext_OpenGL::select_gc()
{
	if (canvas) canvas->select_context();
	else window->select_context();
}

void CL_GraphicContext_OpenGL::setup_2d()
{
	bool use_2d_projection = true;

	in_2d = true;

	glPushAttrib(GL_ALL_ATTRIB_BITS);	// todo: narrow range of what is saved

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glDisable(GL_FOG);
	glDisable(GL_DEPTH_TEST);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glColor4f(1.0, 1.0, 1.0, 1.0);

	if (use_2d_projection)
	{
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();

		GLdouble viewport[4];
		glGetDoublev(GL_VIEWPORT, viewport);

		GLdouble width = viewport[2];
		GLdouble height = viewport[3];

		gluOrtho2D(0.0, width, height, 0.0);
	}

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	if (use_2d_projection)
	{
		glLoadIdentity();
	}

	glPushMatrix();
	glTranslatef(get_translate_x(), get_translate_y(), 0);

	const CL_Rect &clip = get_cliprect();

	glEnable(GL_SCISSOR_TEST);
	update_scissor();
}

void CL_GraphicContext_OpenGL::setup_3d()
{
	CL_GLStateChecker::flush();

	in_2d = false;

	glPopMatrix();
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glPopAttrib();
}

void CL_GraphicContext_OpenGL::set_cliprect(const CL_Rect &rect)
{
	CL_GraphicContext_Generic::set_cliprect(rect);
	if (in_2d)update_scissor();
}

void CL_GraphicContext_OpenGL::push_cliprect(const CL_Rect &rect)
{
	CL_GraphicContext_Generic::push_cliprect(rect);
	if (in_2d) update_scissor();
}

void CL_GraphicContext_OpenGL::push_cliprect()
{
	CL_GraphicContext_Generic::push_cliprect();
	if (in_2d) update_scissor();
}

void CL_GraphicContext_OpenGL::pop_cliprect()
{
	CL_GraphicContext_Generic::pop_cliprect();
	if (in_2d) update_scissor();
}

void CL_GraphicContext_OpenGL::set_translate_offset(int x, int y)
{
	CL_GraphicContext_Generic::set_translate_offset(x, y);
	if (in_2d)
	{
		select_gc();
		CL_GLStateChecker::flush();
		glPopMatrix();
		glPushMatrix();
		glTranslatef(get_translate_x(), get_translate_y(), 0);
	}
}

void CL_GraphicContext_OpenGL::push_translate_offset(int delta_x, int delta_y)
{
	CL_GraphicContext_Generic::push_translate_offset(delta_x, delta_y);
	if (in_2d)
	{
		select_gc();
		CL_GLStateChecker::flush();
		glPopMatrix();
		glPushMatrix();
		glTranslatef(get_translate_x(), get_translate_y(), 0);
	}
}

void CL_GraphicContext_OpenGL::push_translate_offset()
{
	CL_GraphicContext_Generic::push_translate_offset();
	if (in_2d)
	{
		select_gc();
		CL_GLStateChecker::flush();
		glPopMatrix();
		glPushMatrix();
		glTranslatef(get_translate_x(), get_translate_y(), 0);
	}
}

void CL_GraphicContext_OpenGL::pop_translate_offset()
{
	CL_GraphicContext_Generic::pop_translate_offset();
	if (in_2d)
	{
		select_gc();
		CL_GLStateChecker::flush();
		glPopMatrix();
		glPushMatrix();
		glTranslatef(get_translate_x(), get_translate_y(), 0);
	}
}

void CL_GraphicContext_OpenGL::update_scissor()
{
	select_gc();
	CL_GLStateChecker::flush();

	GLdouble viewport[4];
	glGetDoublev(GL_VIEWPORT, viewport);

	GLdouble &viewport_width = viewport[2];
	GLdouble &viewport_height = viewport[3];

	const CL_Rect &clip = get_cliprect();
	int clip_width = clip.get_width();
	int clip_height = clip.get_height();

	glScissor(clip.left, viewport_height-clip.top-clip_height, clip_width, clip_height);
}

void CL_GraphicContext_OpenGL::on_resize(int width, int height)
{
	cliprects.pop_back();
	cliprects.push_back(CL_Rect(0, 0, width, height));

	select_gc();
	glViewport(0, 0, (GLint) width, (GLint) height);

	if (in_2d)
	{
		update_scissor();

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluOrtho2D(0.0, width, height, 0.0);
		glMatrixMode(GL_MODELVIEW);
	}
}
