
#ifndef header_graphic_context_opengl
#define header_graphic_context_opengl

#include "Display/graphic_context_generic.h"
#include "API/signals.h"

class CL_DisplayWindow_Generic;
class CL_Canvas_OpenGL;

//: OpenGL graphic context
class CL_GraphicContext_OpenGL : public CL_GraphicContext_Generic
{
//! Construction:
public:
	CL_GraphicContext_OpenGL(CL_DisplayWindow_Generic *window);

	CL_GraphicContext_OpenGL(CL_Canvas_OpenGL *canvas);

	virtual ~CL_GraphicContext_OpenGL();

//! Attributes:
public:
	//: Returns the current width of the context.
	virtual int get_width() const;

	//: Returns the current height of the context.
	virtual int get_height() const;

//! Operations:
public:
	//: Called prior to every drawal command.
	virtual void select_gc();

	//: Draw a pixel at (x, y) using the specified color.
	virtual void draw_pixel(int x, int y, const CL_Color &color);

	//: Draw a line from (x1, y1) to (x2, y2) using the specified color.
	virtual void draw_line(int x1, int y1, int x2, int y2, const CL_Color &color);

	//: Draw a rectangle using the specified color.
	virtual void draw_rect(const CL_Rect &rect, const CL_Color &color);

	//: Draw a gradient rectangle using the specified gradient.
	virtual void fill_rect(const CL_Rect &rect, const CL_Gradient &gradient);

	//: Draw a filled rectangle using the specified color.
	virtual void fill_rect(const CL_Rect &rect, const CL_Color &color);

	//: Clears the whole context using the specified color.
	virtual void clear(const CL_Color &color);

	//: Set the current clipping rectangle.
	virtual void set_cliprect(const CL_Rect &rect);

	//: Push current clipping rectangle to stack; then set clipping rectangle to the
	//: union of the current rectangle and the passed rectangle.
	virtual void push_cliprect(const CL_Rect &rect);

	//: Push current clipping rectangle to the stack.
	virtual void push_cliprect();

	//: Pop current clipping rectangle from the stack.
	virtual void pop_cliprect();

	//: Sets the effective translate offset to (x,y).
	virtual void set_translate_offset(int x, int y);
	
	//: Push translation offset onto translation stack.
	//- <p>This offset will
	//- affect any subsequent display operations on the current displaycard, by
	//- translating the position of the display operation with the offset.
	//- The offset will be offset by any previous offsets pushed onto the stack,
	//- eg. it inherits the previous offset.</p>
	virtual void push_translate_offset(int delta_x, int delta_y);
	
	//: Sets the translation offset as a new absolute translation offset.
	//- <p>The new offset will disregard any previous offset's, but will not
	//- empty the translation stack. The new translation offset will affect
	//- any subsequent display operations on the current displaycard, by
	//- translating the position of the display operation with the offset.</p>
	virtual void push_translate_offset();
	
	//: Pops the last pushed translation offset from the translation offset
	//: stack.
	//- <p>If the stack is empty, nothing will happen, and if the last
	//- translation offset is popped, the translation offset will be set to 0,0.</p>
	virtual void pop_translate_offset();

	//: Called when 2D operations are begun.
	virtual void setup_2d();

	//: Called when 2D operations are ended.
	virtual void setup_3d();

//! Implementation:
private:
	//: Updates glScissor with current clip rect.
	void update_scissor();

	//: Called when window is resized.
	void on_resize(int width, int height);

	//: True when currently in 2d mode.
	bool in_2d;

	//: OpenGL window.
	//- Can be null if attached to a canvas instead.
	CL_DisplayWindow_Generic *window;

	//: OpenGL canvas.
	//- Can be null if attached to a window instead.
	CL_Canvas_OpenGL *canvas;

	//: Slots.
	CL_SlotContainer slots;
};

#endif
