
#include "Display/display_precomp.h"
#include "surface_opengl.h"
#include "API/Display/pixel_buffer.h"
#include "API/Display/graphic_context.h"
#include "API/Display/pixel_format.h"
#include <Hermes/H_Conv.h>

#include <GL/glu.h>

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_OpenGL construction:

CL_Surface_OpenGL::CL_Surface_OpenGL(
	CL_PixelBuffer *provider,
	bool delete_provider,
	CL_Surface::Hint hint)
{
	// Convert provider to OpenGL texture format ABGR:
	CL_PixelFormat abgr_format;
	abgr_format.set_depth(32);
	abgr_format.enable_colorkey(false);
	abgr_format.set_alpha_mask(0xFF000000);
	abgr_format.set_blue_mask (0x00FF0000);
	abgr_format.set_green_mask(0x0000FF00);
	abgr_format.set_red_mask  (0x000000FF);

	provider->lock();

	screen_size.width = (float)provider->get_width();
	screen_size.height = (float)provider->get_height();
	
	texture_size.width = 1;
	texture_size.height = 1;
	while (texture_size.width < screen_size.width) texture_size.width *= 2;
	while (texture_size.height < screen_size.height) texture_size.height *= 2;

	CL_PixelBuffer *buffer = new CL_PixelBuffer(
		texture_size.width,
		texture_size.height,
		texture_size.width * 4,
		abgr_format);

	provider->convert(
		buffer->get_data(), abgr_format,
		buffer->get_pitch(),
		CL_Rect(0, 0, screen_size.width, screen_size.height),
		CL_Rect(0, 0, screen_size.width, screen_size.height));

	provider->unlock();
	if (delete_provider) delete provider;

	buffer->lock();

	// Upload to OpenGL:
	glGenTextures(1, &handle);
	glBindTexture(GL_TEXTURE_2D, handle);

	glTexImage2D(
		GL_TEXTURE_2D,          // target
		0,                      // level
		GL_RGBA,                // internalformat
		texture_size.width,		// width
		texture_size.height,	// height
		0,                      // border
		GL_RGBA,                // format
		GL_UNSIGNED_BYTE,       // type
		buffer->get_data());    // texels

	// Texture mode parameters must be setup before creating textures!
	// Dunno if this is a driver bug in both nvidia and my laptops s3.
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);

	buffer->unlock();
	delete buffer;
}

CL_Surface_OpenGL::~CL_Surface_OpenGL()
{
	glDeleteTextures(1, &handle);
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_OpenGL attributes:

int CL_Surface_OpenGL::get_width() const
{
	return (int)screen_size.width;
}
	
int CL_Surface_OpenGL::get_height() const
{
	return (int)screen_size.height;
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_OpenGL operations:

//: Draw surface on screen.
void CL_Surface_OpenGL::draw(
	CL_Surface_Generic *attributes,
	int x, int y,
	CL_GraphicContext *gc)
{
	float w = texture_size.width;
	float h = texture_size.height;
	float sw = screen_size.width;
	float sh = screen_size.height;

	// Calculate translation hotspot
	CL_Pointf translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		screen_size);

//	CL_GLState state = init_attributes(attributes, screen_size);
	init_attributes(attributes, screen_size);

	if (attributes->scale_x == 1.0f && attributes->scale_y == 1.0f)
	{
		float x_pos = x - translation.x;
		float y_pos = y - translation.y;
		state.rotation_hotspot.x += x_pos;
		state.rotation_hotspot.y += y_pos;

		CL_Rectf pos(x_pos, y_pos, x_pos + sw, y_pos + sh);
		CL_Rectf tex_coords(0, 0, sw / w, sh / h);

		CL_GLStateChecker::add_quad(state, pos, tex_coords);
	}
	else
	{
		float x_pos = x - translation.x * attributes->scale_x;
		float y_pos = y - translation.y * attributes->scale_y;
		state.rotation_hotspot.x = x_pos + state.rotation_hotspot.x * attributes->scale_x;
		state.rotation_hotspot.y = y_pos + state.rotation_hotspot.y * attributes->scale_y;

		CL_Rectf pos(
			x_pos,
			y_pos,
			x_pos + sw * attributes->scale_x,
			y_pos + sh * attributes->scale_y);

		CL_Rectf tex_coords(0, 0, sw / w, sh / h);

		CL_GLStateChecker::add_quad(state, pos, tex_coords);
	}
}

//: Draw surface on screen.
void CL_Surface_OpenGL::draw(
	CL_Surface_Generic *attributes,
	const CL_Rect &dest,
	CL_GraphicContext *gc)
{
	float w = texture_size.width;
	float h = texture_size.height;
	float sw = (float)dest.get_width();
	float sh = (float)dest.get_height();
	CL_Sizef size(sw, sh);
	
	// Calculate translation hotspot
	CL_Pointf translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		size);

//	CL_GLState state = init_attributes(attributes, size);
	init_attributes(attributes, size);

	if (attributes->scale_x == 1.0f && attributes->scale_y == 1.0f)
	{
		CL_Rectf pos(dest.left-translation.x, dest.top-translation.y, dest.right-translation.x, dest.bottom-translation.y);
		CL_Rectf tex_coords(0, 0, sw/w, sh/h);
		state.rotation_hotspot.x += pos.left;
		state.rotation_hotspot.y += pos.top;
		CL_GLStateChecker::add_quad(state, pos, tex_coords);
	}
	else
	{
		translation.x *= attributes->scale_x;
		translation.y *= attributes->scale_y;

		CL_Rectf pos(
			dest.left-translation.x,
			dest.top-translation.y,
			dest.left+(dest.right-dest.left)*attributes->scale_x-translation.x,
			dest.top+(dest.bottom-dest.top)*attributes->scale_y-translation.y);

		CL_Rectf tex_coords(0, 0, sw/w, sh/h);
		state.rotation_hotspot.x = pos.left + state.rotation_hotspot.x * attributes->scale_x;
		state.rotation_hotspot.y = pos.top + state.rotation_hotspot.y * attributes->scale_y;
		CL_GLStateChecker::add_quad(state, pos, tex_coords);
	}
}

//: Draw surface on screen.
void CL_Surface_OpenGL::draw(
	CL_Surface_Generic *attributes,
	const CL_Rect &src,
	const CL_Rect &dest,
	CL_GraphicContext *gc)
{
	float w = texture_size.width;
	float h = texture_size.height;
	CL_Sizef size((float)dest.get_width(), (float)dest.get_height());

	// Calculate translation hotspot
	CL_Pointf translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		size);

	init_attributes(attributes, size);

	if (attributes->scale_x == 1.0f && attributes->scale_y == 1.0f)
	{
		CL_Rectf pos(dest.left-translation.x, dest.top-translation.y, dest.right-translation.x, dest.bottom-translation.y);
		CL_Rectf tex_coords(src.left / w, src.top / h, src.right / w, src.bottom / h);
		state.rotation_hotspot.x += pos.left;
		state.rotation_hotspot.y += pos.top;

		CL_GLStateChecker::add_quad(state, pos, tex_coords);
	}
	else
	{
		translation.x *= attributes->scale_x;
		translation.y *= attributes->scale_y;

		CL_Rectf pos(
			dest.left-translation.x,
			dest.top-translation.y,
			dest.left+(dest.right-dest.left)*attributes->scale_x-translation.x,
			dest.top+(dest.bottom-dest.top)*attributes->scale_y-translation.y);

		CL_Rectf tex_coords(src.left / w, src.top / h, src.right / w, src.bottom / h);
		state.rotation_hotspot.x = pos.left + state.rotation_hotspot.x * attributes->scale_x;
		state.rotation_hotspot.y = pos.top + state.rotation_hotspot.y * attributes->scale_y;

		CL_GLStateChecker::add_quad(state, pos, tex_coords);
	}
}

/////////////////////////////////////////////////////////////////////////////
// CL_Surface_OpenGL implementation:

/*
	glBindTexture(GL_TEXTURE_2D, handle);

	// Store matrix
	glPushMatrix();

	// Calculate translation hotspot
	CL_Point translation = calc_hotspot(
		attributes->translation_origin,
		attributes->translation_hotspot,
		CL_Size(width, height));

	// Translate to translation hotspot
	glTranslatef(-translation.x + x, -translation.y + y, 0.0f);

	// Scale surface
	if(attributes->scale_x != 1.0f || attributes->scale_y != 1.0f)
	{
		// Translate back to origin
		glTranslatef(translation.x, translation.y, 0.0f);

		// Scale surface
		glScalef(attributes->scale_x, attributes->scale_y, 1.0f);

		// Translate for scaling
		glTranslatef(-translation.x, -translation.y, 0.0f);
	}

	// Rotate surface
	if(attributes->angle)
	{
		// Calculate rotation hotspot
		CL_Point rotation = calc_hotspot(
			attributes->rotation_origin,
			attributes->rotation_hotspot,
			CL_Size(width, height));

		// Translate back to origin
		glTranslatef(rotation.x, rotation.y, 0.0f);

		// Rotate surface
		glRotatef(attributes->angle, 0, 0, 1);

		// Translate to rotation hotspot
		glTranslatef(-rotation.x, -rotation.y, 0.0f);
	}

	// TODO: Set ALL blend func
	if(attributes->blend_src == blend_src_alpha && attributes->blend_dest == blend_one)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	else
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
*/
